-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

----------------------------------------------------------------
-- Abstract C and pretty-printing code
----------------------------------------------------------------

module AbstractC
	( module AbstractC
        , module Id
	) where

import PrettyUtils
import Id

----------------------------------------------------------------
-- Parse trees
----------------------------------------------------------------

type Decl = (QualId,Type)
type Init = (Decl, Initializer)

data Type 
  = Function1 ![Ident]              !Bool !Type [Attribute]
  | Function2 ![(Maybe Ident,Type)] !Bool !Type [Attribute]
  | SimpleTy !QualId
  | TypeOf !Expr
  | Ptr !Type
  | Array !(Maybe Expr) !Type
  | TyEnum   !QualId	-- we always lift the bodies of structs, etc out to top level
  | TyUnion  !QualId	-- leaving just a reference behind.
  | TyStruct !QualId
  | Const !Type
  | TyAttribute ![Attribute] !Type
  | TyApp !Type !Type
  | ParensType !Type
 deriving Show

voidType :: Type
voidType = SimpleTy (mkQualId "void")

intType :: Type
intType = SimpleTy (mkQualId "int")

data Defn
  = Enum   !QualId ![(QualId, Maybe Expr)]
  | Union  !QualId ![(Maybe Ident, Maybe Expr, Type)]
  | Struct !QualId ![(Maybe Ident, Maybe Expr, Type)]
  | Typedef !QualId !Type
  | Extern !Defn
  | Fun !Decl [Decl] !Compound
  | Init !Init
  | TopAsm Bool String [AsmArg] [AsmArg] [AsmArg]
 deriving Show

data Initializer
  = None
  | SimpleInit !Expr
  | FieldInit !Ident !Initializer
  | ArrayInit !Expr !Initializer
  | ListInit ![Initializer]
 deriving Show

data Attribute
  = Mode [Attribute]
  | Format [Attribute]
  | Attr_Constant String
  | SimpleAttr String
{-
  | Aligned
  | Unused
  | Transparent_Union
  | DI
  | Printf
  | Scanf
  | NoReturn
  | Attr_Const
  | StdCall
  | COM
  | C
  | CPU
-}
 deriving Show

type Op = String
ppOp = text

data Expr 
  = Ident !QualId
  | Constant !String
  | String !String
  | Constructor !Type !Initializer
  | App !Expr [Expr]
  | Select !Ident !Expr
  | Arrow  !Ident !Expr -- redundant
  | Cast Type !Expr
  | SizeOf_E !Expr
  | SizeOf_T !Type
  | LabelAddr !Label
  | Cond !Expr !Expr !Expr
  | Cond2 !Expr !Expr -- GNU extension x ?: y == x ? x : y
  | UnOp !Op !Expr
  | BinOp !Op !Expr !Expr
  | StatementExpr !Compound
 deriving Show

parensExpr = UnOp "(_)"

type Label = Ident

-- ToDo: can't handle local typedef/struct decls :-(
type Compound = ([Label], [Init], [Statement])
data Statement 
  = Return !(Maybe Expr)
  | Break
  | Continue
  | Goto !Ident
  | If Expr !Statement !Statement
  | Switch !Expr !Statement
  | Label !Label !Statement
  | Case !Expr !Statement
  | Default !Statement
  | While !Expr !Statement
  | Do !Statement !Expr
  | For !(Maybe Expr) !(Maybe Expr) !(Maybe Expr) !Statement
  | Compound !Compound
  | Expr !Expr
  | Asm Bool String [AsmArg] [AsmArg] [AsmArg]
  | Empty
 deriving Show

type AsmArg = (String, Maybe Expr)

----------------------------------------------------------------
-- list of names defined in a bunch of Defns
-- excluding those from externs!
----------------------------------------------------------------

names :: [Defn] -> ([QualId],[QualId])
names ds = foldr name ([],[]) ds

name :: Defn -> ([QualId],[QualId]) -> ([QualId],[QualId])
name (Enum   t _)        (x,y) = (x,t:y)
name (Union  t _)        (x,y) = (x,t:y)
name (Struct t _)        (x,y) = (x,t:y)
name (Typedef i _)       (x,y) = (i:x,y)
name (Extern d)          (x,y) = name d (x,y)
name (Fun (i,_) _ _)     (x,y) = (i:x,y)
name (Init ((i,_),_))    (x,y) = (i:x,y)
name _ xy = xy

----------------------------------------------------------------
-- End
----------------------------------------------------------------