-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module AbstractUnits
	( BundleTIdent, UnitIdent, BundleIdent, QualIdent, FlagsIdent
        , PackageIdent, MemberIdent
        , Global
	, FileName
	, Depend, ObjectSet(..), DependRelation(..)
        , Constraint(..), Property(..), PropIdent, TypeIdent
        , Clause, Formula(..), Atomic(..), AProp(..)
        , RuleIdent
	, Bundle(..), BundleElement
	, UnitInstance(..)
	, PosArgs, NameArgs, Arg
        , Glue(..)
	, Bindings, Binding(..)
	, BTElement(..), BundleType
        , Flag(..), AtomicUnit, RelativeFileName(..), relativeName
        , FileWithFlags, Rename(..)
	, Import, Export, Unit, Flags
        , ModUnit, RemUnit, AddUnit
	, Defn(..)
        , module Id
	) where

import Id

----------------------------------------------------------------
-- The abstract syntax of units
----------------------------------------------------------------

data Defn
  = BundleT BundleType
  | Unit    Unit
  | ModUnit ModUnit
  | Flags   Flags
  | Package [Glue]
  | Prop  
  | Type    [TypeIdent]
  | Rule    [Clause]
  | Doc     String

----------------------------------------------------------------
-- Bundle Types
----------------------------------------------------------------

type BundleType = ([BTElement],Flags)
data BTElement
  = Ident       MemberIdent
  | Extends     BundleTIdent  -- more or less #include
  | Include     FileName      -- suck in this file to get C type info

----------------------------------------------------------------
-- Units
----------------------------------------------------------------

-- atomic or compound units
type Unit =
  ( [Import]
  , [Export]
  , [RuleIdent]
  , [Constraint]
  , [(QualIdent,ObjectSet)] -- concrete initializers
  , [(QualIdent,ObjectSet)] -- concrete finalizers
  , [Depend]
  , Maybe Bool                 -- should it be flattened
  , Either AtomicUnit Bindings
  )

type Import = (BundleIdent, BundleType)
type Export = (BundleIdent, BundleType)

-- modified units
type ModUnit = (UnitIdent, RemUnit, AddUnit)

type RemUnit =
  ( [BundleIdent] -- imports
  , [BundleIdent] -- exports
  , [Constraint]
  , [Depend]
  , [BundleIdent] -- globals
  , [Glue]
  , [BundleIdent] -- defaults
  , [BundleIdent] -- bindings
  )

type AddUnit =
  ( [Import]
  , [Export]
  , [Constraint]
  , [Depend]
  , [BundleIdent] -- globals
  , [Glue]
  , [BundleIdent] -- defaults
  , [Binding]     
  )

----------------------------------------------------------------
-- Atomic units
----------------------------------------------------------------

type AtomicUnit = ([FileWithFlags], [Rename])

type FileWithFlags = (RelativeFileName, Flags)

data Rename
  = RenameTo     (BundleIdent,Ident) Ident -- rename x.y to z
  | RenamePrefix BundleIdent Ident         -- rename x with prefix
  | RenameSuffix BundleIdent Ident         -- rename x with suffix

data RelativeFileName
   = SrcDir   (Maybe FileName) FileName
   | AnonCode [String]

type Flags = [Flag]

data Flag
  = UseFlags FlagsIdent
  | RawFlag  String

----------------------------------------------------------------
-- Compound units
----------------------------------------------------------------

type Bindings = ([Global], [Glue], [BundleIdent], [Binding])

type Global = BundleIdent

data Glue 
  = GlueIdent   UnitIdent
  | GluePackage PackageIdent
 deriving (Eq)

data Binding 
  = MultiBind  [BundleIdent] UnitInstance
  | SingleBind BundleIdent   Bundle

-- applications allow matching by position or by name
data UnitInstance 
  = App        UnitIdent (Either PosArgs NameArgs)
  | RhsUnit    Unit
  | RhsFlatten Bool UnitInstance

type PosArgs  = [Arg]
type NameArgs = ([(BundleIdent,Arg)],Bool)
type Arg      = Either Bundle UnitInstance

data Bundle
  = BundleIdent BundleIdent
  | BundleBits  [BundleElement]
  | BundleFind  BundleType
  | BundleUnion [Bundle]

type BundleElement = (Ident, QualIdent)  -- x = y.x

----------------------------------------------------------------
-- Constraints
----------------------------------------------------------------

data Constraint
  = Eq Property Property
  | Le Property Property

data Property
  = PropApp PropIdent ObjectSet
  | PropConst TypeIdent

----------------------------------------------------------------
-- Rules for deriving new constraints (horn clauses)
----------------------------------------------------------------

-- what is range of quantification?
-- do I need fun-vars and type-vars (type-vars quantify over
-- typeident and fun-vars over qualidents)???
-- forall vars. F(vars) => A(vars)    -- horn clauses
type Clause = ([Ident], Maybe Formula, Atomic)

-- no quantifiers because range of quantification is unclear
-- maybe we want to define "is_import(x)" and the like?
data Formula
  = Atomic Atomic
  | Formula :|: Formula  -- disjunction
  | Formula :&: Formula  -- conjunction
  | Not Formula          -- negation
  | Parens Formula       -- preserve users parentheses

data Atomic 
  = ALe AProp AProp  -- p1 <= p2
  | Import Ident
  | Export Ident
  | Init Ident
  | Fini Ident

data AProp 
  = APropApp PropIdent Ident
  | APropConst TypeIdent  -- not sure that this is useful

----------------------------------------------------------------
-- Initialisation dependencies
----------------------------------------------------------------

type Depend = (ObjectSet, DependRelation, ObjectSet) 

data ObjectSet 
  = ObjIdent  BundleIdent 
  | ObjImports
  | ObjExports
  | ObjInits
  | ObjFinis
  | ObjSet    [QualIdent]
  | ObjUnion  ObjectSet ObjectSet
  | ObjDiff   ObjectSet ObjectSet
  | ObjParens ObjectSet

data DependRelation
  = Calls
  | Precedes

----------------------------------------------------------------
-- Identifiers and qualified identifiers
----------------------------------------------------------------

type QualIdent   = (Maybe BundleIdent, Ident)
type MemberIdent = Ident
type BundleIdent = Ident
type UnitIdent   = Ident
type BundleTIdent= Ident
type PackageIdent= Ident
type FlagsIdent  = Ident
type PropIdent   = Ident
type TypeIdent   = Ident
type RuleIdent   = Ident

type FileName = String

----------------------------------------------------------------
-- Utilities
----------------------------------------------------------------

relativeName :: RelativeFileName -> String
relativeName (SrcDir Nothing  x) = x
relativeName (SrcDir (Just d) x) = d ++ "/" ++ x
relativeName (AnonCode x) = error $ 
  "relativeName" ++ unlines x

----------------------------------------------------------------
-- End 
----------------------------------------------------------------
