-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module Constraints( 
        Class(..),  Property''(..),
        Constraint'(..), ppQConstraints, ppConstraint,
        Property'(..), ppProp,
        mkClasses,
        checkConstraints,
        ) where

import Id
import AbstractUnits
import qualified PPUnits as PP
import PrettyUtils
import Sequence
import IOExts 
import MonadUtils
import Maybe
import List( find, nub, nubBy, partition, sortBy )

----------------------------------------------------------------
-- Some datatypes
----------------------------------------------------------------

type CIdent = QualId
ppCIdent = ppQualid

-- A class includes a list of its superclasses
data Class = MkClass TypeIdent [Class]

instance Eq Class where MkClass c1 _ == MkClass c2 _ = c1 == c2

data Property'' = MkProp PropIdent

instance Eq Property'' where MkProp p1 == MkProp p2 = p1 == p2

isSubClassOf :: Class -> Class -> Bool
cl1@(MkClass c1 supers1) `isSubClassOf` cl2@(MkClass c2 supers2) 
  = c1 == c2 || any (`isSubClassOf` cl2) supers1

ppClass :: Class -> Doc
ppClass (MkClass c supers) = PP.typeIdent c

mkClasses :: [(TypeIdent, [TypeIdent])] -> [Class]
mkClasses decls = classes
 where
  -- note cyclic definition on next line
  classes = [ MkClass c (map findClass supers) | (c,supers) <- decls ]

  findClass c = fromMaybe (error err) (find (\ (MkClass c' _) -> c == c') classes)
   where
    err = show $  text "Unrecognised superclass"
               <+> PP.typeIdent c

ppProperty :: Property'' -> Doc
ppProperty (MkProp p) = PP.propIdent p

data Constraint' 
  = Eq' Property' Property'
  | Le' Property' Property'

data Property' 
  = PropApp' Property'' [CIdent]
  | PropClass Class

ppQConstraints :: [(Qualifier,[Constraint'])] -> Doc
ppQConstraints ds = vlist2 lbrace semi rbrace (map ppQConstraint ds)

ppQConstraint :: (Qualifier,[Constraint']) -> Doc
ppQConstraint (q, cs) = ppQualifier q <+> ppConstraints cs

ppConstraints :: [Constraint'] -> Doc
ppConstraints ds = vlist2 lbrace semi rbrace (map ppConstraint ds)

ppConstraint :: Constraint' -> Doc
ppConstraint (Eq' p1 p2) = ppProp p1 <+> equals <+> ppProp p2
ppConstraint (Le' p1 p2) = ppProp p1 <+> text "<=" <+> ppProp p2

ppProp :: Property' -> Doc
ppProp (PropApp' f xs) = ppProperty f <> braces (hlist ppCIdent xs)
ppProp (PropClass c)   = ppClass c

----------------------------------------------------------------
-- The constraint checker just does a bunch of unifications
----------------------------------------------------------------

checkConstraints :: [(Qualifier,[Constraint'])] -> IO ()
checkConstraints qcs = do
  print $ int (length cs) <+> text "constraints"
  unless (null raw_undefineds) $
    print $ text "Ill defined constraints:" 
          $$ nest 2 (vlist (braces . hlist (ppThing . unNode)) undefineds)
  print $ vlist ppError $ take 5 $ errors
  when (length errors > 5) $ putStrLn "<more errors omitted>"
 where
  cs = flattenConstraints qcs

  g  = mkGraph cs
  types = [ (n, simplify $ rights $ map unNode $ reachable n) | n@(Node x _) <- g ]

  -- It's undefined (underconstrained) if it isn't related to any types  
  -- ToDo: actually, we should only complain about things which are 
  --       contained on one side but not another (whatever that means)
  raw_undefineds = [ n | (n@(Node (Left x) _), []) <- types ]

  -- We get a better error message if we group the undefined symbols
  -- so that you can tell that stomping on one of them will get them all.
  -- ToDo: I'm not sure this is what actually happens
  -- 
  -- This is pretty slow but the first entry is often all you need and
  -- it doesn't get delayed at all.
--  undefineds = equivsBy equiv raw_undefineds
  undefineds = [raw_undefineds]

  errors = sortBy isSimplerError raw_errors
  raw_errors 
        =  {-nub-} type_errors 
        ++ [ (x, [ (c, filter isRealThing $ path (Right c) x) | c <- ys ]) 
           | (x, ys@(a:b:_)) <- types
           , isRealThing (unNode x)
           ]

  type_errors :: [Error]
  type_errors = [ (n1, [(c2, filter isRealThing $ path (Right c2) n1)])
                | n1@(Node (Right c1) _) <- g
                , n2@(Node (Right c2) _) <- reachable n1
                , not (c1 `isSubClassOf` c2)
                ]

  equiv n1 n2 = n1 `elem` reachable n2

  ppError :: Error -> Doc
  ppError (n, cs)    = text "Conflict:" 
                     $$ nest 2 (vcat [ ppConflict n c exp | (c,exp) <- cs ])
  ppConflict (Node (Left  q) _) c' exp = ppQualid q <+> text "<=" <+> ppClass c'
                                       $$ nest 2 (explain exp)
  ppConflict (Node (Right c) _) c' exp
    | c == c'
    = empty     -- eliminate pointless explanations that C <= C
    | otherwise
    =  ppClass  c <+> text "<=" <+> ppClass c'
    $$ nest 2 (explain exp)
  explain exp = text "because" <+> ppPath exp

type Error = (Node Thing, [(Class,[Thing])])

-- Heuristic: Complexity of an error is number of steps in paths.
complexity :: Error -> Int
complexity (_, cs) = sum [ length ts - 1 | (_,ts) <- cs ]

-- Used to list "simple" errors first
isSimplerError :: Error -> Error -> Ordering
isSimplerError e1 e2 = complexity e1 `compare` complexity e2

isRealThing :: Thing -> Bool
isRealThing (Left q) = isRealIdent q
isRealThing _ = True

equivsBy eq [] = []
equivsBy eq (x:xs) = (x:ys) : equivsBy eq zs
 where
  (ys,zs) = partition (eq x) xs

ppPath :: [Thing] -> Doc
ppPath xs = map ppThing xs `hsepdBy` text " <= "

ppThing :: Thing -> Doc
ppThing = either ppQualid ppClass

path :: Eq a => a -> Node a -> [a]
path target n = reverse $ map unNode $ head $ paths target [] [[n]]

unNode (Node n _) = n

-- extends paths towards target
paths :: Eq a => a -> [a] -> [[Node a]] -> [[Node a]] 
paths target seen [] = []
paths target seen (p@(Node n ns:_) : ps)
  | n == target 
  = p : paths target (n:seen) ps
  | n `elem` seen
  = paths target (n:seen) ps
  | otherwise 
  = paths target (n:seen) ([ x:p | x <- ns ] ++ ps)

rights :: [Either a b] -> [b]
rights xs = [ x | Right x <- xs ]

type Thing = Either QualId Class

-- yoiks - quadratic explosion
cross :: [a] -> [b] -> [(a,b)]
cross as bs = [ (a,b) | a <- as, b <- bs ]

-- Improved version of cross to avoid combinatorial explosion
crossLe :: Qualifier -> Int -> [Thing] -> [Thing] -> List (Thing,Thing)
crossLe q i [] bs = mzero
crossLe q i as [] = mzero
crossLe q i as@(a:_) bs@(b:_) = fromList ([ (a,x) | a <- as ] ++ [ (x,b) | b <- bs ])
 where
  x = Left x'
  x' = qualify q (mkIdent ("$$con_" ++ show i))

isRealIdent :: QualId -> Bool
isRealIdent q = take 6 name /= "$$con_"
 where 
  name = show (identOf q)

-- Everything in as and bs is equal.
-- Carefully written to:
-- 1) Avoid quadratic explosion
-- 2) Ensure that every edge we introduce corresponds directly to
--    something that the programmer wrote so that error messages
--    make some kind of sense.
crossEq :: [Thing] -> [Thing] -> List (Thing,Thing)
crossEq [] bs = mzero
crossEq as [] = mzero
crossEq (a:as) (b:bs) = fromList (  [ (a, x) | x <- b:bs ]
                                 ++ [ (x, a) | x <- b:bs ]
                                 ++ [ (x, b) | x <- a:as ]
                                 ++ [ (b, x) | x <- a:as ]
                                 )

flattenConstraints :: [(Qualifier,[Constraint'])] -> [(Thing,Thing)]
flattenConstraints = toList . foldr mplus mzero . map qc
 where
  qc :: (Qualifier,[Constraint']) -> List (Thing,Thing)
  qc (q,cs) = foldr mplus mzero (zipWith (con q) [0..] cs)

  con :: Qualifier -> Int -> Constraint' -> List (Thing,Thing)
  con q i (Eq' p1 p2) = crossEq (prop p1) (prop p2)
  con q i (Le' p1 p2) = crossLe q i (prop p1) (prop p2)

  prop :: Property' -> [Thing]
  prop (PropApp' (MkProp p) qs) = map (Left . (`qualify` p) . asQualifier) qs
  prop (PropClass c)            = [Right c]

data Node a = Node a [Node a] deriving Show

instance Eq a => Eq (Node a) where (Node x _) == (Node y _) = x == y

mkGraph :: Eq a => [(a,a)] -> [Node a]
mkGraph xys = g
 where
  xs = map fst xys
  ys = map snd xys
  g = map mkNode (nub (xs ++ ys))
  mkNode x = Node x (map mkNode' (follow x))
  mkNode' x' = case [ n | n@(Node x _) <- g, x == x' ] of
              (a:_) -> a
              _     -> panic "mkGraph"

  follow x' = nub [ y | (x,y) <- xys, x == x' ]

reachable :: Eq a => Node a -> [Node a]
reachable n = reachable' [] [n] []

reachable' :: Eq a => [Node a] -> [Node a] -> [[Node a]] -> [Node a]
reachable' seen [] [] = seen
reachable' seen [] (ns:nss) = reachable' seen ns nss
reachable' seen (n@(Node x xs) : ns) nss 
  | n `elem` seen 
  = reachable' seen ns nss
  | otherwise
  = reachable' (n:seen) xs (ns:nss)

simplify :: [Class] -> [Class]
simplify cs = foldl min' [] cs
 where
  min' mins x
--    | any (x ==) mins
--    = mins -- optimisation to reduce thrashing
    | any (`isSubClassOf` x) mins
    = mins
    | otherwise
    = x : filter (not . (x `isSubClassOf`)) mins


----------------------------------------------------------------
-- Utils
----------------------------------------------------------------         

panic :: String -> a
panic s = error ("Error in compiler: " ++ s ++ "\nPlease report it as a bug.\n")

singleton :: a -> List a
singleton x = fromList [x]

diff :: Eq a => [a] -> [a] -> [a]
diff xs ys = [ x | x <- xs, not (x `elem` ys) ]

-- change entry in association list
rebind :: Eq a => a -> b -> [(a,b)] -> [(a,b)]
rebind x y [] = panic "Constraints.rebind"
rebind x y (p@(x',_):xys) 
  | x == x'
  = (x,y):xys
  | otherwise
  = p : rebind x y xys

-- strict!
modIORef :: (IORef a) -> (a -> a) -> IO ()
modIORef r f = do
  a <- readIORef r
  writeIORef r $! (f a)

fst3 :: (a,b,c) -> a
snd3 :: (a,b,c) -> b
thd3 :: (a,b,c) -> c
fst3 (a,b,c) = a
snd3 (a,b,c) = b
thd3 (a,b,c) = c

----------------------------------------------------------------
-- End 
----------------------------------------------------------------