-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module CostC
	( defn
        ) where
import Prelude hiding (init)
import AbstractC

-- Estimate cost of:
-- 1) inlining a function     == size of body - size of call
-- 2) not inlining a function == size of body

defn :: Defn -> (Int, Int)
defn (Fun (_,ty) _ c) = (inline_cost, cost)
 where
  cost        = compound c
  inline_cost = cost - (arg_cost ty + 1)
defn _ = (0,0)

-- Number of arguments in a type (if it's a function)
arg_cost :: Type -> Int
arg_cost (Function1 as _ _ _) = length as
arg_cost (Function2 as _ _ _) = length as
arg_cost (TyAttribute _ ty) = arg_cost ty
arg_cost (TyApp ty1 ty2) = arg_cost ty1 + arg_cost ty2
arg_cost (ParensType ty) = arg_cost ty
arg_cost _ = 0

-- ToDo: remove bound vars from free vars of is
-- (Don't expect it to have much effect in practice - function names
--  tend not to look like variable names in typical C code._
compound :: Compound -> Int
compound (_,is,ss) = sum (map init is ++ map statement ss)

init :: Init -> Int
init (d,i) = initializer i

initializer :: Initializer -> Int

initializer None            = 0
initializer (SimpleInit e)  = expr e
initializer (FieldInit _ i) = initializer i
initializer (ArrayInit e i) = expr e + initializer i
initializer (ListInit is)   = sum (map initializer is)

expr :: Expr -> Int
expr (Ident q)         = 1
expr (Constant c)      = 1
expr (String s)        = 2 + length s
expr (Constructor _ i) = initializer i
expr (App e es)        = expr e + sum (map expr es) + length es
expr (Select _ e)      = 1 + expr e
expr (Arrow _ e)       = 1 + expr e
expr (Cast _ e)        = expr e
expr (Cond e1 e2 e3)   = 1 + expr e1 + expr e2 + expr e3
expr (Cond2 e1 e2)     = 1 + expr e1 + expr e2
expr (UnOp "&" e)      = 1
expr (UnOp "(_)" e)    = expr e
expr (UnOp _ e)        = 1 + expr e
expr (BinOp _ e1 e2)   = 1 + expr e1 + expr e2
expr (StatementExpr c) = compound c
expr _                 = 1 -- really?

-- we ignore the cost of the return instruction because we assume we
-- can eliminate it when we inline
statement :: Statement -> Int
statement (Return me)           = mbExpr me
statement (If e s1 s2)          = 1 + expr e + statement s1 + statement s2
statement (Switch e s)          = 1 + expr e + statement s
statement (Label _ s)           = statement s
statement (Case e s)            = statement s
statement (Default s)           = statement s
statement (While e s)           = 2 + expr e + statement s
statement (Do s e)              = 2 + statement s + expr e
statement (For me1 me2 me3 s)   = 2 + mbExpr me1 + mbExpr me2 + mbExpr me3 + statement s
statement (Compound c)          = compound c
statement (Expr e)              = expr e
statement (Asm _ s as1 as2 as3) = 1 + length (filter (==';') s)  -- approx # of instructions in asm
                                 + sum (map asmArg as1 ++ map asmArg as2 ++ map asmArg as3)
statement Empty                 = 0
statement _                     = 1

asmArg :: AsmArg -> Int
asmArg (_, me) = mbExpr me

mbExpr :: Maybe Expr -> Int
mbExpr = maybe 0 expr
