-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

-- A crude attempt at extracting symbols from assembly code
-- using a crude tokeniser and a stop list.
-- It's not clear what the reserved words are in assembly code.
-- If it turned out (say) that instruction names could also be used as
-- variable names then we could remove them from the stoplist and
-- rely on the fact that C programmers are unlikely to use instruction
-- names as (unrelated) symbols.

module DepAsm( depAsm ) where

import Maybe
import List

import ZEncode
import System( system, ExitCode(..) )
import MonadUtils
	( unless
	)

----------------------------------------------------------------
-- Wrapper
----------------------------------------------------------------

depAsm :: [String] -> [String] -> FilePath -> IO [String]
depAsm env flags fn = do
  { my_system $ "rm -f " ++ yyy
  ; rc <- my_system $ unwords $ "env " : env ++ [shell ("gcc" : ["-P","-E","$CPU_CPPFLAGS"] ++ flags ++ [fn,"> " ++ yyy])]
  ; unless (rc == ExitSuccess) (ioError $ userError $ "Can't preprocess " ++ fn)
  ; cs <- readFile yyy
  ; return $ nub (filter (not . isReserved) (tokenise cs))
  }
 where
  yyy = (zEncode fn ++ "yyy")

  shell xs = "sh -c '" ++ unwords xs ++ "'"

my_system xs = do
--  putStrLn xs -- for debugging
  rc <- System.system ("trap '' 26; " ++ xs)
  unless (rc == ExitSuccess) $ putStrLn xs
  return rc
    
----------------------------------------------------------------
--
----------------------------------------------------------------

isReserved x 
  =  x `elem` stop_list 
  || head x `elem` dubious
  || any (x `hasPrefix`) bad_prefix

tokenise = words . map (kill punc) . kill_strings

-- we need to recognise strings as complete tokens so we can drop them
-- on the floor.  There's some minimal support for escaped quotes too
-- but it's not been tested.
kill_strings [] = []
kill_strings ('"':xs) = in_string xs
 where
  in_string [] = []
  in_string ('\\':x:xs) = in_string xs
  in_string ('"':xs) = kill_strings xs
  in_string (x:xs) = in_string xs

kill_strings ('\\':x:xs) = '\\':x:kill_strings xs
kill_strings (x:xs) = x : kill_strings xs



-- we convert any of these to whitespace
punc = ",():;#*|+<-=$"

-- we ignore any tokens starting with these characters
-- note that some of these shadow things in the stop list
dubious = "%@0123456789"

bad_prefix = 
  [ ".L"
  , "Linux_Boot" -- useful when testing to see what slips through
  ]

hasPrefix _ []          = True
hasPrefix (x:xs) (y:ys) = x == y && hasPrefix xs ys
hasPrefix _ _           = False

-- we ignore the following tokens
stop_list =
  [ "%eax"
  , "%ah"
  , "%al"
  , "%ax"
  , "%bh"
  , "%bl"
  , "%bx"
  , "%ch"
  , "%cl"
  , "%cx"
  , "%ebp"
  , "%ebx"
  , "%eci"
  , "%ecx"
  , "%edi"
  , "%edx"
  , "%esi"
  , "%esp"
  , ".align"
  , ".ascii"
  , ".bss"
  , ".byte"
  , ".code16"
  , ".code32"
  , ".comm"
  , ".data"
  , ".file"
  , ".global"
  , ".globl"
  , ".ident"
  , ".local"
  , ".long"
  , ".org"
  , ".p2align"
  , ".rodata"
  , ".section"
  , ".size"
  , ".space"
  , ".stabs"
  , ".stabn"
  , ".string"
  , ".text"
  , ".type"
  , ".value"
  , ".version"
  , ".weak"
  , ".word"
  , ".zero"
  , "@function"
  , "APP"    -- was #APP in gcc generated code
  , "NO_APP" -- was #NO_APP in gcc generated code
  , "addb"
  , "addl"
  , "addw"
  , "andb"
  , "andl"
  , "andw"
  , "btl"
  , "btsl"
  , "call"
  , "cdq"
  , "cld"
  , "cli"
  , "cltd"
  , "clts"
  , "cmpb"
  , "cmpl"
  , "cmpsb"
  , "cmpw"
  , "cpuid"
  , "decb"
  , "decl"
  , "divl"
  , "divw"
  , "fninit"
  , "hlt"
  , "idivl"
  , "imull"
  , "inb"
  , "incb"
  , "incl"
  , "incw"
  , "int"
  , "iret"
  , "ja"
  , "jae"
  , "jb"
  , "jbe"
  , "je"
  , "jg"
  , "jge"
  , "jl"
  , "jle"
  , "jmp"
  , "jnc"
  , "jne"
  , "jns"
  , "jnz"
  , "js"
  , "jz"
  , "leal"
  , "leave"
  , "lgdt"
  , "lidt"
  , "ljmp"
  , "ltr"
  , "mov"
  , "movb"
  , "movl"
  , "movsb"
  , "movsbl"
  , "movsl"
  , "movsw"
  , "movw"
  , "movzbl"
  , "movzbw"
  , "movzwl"
  , "negl"
  , "notl"
  , "notw"
  , "orb"
  , "orl"
  , "orw"
  , "outb"
  , "popa"
  , "popfl"
  , "popl"
  , "pusha"
  , "pushfl"
  , "pushl"
  , "rep"
  , "repnz"
  , "repz"
  , "ret"
  , "retl"
  , "roll"
  , "rorl"
  , "rorw"
  , "sall"
  , "salw"
  , "sarl"
  , "scasb"
  , "shrb"
  , "shrl"
  , "shrw"
  , "smsw"
  , "std"
  , "sti"
  , "stosb"
  , "stosl"
  , "sub"
  , "subb"
  , "subl"
  , "subw"
  , "testb"
  , "testl"
  , "testw"
  , "wbinvd"
  , "xor"
  , "xorb"
  , "xorl"
  , "xorw"
  ]


kill punc x = if x `elem` punc then ' ' else x

subst theta x = fromMaybe x (lookup x theta)
