-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module EscC
	( defn
        ) where
import Prelude hiding (init)
import AbstractC

-- Calculate what functions fun might take the address of (so we
-- can't inline it).  This is like finding functions which 'escape'
-- their definition in lambda calculus.  (Well, it's a little
-- like it  :-))
-- Also calculates which variables are accessed by code - this is used in
-- dead variable elimination.
--
-- The only difference from CostC is the handling of Ident and &
-- & switches us into an alternate mode where we look for function names

-- ToDo: remove parameters from free vars of c
defn :: Defn -> [QualId]
defn (Fun _ _ c)  = compound c
defn (Init (_,i)) = initializer i
defn (Extern d)   = defn d
defn _            = []

-- ToDo: remove bound vars from free vars of is
-- (Don't expect it to have much effect in practice - function names
--  tend not to look like variable names in typical C code._
compound :: Compound -> [QualId]
compound (_,is,ss) = concat (map init is ++ map statement ss)

init :: Init -> [QualId]
init (d,i) = initializer i

initializer :: Initializer -> [QualId]

initializer None            = []
initializer (SimpleInit e)  = expr e
initializer (FieldInit _ i) = initializer i
initializer (ArrayInit e i) = expr e ++ initializer i
initializer (ListInit is)   = concatMap initializer is

expr :: Expr -> [QualId]
expr (Ident q)         = [q]   -- could be an escaping function (silent &)
expr (Constructor _ i) = initializer i
expr (App e es)        = call e ++ concatMap expr es
expr (Select _ e)      = expr e
expr (Arrow _ e)       = expr e
expr (Cast _ e)        = expr e
expr (Cond e1 e2 e3)   = expr e1 ++ expr e2 ++ expr e3
expr (Cond2 e1 e2)     = expr e1 ++ expr e2
-- expr (UnOp "&" e)      = expr e
-- expr (UnOp "*" e)      = expr e -- ??
expr (UnOp _ e)        = expr e
expr (BinOp _ e1 e2)   = expr e1 ++ expr e2
expr (StatementExpr c) = compound c
expr _                 = []

-- Expression in a function call context
call :: Expr -> [QualId]
call (Ident q)         = [] -- doesn't count
call (Constructor _ i) = initializer i
call (App e es)        = call e ++ concatMap expr es
call (Select _ e)      = expr e
call (Arrow _ e)       = expr e
call (Cast _ e)        = call e
call (Cond e1 e2 e3)   = expr e1 ++ call e2 ++ call e3
call (Cond2 e1 e2)     = call e1 ++ call e2
call (UnOp "&" e)      = expr e
call (UnOp "*" e)      = call e
call (UnOp _ e)        = call e
call (BinOp _ e1 e2)   = call e1 ++ call e2
call (StatementExpr c) = compound c
call _                 = []

statement :: Statement -> [QualId]
statement (Return me)           = mbExpr me
statement (If e s1 s2)          = expr e ++ statement s1 ++ statement s2
statement (Switch e s)          = expr e ++ statement s
statement (Label _ s)           = statement s
statement (Case e s)            = expr e ++ statement s
statement (Default s)           = statement s
statement (While e s)           = expr e ++ statement s
statement (Do s e)              = statement s ++ expr e
statement (For me1 me2 me3 s)   = mbExpr me1 ++ mbExpr me2 ++ mbExpr me3 ++ statement s
statement (Compound c)          = compound c
statement (Expr e)              = expr e
statement (Asm _ _ as1 as2 as3) = concat (map asmArg as1 ++ map asmArg as2 ++ map asmArg as3)
statement _                     = []

asmArg :: AsmArg -> [QualId]
asmArg (_, me) = mbExpr me

mbExpr :: Maybe Expr -> [QualId]
mbExpr = maybe [] expr
