-- Copyright (c) 2000, 2001 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module EvalUnits
        ( Unit', evalUnit, ppUnit
        , Binding', ppBinding'
        , Constraint'(..), Property'(..)
        , BundleType', Bundle'', Flag'
        , Depend', ppDepends, ppDepend
        , Initializer', Finalizer', ppInitializer
        ) where

import AbstractUnits
import Monads( RO, runRO, outRO, fixRO, getRO, withROEx, redirectRO )
import qualified PPUnits as PP
import ParseUnits( readDefns )
import Constraints

import Maybe(fromMaybe, maybeToList, mapMaybe)
import MonadUtils
import PrettyUtils
import List( nub, (\\) )
import TopSort

import IOExts(unsafePerformIO) -- debugging only

import IOExts(trace) -- debugging

----------------------------------------------------------------
-- Tests
----------------------------------------------------------------

-- just print the result
t1 :: String -> IO ()
t1 x = do
  xs <- readDefns ["../units/oskit"] ["Clack.unit"]
--  print $ PP.ts xs
  let u' = evalUnit xs (mkIdent x)
  print $ ppUnit u'

-- just print the result
t2 :: String -> IO ()
t2 x = do
  xs <- readDefns ["."] ["test"]
--  print $ PP.ts xs
  let u' = evalUnit xs (mkIdent x)
  print $ ppUnit u'

----------------------------------------------------------------
-- Evaluated Interfaces and Bindings
----------------------------------------------------------------

type Unit' =
  ( [QualId]  -- exports
  , [(Qualifier,[Constraint'])]
  , [Depend']      
  , [Initializer'] 
  , [Finalizer']
  , [Binding']      -- files to be mangled
  )

type BundleType' = 
  ( [Ident]  -- names (including inits and finis)
  )

type Bundle'' = [(Ident, QualId)]
type Bundle' = [(Ident, CIdent)]
type Binding' = 
  ( Qualifier            -- prefix for local names
  , Bundle''              -- requires
  , Bundle''              -- provides
  , Bool                 -- should this be flattened into giant C file?
  , [(RelativeFileName,[Flag'])]
  )
type Initializer' = (QualId, [QualId])
type Finalizer'   = (QualId, [QualId])
type Depend'      = ([QualId], DependRelation, [QualId])
type Flag'        = String

mkCIdent :: QualId -> E CIdent
fromCIdent :: CIdent -> QualId
type CIdent = QualId
-- newtype CIdent = CIdent QualId deriving (Eq , Show) -- show is for debug
-- mkCIdent q = return (CIdent q)
-- fromCIdent (CIdent q) = q
mkCIdent = return
fromCIdent = id

ienvToCIdents :: IEnv -> [CIdent]
ienvToCIdents = concatMap ((map snd) . snd)

ppUnit :: Unit' -> Doc
ppUnit (exps,cs,ds,is,fs,bs)
  =  text "Exports:" <+> hlist ppQualid exps
  $$ text "Constraints:" <+> ppQConstraints cs
  $$ text "Dependencies:" <+> ppDepends ds
  $$ hlist ppInitializer is
  $$ hlist ppInitializer fs
  $$ vmap ppBinding' bs  

ppBinding' :: Binding' -> Doc
ppBinding' (qual, rs, ps, flatten, fs)
  = ppQualifier qual <+> equals <+>
      (  text "requires" <+> ppBundle'' rs  <> semi
      $$ text "provides" <+> ppBundle'' ps  <> semi
      $$ if flatten then text "flatten;" else empty
      $$ text "files"    <+> braces (vlist fileAndFlags fs) <> semi
      )
 where
  fileAndFlags (f,flags) = 
       ppFile f
   <+> text "with flags" <+> braces (hlist text flags) <> semi

  ppFile (AnonCode c) =  text "%{" <+> vmap text c 
                      $$ text "%}"
  ppFile f = text (relativeName f)

ppCIdent = ppQualid . fromCIdent

ppBundle'' :: Bundle'' -> Doc
ppBundle'' i = braces (hlist ppX i)
 where
  ppX (x,y) = PP.ident x <> equals <> ppQualid y

ppBundle' :: Bundle' -> Doc
ppBundle' i = braces (hlist ppX i)
 where
  ppX (x,y) = PP.ident x <> equals <> ppCIdent y

ppBundleType' :: BundleType' -> Doc
ppBundleType' s = braces (hlist PP.ident s)

ppDepends :: [Depend'] -> Doc
ppDepends ds = vlist2 lbrace semi rbrace (map ppDepend ds)

ppDepend :: Depend' -> Doc
ppDepend (x,r,y) =   braces (hlist ppQualid x) 
                 <+> PP.depRel r 
                 <+> braces (hlist ppQualid y)

ppInitializer :: Initializer' -> Doc
ppInitializer (x,ds) = ppQualid x <+> braces (hlist ppQualid ds)

----------------------------------------------------------------
-- The evaluator
--
-- All the real work takes place inside unitInstance and evalUnit
----------------------------------------------------------------

evalUnit      :: TEnv -> Ident -> Unit'

bundleType    :: BundleType    -> E BundleType'
binding       :: Binding       -> E IEnv
unitInstance  :: UnitInstance  -> E [Bundle']
bundle        :: Bundle        -> E Bundle'
bundleElement :: BundleElement -> E (Ident, CIdent)
file          :: FileWithFlags -> E (RelativeFileName, [Flag'])
flags         :: Flags         -> E [Flag']
flag          :: Flag          -> E [Flag']

ident         :: Ident         -> E CIdent
qualIdent     :: QualIdent     -> E CIdent
bundleIdent   :: BundleIdent   -> E Bundle'
unitIdent     :: UnitIdent     -> E Unit
modunit       :: ModUnit       -> E Unit
bundleTIdent  :: BundleTIdent  -> E BundleType'
packageIdent  :: PackageIdent  -> E [Glue]
flagsIdent    :: FlagsIdent    -> E [Flag']
propIdent     :: PropIdent     -> E Property''
typeIdent     :: TypeIdent     -> E Class
ruleIdent     :: RuleIdent     -> E [Clause']

constraints   :: [CIdent] -> [CIdent] -> [CIdent] -> [CIdent] -> [Clause']
              -> [Constraint]  -> E [Constraint']
objectSet     :: [CIdent] -> [CIdent] -> [CIdent] -> [CIdent] 
              -> ObjectSet       -> E [CIdent]
initializer   :: [CIdent] -> [CIdent] -> [CIdent] -> [CIdent] 
              -> (QualIdent, ObjectSet) -> E ()
finalizer     :: [CIdent] -> [CIdent] -> [CIdent] -> [CIdent] 
              -> (QualIdent, ObjectSet) -> E ()
depend        :: [CIdent] -> [CIdent] -> [CIdent] -> [CIdent] 
              -> Int -> Depend -> E ()

-- This is applied to the top unit in the build.
--
-- The main trick in this is that the imports and exports of the top unit
-- have hardwired names: if it provides 'main', it is definitely not cool to
-- rename it to export1.main.  Imports just work but exports try to do the
-- Bad Thing (ie exported names are uniquified).
-- 
-- The way we do this is that we merrily rename everything we find then,
-- at the end we replace the exports with their "hardwired" names.
--
-- In case it's not obvious, this is a hack...
evalUnit te unitName = (map rn exps, map rnQConstraint cs, map rnDep ds, map rnInit is', map rnInit fs', map rnBinding bs)
 where
  bs :: [Binding']
  ds :: [Depend']
  is' :: [Initializer']
  fs' :: [Finalizer']
  is :: [Bundle']
  (bs,cs,ds,is',fs',is) = runE te $ extendQualifier unitName eval

  exps :: [QualId]
  exps = concatMap (map (fromCIdent . snd)) is

  eval :: E [Bundle']
  eval = do
    u@(rs,ps,rules,cs,vis,vfs,ds,flatten,body) <- unitIdent unitName
    ie <- mapM mkEnv rs
    extendEnv ie $ do
      u' <- case body of
            Left  fs -> atomicUnit (rs,ps,rules,cs,vis,vfs,ds,flatten,fs) 
            Right bs -> compound   (rs,ps,rules,cs,vis,vfs,ds,flatten,bs) 
      return (map fst u')

  mkEnv	:: Import -> E (BundleIdent, Bundle')
  mkEnv (i, bt) = do
    bt' <- bundleType bt
    b   <- bundleTypeToBundle bt'
    return (i, b)
  
  bundleTypeToBundle :: BundleType' -> E Bundle'
  bundleTypeToBundle ns = mapM aux ns
   where
    aux n = do
      n' <- mkCIdent (qualify emptyQual n)
      return (n,n')

  rn        :: QualId -> QualId
  rnInit    :: Initializer' -> Initializer'
  rnDep     :: Depend' -> Depend'
  rnBundle'   :: Bundle'' -> Bundle''
  rnBinding :: Binding' -> Binding'
  rnQConstraint :: (Qualifier,[Constraint']) -> (Qualifier,[Constraint'])
  rnConstraint :: Constraint' -> Constraint'

  rn q = case nub [ x | e <- is, (x,q') <- e, q == fromCIdent q' ] of
         []  -> q
         [r] -> qualify emptyQual r
         rs  -> error $ "Ambiguous export (possibly due to exporting the same thing under 2 different names?)" 
                     ++ show (ppQualid q <+> equals <+> braces (hlist ppIdent rs))

  rnGen  (f,xs) = (f, map rn xs)
  rnInit (x,ds) = (rn x, map rn ds)
  rnDep  (x,r,y)  = (map rn x, r, map rn y)
  rnBundle' ifc = [ (x, rn y) | (x,y) <- ifc ]
  rnBinding (q,rs',ps',flatten,fs) = (q, rnBundle' rs', rnBundle' ps', flatten, fs)
  rnQConstraint (q,cs) = (q, map rnConstraint cs)
  rnConstraint (Eq' p1 p2) = Eq' (rnProp p1) (rnProp p2)
  rnConstraint (Le' p1 p2) = Le' (rnProp p1) (rnProp p2)
  rnProp (PropApp' f xs) = PropApp' f (map rn xs)
  rnProp x = x

-- Atomic units and compound units are so different from each other
-- that we make the split between them here.  (The types might have
-- led you to believe we'd do it only when dealing with the body/files.)
unitInstance (App u as) = do
  (rs,ps,rules,cs,vis,vfs,ds,flatten,body) <- unitIdent u
  ie <- imports rs as
  is <- freshEnv $ extendEnv ie $ do
    case body of
     Left  fs -> atomicUnit (rs,ps,rules,cs,vis,vfs,ds,flatten,fs) 
     Right bs -> compound   (rs,ps,rules,cs,vis,vfs,ds,flatten,bs) 
  return (map fst is)
unitInstance (RhsFlatten f r) = modFlatten (const f) (unitInstance r)
unitInstance (RhsUnit u@([],ps,rules,cs,vis,vfs,ds,flatten,body)) = freshEnv $ do
  is <- case body of
         Left  fs -> atomicUnit ([],ps,rules,cs,vis,vfs,ds,flatten,fs) 
         Right bs -> compound   ([],ps,rules,cs,vis,vfs,ds,flatten,bs) 
  return (map fst is)  
unitInstance (RhsUnit (rs,ps,rules,cs,vis,vfs,ds,flatten,body)) 
  = report $ text "Can't handle nested unit with non-empty imports"

isLocal (Nothing,_) = True
isLocal (Just _, _) = False

flattenObjectSet (ObjIdent i)     = []
flattenObjectSet ObjImports       = []
flattenObjectSet ObjExports       = []
flattenObjectSet ObjInits         = []
flattenObjectSet ObjFinis         = []
flattenObjectSet (ObjSet   qs)    = qs
flattenObjectSet (ObjUnion d1 d2) = flattenObjectSet d1 ++ flattenObjectSet d2
flattenObjectSet (ObjDiff  d1 d2) = flattenObjectSet d1 ++ flattenObjectSet d2
flattenObjectSet (ObjParens d)    = flattenObjectSet d

atomicUnit :: 
  ( [Import]
  , [Export]
  , [RuleIdent]
  , [Constraint]
  , [(QualIdent,ObjectSet)] -- concrete initializers
  , [(QualIdent,ObjectSet)] -- concrete finalizers
  , [Depend]
  , Maybe Bool
  , AtomicUnit
  )
  -> E [(Bundle',BundleType')]
atomicUnit (rs, ps, rules, cs, vis, vfs, ds, flatten, (fs, rns)) = do
  flatten' <- getFlatten -- does parent want us flattened?
  q  <- getQualifier

  -- we add the global symbols into the list of imported symbols
  rs0 <- getIEnv 
  globs <- getGlobals
  let rs' = [ (nullIdent,g) | g <- globs ] ++ rs0

  p_env <- mapM (\(i,s) -> do{ s' <- bundleType s; ifc <- bundleTypeToBundle (extendQual i q) s'; return (i,ifc) }) ps
  let
    -- We have to build a up a list of all local identifiers.
    -- The rule is that all unqualified names in the initialisers,
    -- finalisers and exports are local.
    locals = nub $ map snd $ filter isLocal $ (map fst vis ++ map fst vfs)

    mkLocal :: Ident -> E (Ident, CIdent)
    mkLocal x = do
      x' <- mkCIdent (qualify q x)
      return (x,x')

  locals' <- mapM mkLocal locals

  extendEnv [(nullIdent,locals')] $ extendEnv p_env $ do
    ps' <- mapM (export' q) ps

    -- dependencies may depend on things we provide and on virtual initialisers
    let 
      ifcs = [ (ifc,sig) | (i,ifc,sig) <- ps' ]
      ie   = [ (i, ifc)  | (i,ifc,sig) <- ps' ]

      imports = ienvToCIdents rs'
      exports = ienvToCIdents ie

    inits <- mapM (qualIdent . fst) vis
    finis <- mapM (qualIdent . fst) vfs

    extendEnv ie $ do 
      rules' <- mapM ruleIdent rules
      cs' <- constraints imports exports inits finis (concat rules') cs
      addConstraints q cs'
      zipWithM_ (depend imports exports inits finis) [0..] ds
      mapM_ (initializer imports exports inits finis) vis
      mapM_ (finalizer   imports exports inits finis) vfs

    fs' <- mapM file fs

    let rs'' = [ rename i x q | (i, ifc)    <- rs', (x,q) <- ifc ]
        ps'' = [ rename i x q | (i, ifc, _) <- ps', (x,q) <- ifc ]


    -- ToDo: should watch for clashes between rs' and ifcs
    addBinding (q, rs'', ps'', combFlatten flatten flatten', fs')
    return ifcs
 where

  rename :: BundleIdent -> Ident -> QualId -> (Ident,QualId)
  rename i x q = case mapMaybe (match i x) rns of 
    []  -> (x,q)
    [y] -> (y,q)
    _  -> error "Ambiguous renaming"  -- ToDo: better error message

  match :: BundleIdent -> Ident -> Rename -> Maybe Ident
  match i x (RenameTo (p,q) r)   | i == p && x == q = Just r
  match i x (RenamePrefix p pre) | i == p           = Just (mkIdent (fromIdent pre ++ fromIdent x))
  match i x (RenameSuffix p suf) | i == p           = Just (mkIdent (fromIdent x ++ fromIdent suf))
  match _ _ _ = Nothing

  export' :: Qualifier -> Export -> E (BundleIdent, Bundle', BundleType')
  export' q (i, s) = do
    s'  <- bundleType s
    ifc <- bundleTypeToBundle (extendQual i q) s'
    return (i, ifc, s')

  bundleTypeToBundle :: Qualifier -> BundleType' -> E Bundle'
  bundleTypeToBundle q ns = mapM aux ns
   where
    aux n = do
      n' <- mkCIdent (qualify q n)
      return (n,n')

initializer  imports exports inits finis (x, ds) = do
  x' <- qualIdent x
  ds' <- objectSet imports exports inits finis ds
  addInitializer x' ds'
finalizer  imports exports inits finis (x, ds) = do
  x' <- qualIdent x
  ds' <- objectSet imports exports inits finis ds
  addFinalizer x' ds'

compound :: 
  ( [Import]
  , [Export]
  , [RuleIdent]
  , [Constraint]
  , [(QualIdent,ObjectSet)] -- concrete initializers
  , [(QualIdent,ObjectSet)] -- concrete finalizers
  , [Depend]
  , Maybe Bool
  , Bindings
  )
  -> E [(Bundle',BundleType')]
compound (rs, ps, rules, cs, vis, vfs, ds, flatten, bs) = modFlatten (combFlatten flatten) $ do
  q   <- getQualifier
  ie'' <- getIEnv

  do
    (ie',ps') <- bindings bs ps
    extendEnv ie' $ do
      -- dependencies may depend on names we give to exports
      -- but should compound units be adding dependencies?
      let 
        ifcs = [ (ifc,sig) | (i,sig,ifc) <- ps' ]
        ie   = [ (i,  ifc) | (i,sig,ifc) <- ps' ]
  
        imports = ienvToCIdents ie''
        exports = ienvToCIdents ie
        inits   = map (lkup . fst) vis
        finis   = map (lkup . fst) vfs
  
        lkup :: QualIdent -> CIdent
        lkup (Nothing, y) = panic "compound.lkup"  -- ToDo
        lkup (Just x,  y) = panic "compound.lkup"  -- ToDo
  
      extendEnv ie' $ do
        rules' <- mapM ruleIdent rules
        cs' <- constraints imports exports inits finis (concat rules') cs
        addConstraints q cs'
        zipWithM_ (depend imports exports inits finis) [0..] ds
        mapM_ (initializer imports exports inits finis) vis
        mapM_ (finalizer   imports exports inits finis) vfs
  
      return ifcs
  
export2 :: Export -> E Default
{-
export2 (i,s) = do
  s'  <- bundleType s
  ifc <- bundleIdent i
  return (i, ifc, s')
-}
export2 (i, s) = do
  ie <- getIEnv
  s' <- bundleType s
  ifc0 <- case [ y | (x,y) <- ie, i==x ] of
           [y] -> return y
           []  -> defaultArg i s'
           _   -> report $  text "Ambiguous argument" <+> PP.ident i
  let ifc = restrict s' ifc0
  ifc `hasType` s'
  return (i, s', ifc)
  
-- glue units are added to the defaults
bindings :: Bindings -> [Export] -> E (IEnv, [Default])
bindings (gls,us,ds,bs) ps = do
  (bs,_,ps) <- fixE $ 
    \ ~(ie,gs,_) -> extendEnv ie $
    getIEnv >>= \ ie' -> 
    withDefaults (gs ++ map (f ie') ds) $ 
    withGlobals (map (g ie') gls) $ 
    mapM glue us >>= \ gss -> do
      bss <- mapM binding bs
      ps' <- mapM export2 ps
      return (concat bss, concat gss, ps')
  return (bs,ps)
 where
  f ie d = (d, bundleToBundleType ifc, ifc) where ifc = find' ie d
  g ie x = find' ie x

-- Glue units generate instances of the form [x,y,z] <- glue <- {...}
glue :: Glue -> E [Default]
glue (GlueIdent x) = glue' x
glue (GluePackage x) = do
  us  <- packageIdent x
  dss <- mapM glue us
  return (concat dss)

glue' :: UnitIdent -> E [Default]
glue' u = extendQualifier qual $ do
  (rs,ps,rules,cs,vis,vfs,ds,flatten,body) <- unitIdent u
  ie <- mapM glueArg rs
  is <- freshEnv $ extendEnv ie $ do
    case body of
     Left  fs -> atomicUnit (rs,ps,rules,cs,vis,vfs,ds,flatten,fs) 
     Right bs -> compound   (rs,ps,rules,cs,vis,vfs,ds,flatten,bs) 
  return (zipWith nameIt is [0..])
 where
  u' = fromIdent u
  qual = mkIdent $ "glue_" ++ u'
  nameIt (i, s) n = (mkIdent (u' ++ "_" ++ show n), s, i)

glueArg :: Import -> E (BundleIdent, Bundle')
glueArg (i, s) = do
  ie <- getDefaults
  s' <- bundleType s
  ifc <- case nub [ restrict s' i 
                     | (_,sig,i) <- ie, s' `isSubBundleTypeOf` sig ] of
            [i] -> return i
            []  -> report $  text "Can't find glue argument" <+> PP.ident i <+> colon <+> ppBundleType' s'
                          $$ text "in"                  <+> ppDefaults ie
                          $$ text "Can't find glue argument" <+> PP.ident i <+> colon <+> ppBundleType' s'
            is  -> report $  text "Ambiguous glue bundleType" <+> ppBundleType' s'
                          $$ text "in"                  <+> braces (hlist ppBundle' is)
  ifc `hasType` s'
  return (i, ifc)

imports :: [Import] -> (Either PosArgs NameArgs) -> E IEnv
imports rs as = either (posArgs rs) (nameArgs rs) as

posArgs :: [Import] -> PosArgs -> E IEnv
posArgs rs as = do
  unless (length rs == length as) $
    report $ text "Mismatched arguments"
           $$ text "Expected:" <+> vmap PP.import' rs
           $$ text "Actual:  " <+> PP.posArgs as
  zipWithM posArg rs (zip as [0..])

posArg :: Import -> (Arg,Int) -> E (BundleIdent, Bundle')
posArg (i, s) (a,n) = do
  s'   <- bundleType s
  ifc0 <- arg a n
  let ifc = restrict s' ifc0
  ifc `hasType` s'
  return (i, ifc)

restrict :: BundleType' -> Bundle' -> Bundle'
restrict sig@(ns) ifc = [ (x,y) | (x, y) <- ifc, x `elem` ns ]

nameArgs :: [Import] -> NameArgs -> E IEnv
nameArgs rs (as, ellipsis) = do
  as' <- mapM lookupArg (zip as [0..])
  nameArgs' rs as' ellipsis

lookupArg :: ((BundleIdent,Arg),Int) -> E (BundleIdent, Bundle')
lookupArg ((i,r),n) = do
  r' <- arg r n
  return (i, r')

arg :: Arg -> Int -> E Bundle'
arg (Left  b) n = bundle b
arg (Right i) n = extendQualifier id $ do
  es <- unitInstance i
  case es of
   [e] -> return e
   _   -> report $ text "Anonymous unit instances must export exactly one bundle"
 where
  id = mkIdent ("arg__" ++ show n)

nameArgs' :: [Import] -> [(BundleIdent, Bundle')] -> Bool -> E IEnv
nameArgs' rs as ellipsis = do
  mapM (nameArg as) rs
 where
  nameArg :: IEnv -> Import -> E (BundleIdent, Bundle')
  nameArg ie (i, s) = do
    s' <- bundleType s
    ifc0 <- case [ y | (x,y) <- ie, i==x ] of
             [y] -> return y
             []  -> if ellipsis
                    then defaultArg i s'
                    else report $ text "Missing argument" <+> PP.ident i
             _   -> report $  text "Ambiguous argument" <+> PP.ident i
    let ifc = restrict s' ifc0
    ifc `hasType` s'
    return (i, ifc)

defaultArg :: BundleIdent -> BundleType' -> E Bundle'
defaultArg i s = do
  ie   <- getDefaults
  lookupBundleByBundleType ie i s

lookupBundleByBundleType :: [Default] -> BundleIdent -> BundleType' -> E Bundle'
lookupBundleByBundleType ie i s = do
  case nub [ i | (_,sig,i) <- ie, s `isSubBundleTypeOf` sig ] of
   [i] -> return i
   []  -> report $  text "Can't find argument" <+> PP.ident i <+> colon <+> ppBundleType' s
                 $$ text "in"                  <+> ppDefaults ie
   is  -> report $  text "Ambiguous bundleType" <+> ppBundleType' s
                 $$ text "in"                  <+> braces (hlist ppBundle' is)

hasType :: Bundle' -> BundleType' -> E ()
hasType i s 
  | s `isSubBundleTypeOf` bundleToBundleType i
  = return ()
  | otherwise
  = report $  text "Argument mismatch"
           $$ text "Interface" <+> ppBundle' i
           $$ text "does not have bundleType" <+> ppBundleType' s

bundleType (xs,_) = foldM btElement [] xs

btElement :: BundleType' -> BTElement -> E BundleType'
btElement ns (Ident   i) = return (i:ns)
btElement ns (Include f) = return ns  -- ignored for now
btElement ns (Extends s) = do
  ns' <- bundleTIdent s
  return (ns++ns')

binding (SingleBind x b) = do
  b' <- extendQualifier x $ bundle b
  return [(x,b')]
binding (MultiBind l r) = do
  r' <- extendQualifier (head l) $ unitInstance r
  bindBundles l r'
 where
  bindBundles xs is 
    | length xs /= length is
    -- ToDo: improve error message using qualifier??
    = report $ text "Wrong number of result bindings" 
             $$ text "Actual:  " <+> int (length is) <+> text "results"
             $$ text "Expected:" <+> braces (vlist PP.bundleIdent xs)
    | otherwise
    = return (zip xs is)

bundle (BundleBits xs)  = mapM bundleElement xs
bundle (BundleIdent i)  = bundleIdent i
bundle (BundleUnion bs) = do
   bs' <- mapM bundle' (zip bs [0..])
   return (foldr union [] bs')
 where
  bundle' (b,i) = extendQualifier (mkIdent ("u" ++ show i)) (bundle b)
bundle (BundleFind s) = do
  s' <- bundleType s
  ie <- getDefaults
  ifc <- case nub [ restrict s' i 
                     | (_,sig,i) <- ie, s' `isSubBundleTypeOf` sig ] of
            [i] -> return i
            []  -> report $  text "Can't resolve `find'" <+> ppBundleType' s'
                          $$ text "in"                   <+> ppDefaults ie
                          $$ text "Can't resolve `find'" <+> ppBundleType' s'
            is  -> report $  text "Ambiguous glue bundleType" <+> ppBundleType' s'
                          $$ text "in"                  <+> braces (hlist ppBundle' is)
  ifc `hasType` s'
  return ifc

bundleElement (lhs,x) = do
  x' <- qualIdent x
  return (lhs,x')

-- instead of generating x1 : y1, x2 : y1, x1 : y2, x2 : y2,
-- we generate x1 : $$$.l, x2 : $$$.l,  $$$.l : $$$.r, $$$.r : y1, $$$.r : y2
-- This reduces the number of dependencies in large examples.
depend imports exports inits finis i (xs,r,ys) = do
  q <- getQualifier
  xs' <- objectSet imports exports inits finis xs
  ys' <- objectSet imports exports inits finis ys
  addDepend (map fromCIdent xs', r, map fromCIdent ys')

objectSet imports exports inits finis = objectSet'
 where
  objectSet' (ObjIdent i)     = do{ xs <- bundleIdent i; return (map snd xs) }
  objectSet' ObjImports       = return imports
  objectSet' ObjExports       = return exports
  objectSet' ObjInits         = return inits
  objectSet' ObjFinis         = return finis
  objectSet' (ObjSet   qs)    = mapM qualIdent qs
  objectSet' (ObjUnion d1 d2) = liftM2 union (objectSet' d1) (objectSet' d2)
  objectSet' (ObjDiff  d1 d2) = liftM2 diff  (objectSet' d1) (objectSet' d2)
  objectSet' (ObjParens d)    = objectSet' d

isSubBundleTypeOf :: BundleType' -> BundleType' -> Bool
isSubBundleTypeOf s1 s2 = all (`elem` s2) s1

bundleToBundleType :: Bundle' -> BundleType'
bundleToBundleType = map fst

modunit (x,rs,as) = do
  u <- unitIdent x
  case u of
   (is,es,[],cs,[],[],deps,f,Right (gls,gs,defs,bs)) -> do
     let
       (ris,res,rcs,rdeps,rgls,rgs,rdefs,rbs) = rs
       (ais,aes,acs,adeps,agls,ags,adefs,abs) = as
     ogs  <- concatMapM glue'' gs
     orgs <- concatMapM glue'' rgs
     let
       is'   = ais ++ rm ris is
       es'   = aes ++ rm res es
       cs'   = acs   ++ cs
       deps' = adeps ++ deps
       gls'  = agls ++ (gls \\ rgls)
       gs'   = ags   ++ map GlueIdent (ogs \\ orgs)
       defs' = adefs ++ (defs \\ rdefs)
       bs'   = abs   ++ rmbs rbs bs
     return (is',es',[],cs',[],[],deps',f,Right (gls',gs',defs',bs'))
   _ -> report $ text "Can't modify unit" <+> PP.unitIdent x
 where  
  glue'' :: Glue -> E [UnitIdent]
  glue'' (GlueIdent   x) = return [x]
  glue'' (GluePackage x) = do
    us <- packageIdent x
    concatMapM glue'' us

  rm :: Eq a => [a] -> [(a,b)] -> [(a,b)]
  rm as abs = [ (a,b) | (a,b) <- abs, not $ a `elem` as ]

  rmbs :: [BundleIdent] -> [Binding] -> [Binding]
  rmbs bis bs = filter (not . match) bs
   where
    match (MultiBind (x:_) _) = x `elem` bis
    match (SingleBind x _)    = x `elem` bis

-- local identifiers can be named without a qualifier
-- since the ienv doesn't identify local identifiers, we 
-- look for ones with empty qualifiers - bit of a hack

ident y = do
  ie   <- getIEnv
  case nub [ q | (x,i) <- ie, x==nullIdent, (y',q) <- i, y==y' ] of
   [q] -> return q
   []  -> report $  text "Can't find ident" <+> PP.ident y 
                 $$ text "in"               <+> ppIEnv ie
   qs  -> report $ text "Ambiguous ident" <+> PP.ident y
                 $$ text "in"             <+> braces (hlist ppCIdent qs)
  
nullIdent :: Ident
nullIdent = mkIdent ""

qualIdent (Nothing, y) = ident y
qualIdent qid@(Just x,  y) = do
  ie   <- getIEnv
  case nub [ q | (x',i) <- ie, x==x', (y',q) <- i, y==y' ] of
   [q] -> return q
   []  -> report $  text "Can't find qualident" <+> PP.qualIdent qid
                 $$ text "in"                   <+> ppIEnv ie
   qs  -> report $ text "Ambiguous qualident"   <+> PP.qualIdent qid
                 $$ text "in"                   <+> braces (hlist ppCIdent qs)

bundleIdent x = do
  ie <- getIEnv
  find ppBundle' ie x

unitIdent x = do
  te <- getTEnv
  find2' unitFromT te x return $ find2' modunitFromT te x modunit $ err te
 where
  err env = report $  text "Can't find" <+> PP.ident x 
                   $$ text "among" <+> braces (hlist PP.ident (nub (map fst env)))

bundleTIdent x = do
  te <- getTEnv
  bt <- find2 btFromT te x
  bundleType bt

packageIdent x = do
  te <- getTEnv
  find2 packageFromT te x

flagsIdent x = do
  te <- getTEnv
  f <- find2 flagsFromT te x
  flags f

propIdent x = do
  te <- getTEnv
  find2 propFromT te x
  return (MkProp x)

ruleIdent x = do
  te <- getTEnv
  rs <- find2 ruleFromT te x
  mapM clause rs

----------------------------------------------------------------
-- Before we process a clause, we have to resolve all the references in it
----------------------------------------------------------------

-- ToDo: parameterise wrt kind of property and type?
type Clause' = ([Ident], Maybe Formula', Atomic')
data Formula'
  = Atomic' Atomic'
  | Formula' :||: Formula'  
  | Formula' :&&: Formula'  
  | Not' Formula'          
  | Parens' Formula'       
data Atomic' 
  = ALe' AProp' AProp'
  | Import' Ident
  | Export' Ident
  | Init' Ident
  | Fini' Ident
data AProp' 
  = APropApp' Property'' Ident
  | APropConst' Class

clause  :: Clause  -> E Clause'
formula :: Formula -> E Formula'
atomic  :: Atomic  -> E Atomic'
aprop   :: AProp   -> E AProp'

-- ToDo: vs could mask propicate names
clause (vs,f,a) = do
  f' <- maybeM formula f
  a' <- atomic a
  return (vs,f',a')

formula (Atomic a)  = liftM1 Atomic' (atomic a)
formula (f1 :|: f2) = liftM2 (:||:)  (formula f1) (formula f2)
formula (f1 :&: f2) = liftM2 (:&&:)  (formula f1) (formula f2)
formula (Not f)     = liftM1 Not'    (formula f)
formula (Parens f)  = liftM1 Parens' (formula f)

atomic (ALe p1 p2) = liftM2 ALe' (aprop p1) (aprop p2)
atomic (Import x)  = liftM0 (Import' x)
atomic (Export x)  = liftM0 (Export' x)
atomic (Init x)    = liftM0 (Init' x)
atomic (Fini x)    = liftM0 (Fini' x)

aprop (APropApp p x) = do
  p' <- propIdent p
  return (APropApp' p' x)
aprop (APropConst c) = do
  c' <- typeIdent c
  return (APropConst' c')

----------------------------------------------------------------
-- Constraints and inference rules
----------------------------------------------------------------

-- Evaluate all the constraints then apply the rules 
-- to "bulk up" what we have
-- We are able to restrict ourselves to a single pass over
-- the rules by sorting the rules according to dependencies
-- of their RHS on their LHS.  (This requires us to enforce
-- a rule that these dependencies are non-cyclic.)
constraints imports exports inits finis rules cs = do
  let
    defs  = zip [0..] [ nub $ atomicDep a  | (_,_,a) <- rules ]
    needs = zip [0..] [ (nub $ atomicDep a, nub $ formulaDep f) | (_,Just f,a) <- rules ]

    deps  = [ (x,[ y | (y,(_,bs)) <- needs, any (`elem` bs) as]) | (x,(as,_)) <- needs ]

    order = reverse $ topSortR deps (map fst defs)
    rules' = map (rules !!) order

  cs' <- mapM (constraint imports exports inits finis) cs

  return (foldl applyRule cs' rules')

 where
  -- we quantify over imports+exports+inits+finis
  qs = inits ++ finis ++ exports ++ imports
  
  applyRule :: [Constraint'] -> Clause' -> [Constraint']
  applyRule cs (vs,f,(ALe' p1 p2)) = new ++ cs
   where
    new = [ Le' (prop env p1) (prop env p2) 
          | env <- envs vs qs
          , fromMaybe True (fmap (formula env) f)
          ]
  
    formula env (Atomic' a)  = atomic env a
    formula env (f1 :||: f2) = formula env f1 || formula env f2
    formula env (f1 :&&: f2) = formula env f1 && formula env f2
    formula env (Not' f)     = not (formula env f)
    formula env (Parens' f)  = formula env f
  
    atomic env (ALe' p1 p2) = Le' (prop env p1) (prop env p2) `elemC` cs
    atomic env (Import' x)  = var env x `elem` imports
    atomic env (Export' x)  = var env x `elem` exports
    atomic env (Init' x)    = var env x `elem` inits
    atomic env (Fini' x)    = var env x `elem` finis
    
    prop env (APropApp' p x) = PropApp' p [var env x]
    prop env (APropConst' c) = PropClass c
    
    var env x = fromMaybe err (lookup x env)
     where
      err = error $ "Unknown identifier in clause" ++ show (ppIdent x) -- todo

impliedBy :: Constraint' -> Constraint' -> Bool
impliedBy (Le' p1 p2) (Le' p1' p2') = p1 `subProp` p1' && p2 `subProp` p2'
impliedBy (Le' p1 p2) (Eq' p1' p2') = p1 `subProp` p1' && p2 `subProp` p2'
impliedBy (Eq' p1 p2) (Eq' p1' p2') = p1 `subProp` p1' && p2 `subProp` p2'
impliedBy (Eq' p1 p2) (Le' p1' p2') = False -- but it could be true if we look elsewhere!

-- wrong name
subProp :: Property' -> Property' -> Bool
subProp (PropApp' p xs) (PropApp' p' xs') = p == p' && xs `subsetEq` xs'

-- ToDo: if p x <= p y and p y <= p z then px <= pz 
elemC :: Constraint' -> [Constraint'] -> Bool
elemC = elemBy impliedBy

subsetEq :: Eq a => [a] -> [a] -> Bool
subsetEq as bs = all (`elem` bs) as
  
elemBy p x xs = any (p x) xs

envs :: [a] -> [b] -> [[(a,b)]]
envs [] bs = [[]]
envs (a:as) bs = [ (a,b):e | e <- envs as bs, b <- bs ]

formulaDep :: Formula' -> [Property'']
atomicDep  :: Atomic'  -> [Property'']
propDep    :: AProp'   -> [Property'']

formulaDep (Atomic' a)  = atomicDep a
formulaDep (f1 :&&: f2) = formulaDep f1 ++ formulaDep f2
formulaDep (f1 :||: f2) = formulaDep f1 ++ formulaDep f2
formulaDep (Not' f)     = formulaDep f
formulaDep (Parens' f)  = formulaDep f

atomicDep (ALe' p1 p2) = propDep p1 ++ propDep p2
atomicDep _ = []

propDep (APropApp' p _) = [p]
propDep (APropConst' c) = []

constraint imports exports inits finis c = 
  case c of 
  Eq p1 p2 -> liftM2 Eq' (property p1) (property p2)
  Le p1 p2 -> liftM2 Le' (property p1) (property p2)
 where
  property (PropApp p xs) = do
    p'  <- propIdent p
    xs' <- objectSet imports exports inits finis xs
    return (PropApp' p' xs')
  property (PropConst c) = do
    c' <- typeIdent c
    return (PropClass c')
    
typeIdent x = do
  classes <- getClasses
  case [ cls | cls@(MkClass c _) <- classes, x==c ] of
       [y] -> return y
       []  -> report $  text "Can't find type" <+> PP.typeIdent x 
       _   -> report $  text "Ambiguous type identifier" <+> PP.typeIdent x

find :: Eq a => (a -> Doc) -> [(Ident,a)] -> Ident -> E a
find pp env x = case nub [ y | (x',y) <- env, x==x' ] of
  [y] -> return y
  []  -> report $  text "Can't find" <+> PP.ident x 
                $$ text "among" <+> braces (hlist PP.ident (nub (map fst env)))
  ys  -> report $  text "Ambiguous identifier" <+> PP.ident x <> colon
                $$ braces (vlist pp ys)
  
-- we can provide a value to use if can't find it
find3 :: Eq a => [(Ident,a)] -> a -> Ident -> E a
find3 env def x = case nub [ y | (x',y) <- env, x==x' ] of
  [y] -> return y
  []  -> return def
  _   -> report $  text "Ambiguous identifier in renaming" <+> PP.ident x
  
find2 :: (a -> Maybe b) -> [(Ident,a)] -> Ident -> E b
find2 p env x = case [ b | (x',y) <- env, x==x', Just b <- [p y] ] of
  [y] -> return y
  []  -> report $  text "Can't find" <+> PP.ident x 
                $$ text "among" <+> braces (hlist PP.ident (nub (map fst env)))
  _   -> report $  text "Ambiguous identifier 2" <+> PP.ident x
  
find2' :: (a -> Maybe b) -> [(Ident,a)] -> Ident -> (b -> E c) -> E c -> E c
find2' p env x succ fail = case [ b | (x',y) <- env, x==x', Just b <- [p y] ] of
  [y] -> succ y
  []  -> fail
  _   -> report $  text "Ambiguous identifier 2" <+> PP.ident x
  
-- shouldn't use this (because it gives worse error messages)
find' :: [(Ident,a)] -> Ident -> a
find' env x = case [ y | (x',y) <- env, x==x' ] of
  [y] -> y
  []  -> error $ "Can't find " ++ show x ++ " in " ++ show (map fst env)
  _   -> error $ "Ambiguous "  ++ show x ++ " in " ++ show (map fst env)
  
find'' :: (a -> Maybe b) -> [(Ident,a)] -> Ident -> b
find'' p env x = case [ b | (x',y) <- env, x==x', Just b <- [p y] ] of
  [y] -> y
  []  -> error $ "Can't find " ++ show x ++ " in " ++ show (map fst env)
  _   -> error $ "Ambiguous "  ++ show x ++ " in " ++ show (map fst env)
  
---------------------------------------------------------------
-- Flag set flattening
---------------------------------------------------------------

file (fileName,fs) = do
  fs' <- flags fs
  return (fileName, fs')

flags = concatMapM flag

flag (RawFlag str)      = return [str]
flag (UseFlags str)     = flagsIdent str

----------------------------------------------------------------
-- Utils
----------------------------------------------------------------         

btFromT :: Defn -> Maybe BundleType
btFromT (BundleT l) = Just l
btFromT _           = Nothing

unitFromT :: Defn -> Maybe Unit
unitFromT (Unit l)    = Just l
unitFromT _           = Nothing

modunitFromT :: Defn -> Maybe ModUnit
modunitFromT (ModUnit l) = Just l
modunitFromT _           = Nothing

flagsFromT :: Defn -> Maybe Flags
flagsFromT (Flags l) = Just l
flagsFromT _         = Nothing

packageFromT :: Defn -> Maybe [Glue]
packageFromT (Package l) = Just l
packageFromT _           = Nothing

propFromT :: Defn -> Maybe ()
propFromT (Prop  ) = Just ()
propFromT _        = Nothing

ruleFromT :: Defn -> Maybe [Clause]
ruleFromT (Rule cs) = Just cs
ruleFromT _        = Nothing

----------------------------------------------------------------
-- The eval monad
--
-- This uses the Reader+Output monad.
--
-- The Reader part is used to propagate environment info around
-- The Output part is used to accumulate results: bindings, dependencies, etc
----------------------------------------------------------------

type TEnv = [(Ident,Defn)]
type IEnv = [(BundleIdent, Bundle')]
type Default = (BundleIdent, BundleType', Bundle')

data Output
  = B Binding' 
  | C Qualifier [Constraint']
  | D Depend'
  | I CIdent [CIdent]
  | F CIdent [CIdent]

-- ToDo: add a local namespace for unqualified names and kill the
-- nullIdent hack used in atomic and ident
newtype E a = E (RO (TEnv, [Class], [Default], IEnv, Qualifier, Bool, [Bundle']) Output a)
unE (E x) = x

instance Functor E where
  fmap f xs = E (fmap f (unE xs))

instance Monad E where
  return x = E (return x)
  m >>= k = E (unE m >>= (unE . k))


runE            :: TEnv -> E a -> ([Binding'], [(Qualifier,[Constraint'])], [Depend'], [Initializer'], [Finalizer'], a)
fixE        	:: (a -> E a) -> E a
report          :: Doc -> E a
addBinding  	:: Binding'  -> E ()
addConstraints  :: Qualifier -> [Constraint'] -> E ()
addDepend   	:: Depend' -> E ()
addInitializer  :: CIdent -> [CIdent] -> E ()
addFinalizer    :: CIdent -> [CIdent] -> E ()
extendQualifier :: Ident    -> E a -> E a
extendEnv       :: IEnv -> E a -> E a
withDefaults    :: [Default] -> E a -> E a
freshEnv        :: E a -> E a
getQualifier    :: E Qualifier
getTEnv	        :: E TEnv
getIEnv	        :: E IEnv
getDefaults     :: E [Default]
grabIsFs        :: E a -> E (a, [(CIdent,[CIdent])], [(CIdent,[CIdent])])
withFlatten     :: Bool -> E a -> E a
getFlatten      :: E Bool
getClasses      :: E [Class]
withGlobals     :: [Bundle'] -> E a -> E a
getGlobals      :: E [Bundle']

runE te (E e) = (bs, cs, ds, is, fs, a)
 where
  classes = mkClasses [ (cls, supers) | (cls, Type supers) <- te ]
  (a,o) = runRO e (te, classes, [], [], emptyQual, False, [])
  bs = [ x | B x <- o ]
  cs = [ (q,cs) | C q cs <- o ]
  ds = [ x | D x <- o ]
  is = [ (fromCIdent x, map fromCIdent ds) | I x ds <- o ]
  fs = [ (fromCIdent x, map fromCIdent ds) | F x ds <- o ]

outE = E . outRO
getE = E getRO
withEEx f (E m) = E (withROEx f m)

fixE f = E (fixRO (unE . f))

-- ToDo: try to recover enough from error messages that we can keep
-- going and find more.
report msg = do
  qual <- getQualifier              
  error $ show (msg $$ text "while processing" <+> ppQualifier qual)

addBinding b  = outE [B b]
addConstraints q cs = outE [C q cs]
addDepend d = outE [D d]

addInitializer x ds = outE [I x ds]
addFinalizer   x ds = outE [F x ds]
addInitializers is = outE [ I x ys | (x,ys) <- is ]
addFinalizers   fs = outE [ F x ys | (x,ys) <- fs ]

extendQualifier i = withEEx (\ (te, classes, d, ie, q, flatten, gs) -> (te, classes, d,    ie,    extendQual i q, flatten, gs))
--extendEnv     e = withEEx (\ (te, classes, d, ie, q, flatten, gs) -> (te, classes,       e:ie,  q,  flatten, gs))
extendEnv       e = withEEx (\ (te, classes, d, ie, q, flatten, gs) -> (te, classes, d,    e++ie, q,  flatten, gs))
withDefaults    e = withEEx (\ (te, classes, d, ie, q, flatten, gs) -> (te, classes, e,    ie,    q,  flatten, gs))
freshEnv          = withEEx (\ (te, classes, d, ie, q, flatten, gs) -> (te, classes, d,    [],    q,  flatten, gs))
withFlatten     f = withEEx (\ (te, classes, d, ie, q, flatten, gs) -> (te, classes, d,    ie,    q,  f      , gs))
withGlobals     g = withEEx (\ (te, classes, d, ie, q, flatten, gs) -> (te, classes, d,    ie,    q,  flatten, g++gs))

getQualifier = getE >>= \ (te, classes, d, ie, q, flatten, gs) -> return q
getTEnv	     = getE >>= \ (te, classes, d, ie, q, flatten, gs) -> return te
getIEnv	     = getE >>= \ (te, classes, d, ie, q, flatten, gs) -> return ie
getDefaults  = getE >>= \ (te, classes, d, ie, q, flatten, gs) -> return d
getFlatten   = getE >>= \ (te, classes, d, ie, q, flatten, gs) -> return flatten
getClasses   = getE >>= \ (te, classes, d, ie, q, flatten, gs) -> return classes
getGlobals   = getE >>= \ (te, classes, d, ie, q, flatten, gs) -> return gs

grabIsFs (E m) = E $ do
  (a,os) <- redirectRO m
  let 
    is = [ (x,ys) | I x ys <- os ]
    fs = [ (x,ys) | F x ys <- os ]
    ds = [ d | d@(D _) <- os ]
    bs = [ b | b@(B _) <- os ]
  outRO ds
  outRO bs
  return (a,is,fs)

ppIEnv :: IEnv -> Doc
ppIEnv = vlist (\(x,i) -> PP.ident x <+> colon <+> ppBundleType' (bundleToBundleType i))

ppDefaults :: [Default] -> Doc
ppDefaults = vlist (\(x,sig,i) -> PP.ident x <+> colon <+> ppBundleType' sig)

modFlatten      :: (Bool -> Bool) -> E a -> E a
modFlatten f m = do
  flatten <- getFlatten 
  withFlatten (f flatten) m

combFlatten :: Maybe Bool -> (Bool -> Bool)
combFlatten Nothing = id
combFlatten (Just x) = const x

----------------------------------------------------------------
-- Utils
----------------------------------------------------------------         

-- ToDo: maybe complain about overlap
union :: Eq a => [a] -> [a] -> [a]
union as bs = as ++ bs   

diff  :: Eq a => [a] -> [a] -> [a]
diff as bs = [ a | a <- as, not (a `elem` bs) ]

panic :: String -> a
panic s = error ("Error in compiler: " ++ s ++ "\nPlease report it as a bug.\n")

fst3 :: (a,b,c) -> a
snd3 :: (a,b,c) -> b
thd3 :: (a,b,c) -> c
fst3 (a,b,c) = a
snd3 (a,b,c) = b
thd3 (a,b,c) = c

cross :: [a] -> [b] -> [(a,b)]
cross xs ys = [ (x,y) | x <- xs, y <- ys ]

----------------------------------------------------------------
-- End
----------------------------------------------------------------
