-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module Flatten
	( flattenBindings
	) where

import EvalUnits
import AbstractUnits
import qualified PPUnits 

import ParseC
import qualified PP1 as PPC
import qualified AbstractC as C
import RN
import qualified DepC
import qualified EscC
import qualified CostC

import DepAsm

import PrettyUtils
import Maybe(catMaybes, mapMaybe, listToMaybe, fromMaybe, fromJust)
import List(partition, (\\), unzip3, nub, sortBy, minimumBy)
import Monad( mapAndUnzipM )
import Monad -- everything else
import MonadUtils( concatMapM )
import ListUtils( mapAndUnzip3M )
import FiniteMap
import TopSort

-- for debugging
import Id(ppQualid)
import IOExts( trace )

----------------------------------------------------------------
-- There are two parts to the "flattener":
-- 1) Renaming C code: ordering definitions, eliminating duplicates, etc.
--    so that gcc will accept the generated file without too much complaint.
--    After this stage, we are free to output the function definitions in
--    any order we see fit.
-- 2) Choosing the order of function definitions and what to mark for inlining
--    This uses a cost estimate and dependency information to estimate the
--    total cost of a set of inliining decisions and uses various heuristics
--    to try to choose the most profitable inlinings.  
--    It'd be interesting to include profiling information to further
--    improve the decision.  If doing that, we might want to make
--    inlining decisions on a site-by-site basis (see later).
-- 3) Not Done (ToDo): do something sensible with .code16 when it occurs
--    in a function decl or as a top level decl.
----------------------------------------------------------------

----------------------------------------------------------------
-- Renaming C code 
--
-- The important things here are:
-- 1) We must be careful not to reorder declarations too much.
--    It's ok to:
--    o Move type-related declarations to before their use
--      in function/variable prototypes and declarations.
--      (Provided the relative order of type declarations is unchanged.)
--    o Move all function definitions to after variable definitions
--      provided there is a prototype for the function.
--      It's ok to reorder the function definitions.
--      It's not ok to reorder variable declarations because their
--      initialisers might refer to other variables.
--      Variables must come before functions in case the function uses
--      sizeof (quite common for arrays).
-- 2) Function/variable prototypes _must_ match the actual definitions
--    so we generate prototypes from the definitions and use those
--    in preference to any hand-written prototypes.  Prototypes for
--    externally defined variables/functions are kept.
-- 3) Types involved in assignments or function arguments have to match
--    exactly or gcc halts and reports an error.  Renaming with typedefs is ok.
--    Three possible fixes:
--    1) Don't blindly uniquify types 
--       Hard to do for recursive types and bloody tedious besides.
--    2) Cast rhs of assignments and function arguments to appropriate type.
--       Assignments are easy:   "e1 = e2" --> "e1 = (typeof x)(e2)".
--       Function calls are hard because we can't use typeof.
--       We could reimplement the C typechecker - but that sounds no fun!
--    3) Modify gcc's typechecker.
--    We choose (3).
-- 4) Uniquify user_inlines to delete duplicate extern inlines
-- 5) ToDo: Worry more about common variables 
--    (maybe I already worried about that?)
----------------------------------------------------------------

flattenBindings :: [(String,String)] -> [QualId] -> [Binding'] -> IO [Binding']
flattenBindings env exps bs = do
  (code, es, bs') <- mapAndUnzip3M (readBinding env) bs

  putStrLn "Read C code"

  -- We move prototypes up to the front and eliminate duplicates
  -- to kill some fatal gcc warnings.
  let 
    (rss, pss, css) = unzip3 (catMaybes code)

    -- All the symbols which might possibly be exported from C code
    -- plus the functions which gcc might choose to call
    extern = nub (concat es ++ exps)

    css' = map (mkStatic extern) $ concat css
    
    types = justTypes css'
    
    vars' = justVars css'

  (vars, dead, user_inlines, inlines, non_inlines, budget') <- doInlining3 extern css' vars'  undefined budget

--  print $ text "Dead:        " <+> vmap ppQualid [ f | C.Fun (f,_) _ _ <- dead ]
--  print $ text "Inlines:     " <+> vmap ppQualid [ f | C.Fun (f,_) _ _ <- inlines ]
--  print $ text "Non Inlines: " <+> vmap ppQualid [ f | C.Fun (f,_) _ _ <- non_inlines ]

  print $ text "Dead         " <+> int (length dead)
  print $ text "Inlining     " <+> int (length inlines)
  print $ text "Not inlining " <+> int (length non_inlines)

  case budget' `compare` budget of
   LT -> putStrLn ("Undershot budget by " ++ show (budget - budget'))
   GT -> putStrLn ("Overshot budget by "  ++ show (budget' - budget))
   EQ -> putStrLn ("Hit the budget!")

  let
    cs'  =   types 
          ++ uniqDefns [] decls 
          ++ vars 
          ++ uniqDefns [] user_inlines
          ++ funs'
    
    -- machine generated prototypes and var decls override human generated
    -- function decls. and var decls.
    -- given a choice between a human written common var and a human written
    -- extern var, we keep the common-var.
    decls =  mkProtos funs'
          ++ mkVars vars
          ++ justCommons css'
          ++ justExterns css' 
    
    funs' =  map mkFunInline inlines
          ++ non_inlines

  let file    = AnonCode $ map showIt $ map PPC.ppDefn cs'

      -- The import/export lists are only used for renamings
      -- But, we've already renamed everything so all we have to do is:
      -- 1) Rename a few undefined symbols introduced by gcc: memcmp, 
      --    memcpy, etc
      --    Note that we have to use the same renaming for all the code
      --    Unfortunate, but there you go...
      -- 2) Prevent the object file renamer from hiding all the 
      --    symbols which are supposed to be visible.
      rs' = nub [ (x,r) | rs <- rss, (x,r) <- rs, x `elem` gcc_imps ]
      ps' = [ (mkIdent (show (qualident'C x)), x) | x <- exps ] 
      binding = (emptyQual, rs', ps', False, [(file,[])])
  return (binding : bs')
 where
  budget = read (fromMaybe "0" (lookup "KNIT_BUDGET" env))

gcc_imps :: [Ident]
gcc_imps = map mkIdent 
  [ 
  "memcpy",
  "memcmp",
  "memset",
--  "strcpy",
--  "strlen",
--  "alloca",

  -- The following are all the symbols from SigQuad
  -- It's not clear whether they _all_ need special treatment but
  -- at least some of them do.
  "__adddi3",
  "__anddi3",
  "__ashldi3",
  "__ashrdi3",
  "__cmpdi2",
  "__divdi3",
  "__fixdfdi",
  "__fixsfdi",
  "__fixunsdfdi",
  "__fixunssfdi",
  "__floatdidf",
  "__floatdisf",
  "__floatunsdidf",
  "__iordi3",
  "__lshldi3",
  "__lshrdi3",
  "__moddi3",
  "__muldi3",
  "__negdi2",
  "__one_cmpldi2",
  "__qdivrem",
  "__subdi3",
  "__ucmpdi2",
  "__udivdi3",
  "__umoddi3",
  "__xordi3"
  ]

-- PageMode is more readable but probably slower
showIt doc = fullRender PageMode 100 1.5 string_txt "" doc
--showIt doc = fullRender OneLineMode 100 1.5 string_txt "" doc
 where
  string_txt (Chr c)   s  = c:s
  string_txt (Str s1)  s2 = s1 ++ s2
  string_txt (PStr s1) s2 = s1 ++ s2

readBinding :: [(String,String)] -> Binding' 
            -> IO (Maybe (Bundle'',Bundle'',[C.Defn]),[QualId],Binding')
readBinding env b@(qual,reqs,provs,False,fs) = do
  return (Nothing, map snd reqs, b)
readBinding env (qual,reqs,provs,flatten,fs) = do
  let (cs,  not_cs) = partition (isType 'c') fs
  let (asms,other)  = partition (isType 'S') not_cs
  css <- mapM processCFile (zip cs [0..])
  asm_refs <- concatMapM processAsmFile asms
  let cs' = concat css
      builtin = catMaybes $ map (`lookup` renamings) gcc_imps
      externs | null not_cs = builtin
              | null other  = builtin ++ asm_refs
              | otherwise   = builtin ++ map snd reqs ++ asm_refs ++ justGlobals cs'

--  print $ text "Externs: " <+> ppQualifier qual <+> vmap ppQualid externs
--  print $ text "Builtins: " <+> ppQualifier qual <+> vmap ppQualid builtin
  return (Just (reqs,provs,cs'), externs, (qual,reqs,provs,flatten,asms++other))
 where
  -- the renamings aren't very good - but they'll do for now
  renamings = reqs ++ provs

  processCFile :: ((RelativeFileName,[Flag']), Int) -> IO [C.Defn]
  processCFile ((fname,flags), i) = do

    -- Anonymous files get written to anon.c so we can preprocess them
    case fname of { AnonCode f -> writeFile "anon.c" (unlines f); _ -> return () }

    cs <- parseDefnsEx cparser env' flags fname' fname'
--    print $ vmap PPC.ppDefn $ cs
    let is = interns cs
        rename_env = mkEnv renamings qual fq is
    let cs' = map (rnDefn rename_env) cs
--    print $ vmap PPC.ppDefn $ cs'
    return cs'
   where
    bindir  = maybe "" (++"/") $ lookup "KNIT_TOOLS" env
    cparser = bindir ++ "knit_c_parser"
    fname' = fullPath fname
--    fq  = mkQual $ map mkIdent [zEncode fname']
    fq  = mkQual [ mkIdent ("file" ++ show i) ]

  fullPath (SrcDir (Just dir) f) = dir ++ "/" ++ f
  fullPath (SrcDir Nothing    f) = f
  fullPath (AnonCode f)   = "anon.c"

  env' = map (\ (x,y) -> (x ++ "=" ++ y)) env

  processAsmFile :: (RelativeFileName,[Flag']) -> IO [QualId]
  processAsmFile (fname,flags) = do
    cs <- depAsm env' flags (fullPath fname)
    let xs = map (rn . mkIdent) cs
--    print $ PPUnits.fileName fname <> colon <+> vmap ppQualid xs
    return xs
   where    
    rn :: Ident -> QualId
    rn x = fromMaybe (qualify qual x) (lookup x renamings)
      

isType :: Char -> (RelativeFileName,[Flag']) -> Bool
isType ty (SrcDir d f, _) = last f == ty
isType ty (AnonCode f, _) = True

----------------------------------------------------------------
-- Choosing what to inline:
--
-- 1) We estimate the effect of inlining on code size including gains
--    from killing the last use of a function (if not externally visible)
--    and losses from duplicated code.
-- 
--    Inlining cost can be -ve because we subtract the size of a call to
--    the function from the size of the (non-inlined) function.  (This means
--    that every function has 2 costs associated with it: inlining cost
--    and cost of having a non-inlined copy.)
--
--    (Note that inlining cost will be wrong for vararg functions.  In fact,
--    it's possible that gcc disallows inlining of vararg functions?)
--
-- 2) We keep track of calls we can see to a function and calls we can't
--    see either because they're external (the function is exported) or
--    hidden inside a .o or .S file whose imports are unknown or via
--    a function pointer (i.e., we can see the address of the function
--    being taken).
--
--    asms which contain symbols (in the asm string) (and there
--    are a few) are a problem - but they're already a problem since we
--    can't rename them.  (Example of "bad" asm code: "call __put_user_<N>" in
--    oskit/linux/src/include/asm-i386/uaccess.h.)
-- 
-- 3) Local code is either deleted (if we're certain it is dead) or
--    marked static inline (if we're certain it doesn't escape).
--
--    Using -gc-sections kills the dead code too but not if we use 
--    dynamic linking (which we want to do on non-oskit systems).
--    Besides, killing dead code gives more accurate usage counts.
-- 
-- 4) gcc does all inlining in a single pass so we have to choose one
--    function at a time and recalculate the sizes of each function.
--    (We can choose multiple functions at once for leaf nodes.)
--
--    One nice thing about this is that we don't have to worry about
--    mutually recursive inline functions.
--
--    If we want to make inlining decisions at a finer granularity, 
--    we could make multiple copies of each function and rename 
--    individual uses to use the right copy.
--
-- 5) We keep inlining until we reach a user-specified budget for
--    how much code-bloat they can stand.  Savings from killing dead
--    code and inlinings with -ve cost are added to this budget.
--
--    At the end, we report how close to the budget we came.
--    We should also print the estimated size of each function so
--    that we can refine the cost estimate.
--
--
-- ToDo: 
-- o Could do a better job of generating list of symbols that must be defined
-- o Write a wee script which generates dependencies from .S files.
--   Run through cpp, then list tokens which aren't opcodes, register names
--   or punctuation.   Should come pretty close.
--   Applying it to asm statements in C files would also be a good idea.
-- o Refine cost estimate.
--   Consider using least-squares fit (or whatever) to automatically
--   generate costs of each kind of node to make it easier to port to
--   other architectures.
-- o no point in ever considering a function with no local calls
--   but we have to remember its existence because of knock-on effects
-- o ignore functions which escape too
--   unless they have a -ve inlining cost
-- o should definitely ignore things with 0 calls at this stage
-- o handle recursion in inlining algorithms - don't loop!
-- o accurately estimate savings from earlier inlining decisions
-- o consider abandoning the fancy heuristics at the start 
--   (e.g., inlining leaves first) and just using cost estimate.
-- o The "inline" annotations let us make inlining decisions on a 
--   per-function basis.  If we wanted to decide on a per-call-site 
--   basis, we could generate duplicates of the function to be inlined,
--   mark one of them inline, the other no-inline and then rename calls
--   to the function accordingly.
----------------------------------------------------------------

doInlining :: [QualId] -> [C.Defn] -> [C.Defn] -> [C.Defn] -> Int -> ([C.Defn], [C.Defn], [C.Defn], [C.Defn], [C.Defn], Int)
doInlining = doInlining2

----------------------------------------------------------------
-- The trivial inliner: does nothing!
----------------------------------------------------------------

doInlining0 :: [QualId] -> [C.Defn] -> [C.Defn] -> [C.Defn] -> Int -> ([C.Defn], [C.Defn], [C.Defn], [C.Defn], [C.Defn], Int)
doInlining0 externs defns vars funs budget = (vars, [], user_inlines, inlines, non_inlines, budget')
 where
  budget' = budget
  (user_inlines,live_funs) = partition isInline funs
  (inlines,non_inlines) = ([],live_funs)

----------------------------------------------------------------
-- "Inliner" that kills dead code (and nothing else)
----------------------------------------------------------------

doInlining1 :: [QualId] -> [C.Defn] -> [C.Defn] -> [C.Defn] -> Int -> ([C.Defn], [C.Defn], [C.Defn], [C.Defn], [C.Defn], Int)
doInlining1 externs defns vars funs budget = (vars, dead_funs, user_inlines, [], live_funs, budget')
 where
  budget' = budget + death_benefit

  fun_names = [ f | C.Fun (f,_) _ _ <- funs ]

  callgraph :: Map [QualId]
  callgraph = listToMap [ (f, DepC.defn d) | d@(C.Fun (f,_) _ _) <- funs ]

  -- set of all locally defined functions which escape
  -- i.e., have their address taken
  -- ToDo:  ignore accesses from dead code and variables
  escapes :: [QualId]
  escapes = nub 
          $ filter (`elem` fun_names) 
          $ concatMap EscC.defn 
          $ defns
  
  live :: [QualId]
  live = genLive callgraph (escapes ++ externs) [] []

  isLive (C.Fun (f,_) _ _) = f `elem` live

  (live_funs', dead_funs) = partition isLive funs
  (user_inlines,live_funs) = partition isInline live_funs'
  death_benefit = - sum (map (snd . CostC.defn) dead_funs)

----------------------------------------------------------------
-- "Inliner" that kills dead code then spends its savings inlining
-- all the cheap functions.
-- Note that cost is inaccurate because it doesn't take knock-on effect
-- of inlinings into account.
----------------------------------------------------------------

doInlining2 :: [QualId] -> [C.Defn] -> [C.Defn] -> [C.Defn] -> Int -> ([C.Defn], [C.Defn], [C.Defn], [C.Defn], [C.Defn], Int)
doInlining2 externs defns vars funs budget = (vars, dead_funs, user_inlines, cheap_funs, dear_funs, budget')
 where
  budget' = budget + cost

  fun_names = [ f | C.Fun (f,_) _ _ <- funs ]

  callgraph :: Map [QualId]
  callgraph = listToMap [ (f, DepC.defn d) | d@(C.Fun (f,_) _ _) <- funs ]

  -- set of all locally defined functions which escape
  -- i.e., have their address taken
  -- ToDo:  ignore accesses from dead code and variables
  escapes :: [QualId]
  escapes = nub 
          $ filter (`elem` fun_names) 
          $ concatMap EscC.defn 
          $ defns
  
  live :: [QualId]
  live = genLive callgraph (escapes ++ externs) [] []

  isLive (C.Fun (f,_) _ _) = f `elem` live

  (live_funs', dead_funs) = partition isLive funs
  (user_inlines,live_funs) = partition isInline live_funs'
  death_benefit = - sum (map (snd . CostC.defn) dead_funs)

  -- (inline_cost, keep_cost), uses, escapes?
  info :: Map ((Int,Int),[QualId],Bool)
  info = listToMap [ ( f
                    , ( CostC.defn d
                      , uses f
                      , f `elem` escapes || f `elem` externs
                      )
                    )
                  | d@(C.Fun (f,_) _ _) <- funs, f `elem` live ]
  
  uses :: QualId -> [QualId]
  uses f = [ caller | (caller,gs) <- mapToList callgraph, g <- gs, f == g, caller `elem` live ]
  
  costs :: [(QualId, Int)]
  costs = sortBy (comp2 compare snd)
          [ (f, inline_cost * length uses + if escs then keep_cost else 0)
          | (f,((inline_cost,keep_cost),uses,escs)) <- mapToList info
          ]

  (cheap,dear) = break (\(x,y) -> y > 0) (accum (death_benefit-budget) costs)
  cost = case reverse cheap of { [] -> death_benefit; ((_,c):_) -> c }

  isCheap (C.Fun (f,_) _ _) = f `elem` (map fst cheap)

  (cheap_funs, dear_funs) = partition isCheap live_funs

----------------------------------------------------------------
-- "Inliner" that kills dead code then spends its savings inlining
-- all the cheap functions.
-- Note that cost is inaccurate because it doesn't take knock-on effect
-- of inlinings into account.
----------------------------------------------------------------

-- ToDo: 4th argument no longer used
doInlining3 :: [QualId] -> [C.Defn] -> [C.Defn] -> [C.Defn] -> Int -> IO ([C.Defn], [C.Defn], [C.Defn], [C.Defn], [C.Defn], Int)
doInlining3 externs defns vars funs budget = do
--  print escs
  putStrLn $ "Death Benefit: " ++ show (-death_benefit)
  return (live_vars, dead_funs, user_inlines, cheap_funs, dear_funs, budget')
 where
  budget' = budget + cost

  callgraph :: Map [QualId]
  callgraph = listToMap' (++) [ (f, DepC.defn d) | d <- defns, f <- name d ]
--  callgraph = listToMap [ (f, DepC.fun d) | d@(C.Fun (f,_) _ _) <- funs ]

  -- variables and functions which take the address of functions or
  -- which access variables (which might take the address of ...)
  -- Note that we use all definitions here so that user inlined functions
  -- will be scanned too.
  escgraph :: Map [QualId]
  escgraph = listToMap' (++) [ (f, EscC.defn d) | d <- defns, f <- name d ]

  escs = vcat [ (ppQualid f <+> vmap ppQualid gs) | (f,gs) <- mapToList escgraph ]

  name (C.Fun (i,_) _ _)  = [i]
  name (C.Init ((i,_),_)) = [i]
  name (C.Extern d      ) = name d
  name _                  = []

  live :: [QualId]
  live = genLive (plusMap_C (++) escgraph callgraph) externs [] []

  isLiveFun (C.Fun (f,_) _ _)  = f `elem` live
  isLiveFun _                  = False

  isLiveVar (C.Init ((f,_),_)) = f `elem` live
  isLiveVar (C.Extern d      ) = isLiveVar d

  live_vars = filter isLiveVar vars
  (live_funs', dead_funs) = partition isLiveFun (justFuns defns)

  -- We leave functions already marked "inline" alone because
  -- 1) The programmer might be right
  -- 2) If we prevent an extern inline function from being inlined,
  --    the system may not link.  (Of course, this should never happen
  --    but the world ain't perfect and we don't want to have to fix 
  --    the source code.)
  -- Elsewhere, we delete duplicates in user_inlines because
  -- duplicate extern inline defns make gcc unhappy.
  (user_inlines,live_funs) = partition isInline live_funs'

  fun_names = [ f | C.Fun (f,_) _ _ <- live_funs ]

  death_benefit = - sum (map (snd . CostC.defn) dead_funs)

  -- set of all locally defined functions which escape
  -- i.e., have their address taken
  escapes :: [QualId]
  escapes = nub 
          $ filter (`elem` live) -- should check that var that refers to it is live
          $ filter (`elem` fun_names) 
          $ concatMap EscC.defn 
          $ defns
  
  -- ((inline_cost, keep_cost), uses, escapes?)
  info :: Map ((Int,Int),[QualId],Bool)
  info = listToMap [ ( f
                    , ( CostC.defn d
                      , callers
                      , f `elem` escapes || f `elem` externs
                      )
                    )
                  | d@(C.Fun (f,_) _ _) <- live_funs'
                  , let callers = uses f
                  -- we can't inline a function if we can't see the call
                  , not (null callers) 
                  ]
  
  -- if a function calls another multiple times, it is listed multiple times
  uses :: QualId -> [QualId]
  uses f = [ caller | (caller,gs) <- mapToList callgraph, g <- gs, f == g, caller `elem` live ]
  
  costs = inline_lots info
  (cheap,dear) = break (\(x,y) -> y > 0) (accum (death_benefit-budget) costs)
  cost = case reverse cheap of { [] -> death_benefit; ((_,c):_) -> c }

  isCheap (C.Fun (f,_) _ _) = f `elem` (map fst cheap)

  (cheap_funs, dear_funs) = partition isCheap live_funs

-- simulate effect of inlining one function
inline_one_function :: Map ((Int,Int),[QualId],Bool) -> QualId -> Map ((Int,Int),[QualId],Bool) 
inline_one_function m f = updateListMap comb changes (delFromMap f m)
 where
  Just ((inline_cost,keep_cost),uses,escs) = lookupMap f m 

  comb ((a1,b1),c1,d1) ((a2,b2),c2,d2) = ((a1+a2,b1+b2),c1\\c2,d1 && d2)

  changes = [(g,((inline_cost,inline_cost), [], True)) | g <- uses, g /= f ]

-- ToDo: replace with some kind of updateable priority tree/queue
cheapest :: Map ((Int,Int),[QualId],Bool) -> Maybe (QualId, Int)
cheapest m = if null xs then Nothing else Just $ minimumBy (minBy (comp2 compare snd)) xs
 where
  xs = [ (f, inline_cost * length uses + if escs then keep_cost else 0)
       | (f,((inline_cost,keep_cost),uses,escs)) <- mapToList m
       , not (null uses) -- don't mark inline unless callers still exist
       ]

inline_lots :: Map ((Int,Int),[QualId],Bool) -> [(QualId, Int)]
inline_lots m = case cheapest m of 
  Nothing    -> []
  Just (x,c) -> (x,c) : inline_lots (inline_one_function m x)

----------------------------------------------------------------
-- Yet another transitive closure algorithm
----------------------------------------------------------------

-- finds the live code in a bunch of definitions
-- args 2 and 3 are the usual list and list of list worklist
-- arg4 is the usual accumulator
--   genLive ds ws wss ls =~ genLive ds (ws ++ concat wss) [] ls
--   genLive ds ws []  ls =~ genLive ds (ws ++ ls) [] []
--   genLive ds ws []  [] =~ nodes reachable from ws via ds
-- where =~ = set equality

genLive :: Map [QualId] -> [QualId] -> [[QualId]] -> [QualId] -> [QualId]
genLive ds [] []       live = live
genLive ds [] (ws:wss) live = genLive ds ws wss live
genLive ds (w:ws) wss live
  | w `elem` live 
  = genLive ds ws wss live
  | otherwise
  = genLive ds ws (ws':wss) (w:live)
 where
  ws' = fromMaybe [] (lookupMap w ds)


----------------------------------------------------------------
-- Various functions for manipulating C code
----------------------------------------------------------------

mkStatic :: [QualId] -> C.Defn -> C.Defn
mkStatic es (C.Fun   (i,ty) p q)| not (i `elem` es) = C.Fun   (i, mks ty) p q
mkStatic es (C.Init ((i,ty),p)) | not (i `elem` es) = C.Init ((i, mks ty),p)
--mkStatic es (C.Extern (C.Init ((i,ty),p))) | not (i `elem` es) = C.Init ((i, mks ty),p)
mkStatic es (C.Extern (C.Fun   (i,ty) p q))| not (i `elem` es) = C.Fun (i, mks ty) p q
mkStatic es d = d

mks :: C.Type -> C.Type
mks ty | isStaticType ty = ty
mks (C.TyApp (C.SimpleTy x) ty) | x == externQualid = C.TyApp staticTy ty
mks ty | isExternType ty = ty  -- paranoia
       | otherwise       = C.TyApp staticTy ty

mkFunInline :: C.Defn -> C.Defn
mkFunInline (C.Fun (i,ty) p q) = C.Fun (i, mki ty) p q

mki :: C.Type -> C.Type
mki ty = C.TyApp inlineTy ty

inlineTy :: C.Type
inlineTy = C.SimpleTy inlineQualid1

staticTy :: C.Type
staticTy = C.SimpleTy staticQualid

-- symbols with internal linkage (typedefs, enums, unions, structs and statics)
interns :: [C.Defn] -> [QualId]
interns ds = foldr intern [] ds

intern :: C.Defn -> [QualId] -> [QualId]
intern (C.Enum    tag es)  x = [ i | (i,_) <- es ] ++ x
intern (C.Union   tag fs)  x = x
intern (C.Struct  tag fs)  x = x
intern (C.Typedef tag ty)  x = tag : x   
intern (C.Fun (i,rty) _ _) x | isStaticType rty = i:x 
intern (C.Init ((i,ty),_)) x | isStaticType  ty = i:x 
intern _ x = x

isInline :: C.Defn -> Bool
isInline (C.Extern (C.Fun (_, (C.Function1 _ _ ty attrs)) _ _)) = isInlineType ty
isInline (C.Extern (C.Fun (_, (C.Function2 _ _ ty attrs)) _ _)) = isInlineType ty
isInline (C.Fun (_, (C.Function1 _ _ ty attrs)) _ _) = isInlineType ty
isInline (C.Fun (_, (C.Function2 _ _ ty attrs)) _ _) = isInlineType ty
isInline _ = False

isInlineType, isStaticType, isExternType :: C.Type -> Bool
isInlineType ty =  ty `isQualifiedBy` inlineQualid1
                || ty `isQualifiedBy` inlineQualid2
isStaticType ty = ty `isQualifiedBy` staticQualid
isExternType ty = ty `isQualifiedBy` externQualid

inlineQualid1, inlineQualid2, staticQualid :: QualId
inlineQualid1 = mkQualId "inline"
inlineQualid2 = mkQualId "__inline"
staticQualid = mkQualId "static"
externQualid = mkQualId "extern"


isQualifiedBy :: C.Type -> QualId -> Bool
isQualifiedBy (C.Function1 _ _ ty _)  q = ty `isQualifiedBy` q
isQualifiedBy (C.Function2 _ _ ty _)  q = ty `isQualifiedBy` q
isQualifiedBy (C.SimpleTy t)          q = t == q
isQualifiedBy (C.TyApp t1 t2)         q = isQualifiedBy t1 q || isQualifiedBy t2 q
isQualifiedBy (C.Ptr     ty)          q = ty `isQualifiedBy` q
isQualifiedBy (C.Array _ ty)          q = ty `isQualifiedBy` q
isQualifiedBy (C.Const   ty)          q = ty `isQualifiedBy` q
isQualifiedBy (C.TyAttribute _ ty)    q = ty `isQualifiedBy` q
isQualifiedBy (C.ParensType ty)       q = ty `isQualifiedBy` q
isQualifiedBy _                       q = False

-- Generate declarations for things defined in this file
mkProtos ds = [ C.Init (dcl, C.None) | (C.Fun dcl _ _) <- ds ]

-- generate forward decls for variables
mkVars :: [C.Defn] -> [C.Defn]
mkVars ((C.Init (dcl@(q,ty), C.None)):ds) = mkVars ds
mkVars ((C.Init (dcl@(q,ty), i))     :ds) | isStaticType ty = C.Init (dcl,C.None):mkVars ds
mkVars ((C.Init (dcl@(q,ty), i))     :ds) | otherwise       = mkExternVar dcl:mkVars ds
mkVars (_                            :ds) = mkVars ds
mkVars []                                 = []

mkExternVar :: C.Decl -> C.Defn
mkExternVar dcl = C.Extern (C.Init (dcl, C.None))

-- type declarations, but no definitions or prototypes
justTypes :: [C.Defn] -> [C.Defn]
justTypes (d@(C.Fun _ _ _)                 :ds) =     justTypes ds
justTypes (d@(C.Extern (C.Init (_,C.None))):ds) =     justTypes ds
justTypes (d@(C.Extern (C.Init (_,_)))     :ds) =     justTypes ds
justTypes (d@(C.Init (_,C.None))           :ds) =     justTypes ds
justTypes (d@(C.Init (_,_))                :ds) =     justTypes ds
justTypes (d                               :ds) = d : justTypes ds
justTypes [] = []

-- all symbols functions and variables) with program scope (not file scope)
-- we can exclude inline functions because they're not affected by flattening
-- anyway.
justGlobals :: [C.Defn] -> [QualId]
justGlobals (d@(C.Fun (q,ty) _ _)            :ds) | not (isStaticType ty || isInlineType ty) = q : justGlobals ds
justGlobals (d@(C.Init ((q,ty),_))           :ds) = q : justGlobals ds
justGlobals (d@(C.Extern (C.Init ((q,ty),_))):ds) = q : justGlobals ds
justGlobals (d                               :ds) = justGlobals ds
justGlobals [] = []

-- extern var decls only
justExterns :: [C.Defn] -> [C.Defn]
justExterns (d@(C.Extern (C.Init (_,C.None))):ds) = d : justExterns ds
justExterns (d                               :ds) =     justExterns ds
justExterns [] = []

-- common var decls only
justCommons :: [C.Defn] -> [C.Defn]
justCommons ds = [ d | d@(C.Init (dcl, C.None)) <- ds ]

-- non-common var decls
justVars :: [C.Defn] -> [C.Defn]
justVars ((C.Extern d@(C.Init (_,i))):ds) | notNone i = d : justVars ds
justVars (d@(C.Init (_,i))           :ds) | notNone i = d : justVars ds
justVars (d                          :ds)             =     justVars ds
justVars [] = []

justFuns :: [C.Defn] -> [C.Defn]
justFuns ds = [ d | d@(C.Fun _ _ _) <- ds ]

notNone C.None = False
notNone _      = True

uniqDefns :: [QualId] -> [C.Defn] -> [C.Defn]
uniqDefns seen [] = []
uniqDefns seen (d@(C.Extern (C.Fun (q,ty) _ _)):ds)  -- extern functions
  | q `elem` seen
  = uniqDefns seen ds
  | otherwise
  = d : uniqDefns (q:seen) ds
uniqDefns seen (d@(C.Fun (q,ty) _ _):ds)  -- extern functions
  | isExternType ty && q `elem` seen
  = uniqDefns seen ds
  | otherwise
  = d : uniqDefns (q:seen) ds
uniqDefns seen (d@(C.Extern (C.Init ((q,ty),C.None))):ds)
  | q `elem` seen
  = uniqDefns seen ds
  | otherwise
  = d : uniqDefns (q:seen) ds
uniqDefns seen (d@(C.Init ((q,ty),C.None)):ds)
  | q `elem` seen
  = uniqDefns seen ds
  | otherwise
  = d : uniqDefns (q:seen) ds
uniqDefns seen (d:ds) = d : uniqDefns seen ds


----------------------------------------------------------------
-- Mappings (i.e., lookup tables for QualIds)
--
-- We don't just use finitemap because we may want to try to 
-- optimise updates by maintaining a delta-list to be searched
-- before doing a real lookup (and flush the deltas when the
-- list length gets too high).
--
-- Also, it lets us drop the index which is always QualId.
-- And, finally, it lets us make the ordering match corresponding
-- functions on association lists.
----------------------------------------------------------------

newtype Map a = Map (FiniteMap QualId a)

unMap :: Map a -> FiniteMap QualId a
unMap (Map x) = x

emptyMap     :: Map a
listToMap    :: [(QualId,a)] -> Map a
listToMap'   :: (a -> a -> a) -> [(QualId,a)] -> Map a
mapToList    :: Map a -> [(QualId,a)]
mapMap       :: (QualId -> a -> b) -> (Map a -> Map b)
keysMap      :: Map a -> [QualId]
eltsMap      :: Map a -> [a]
elemMap      :: QualId -> Map a -> Bool
lookupMap    :: QualId -> Map a -> Maybe a
filterMap    :: (QualId -> a -> Bool) -> Map a -> Map a
partitionMap :: (QualId -> a -> Bool) -> Map a -> (Map a, Map a)

addToMap       :: QualId -> a   -> Map a -> Map a
addListToMap   :: [(QualId, a)] -> Map a -> Map a
updateMap      :: (a -> a -> a) -> QualId -> a   -> Map a -> Map a
updateListMap  :: (a -> a -> a) -> [(QualId, a)] -> Map a -> Map a
delFromMap     :: QualId        -> Map a -> Map a
delListFromMap :: [QualId]      -> Map a -> Map a
plusMap_C      :: (a -> a -> a) -> Map a -> Map a -> Map a

emptyMap          = Map emptyFM
listToMap         = Map . listToFM
mapToList         = fmToList . unMap
keysMap           = keysFM . unMap
eltsMap           = eltsFM . unMap
elemMap      x xs = elemFM   x (unMap xs)
lookupMap    x xs = lookupFM (unMap xs) x
mapMap       f xs = Map (mapFM f (unMap xs)) -- not quite a monadic map!
filterMap    f xs = Map (filterFM f (unMap xs))
partitionMap f xs = (filterMap f xs, filterMap g xs) where g x y = not (f x y)

addToMap        a b xs = Map (addToFM (unMap xs) a b)
addListToMap     as xs = Map (addListToFM (unMap xs) as)
updateMap     f a b xs = Map (addToFM_C f (unMap xs) a b)
updateListMap f  as xs = Map (addListToFM_C f (unMap xs) as)
delFromMap       a  xs = Map (delFromFM (unMap xs) a)
delListFromMap   as xs = Map (delListFromFM (unMap xs) as)
plusMap_C      c xs ys = Map (plusFM_C c (unMap xs) (unMap ys))

listToMap' f xs = Map (addListToFM_C f emptyFM xs)

----------------------------------------------------------------
-- Random utilities
----------------------------------------------------------------

accum :: Int -> [(a,Int)] -> [(a,Int)]
accum a [] = []
accum a ((x,y):xys) = ((x,y') : accum y' xys)
 where
  y' = a + y

-- a kind of function composition for binary functions
comp2 :: (b -> b -> c) -> (a -> b) -> (a -> a -> c)
comp2 f g x y = g x `f` g y

traceShow :: Show a => String -> a -> a
traceShow msg x = trace (msg ++ show x) x

minBy :: (a -> a -> Ordering) -> (a -> a -> a)
minBy cmp x y = case cmp x y of 
  LT -> x
  EQ -> x -- arbitrary
  GT -> y

----------------------------------------------------------------
-- Test
----------------------------------------------------------------

test x y = do
  bs <- flattenBindings env exps [binding] 
  print (vmap ppBinding' bs)
 where
  env = [("KNIT_BUDGET", show y)]
  binding = (qual, [], [], True, [(SrcDir Nothing x,[])])
  qual = mkQual [mkIdent "qual"]
  exps = [] -- no exports (all dead code?)

----------------------------------------------------------------
-- End
----------------------------------------------------------------
