-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module Main( main ) where

import System( getArgs )
import Monad( when )
import List( sort, partition )

-- import PrettyUtils
import ParseUnits( readDefns )
import qualified PPUnits as PP
import AbstractUnits

import ParseC
import qualified PP1 as PPC
import qualified AbstractC as C
import Maybe(fromMaybe)
import MonadUtils( concatMapM )
import ListUtils( chopList )

main = do
  args <- getArgs
  let (env,args') = getEnv args
      unitDirs    = chopList (== ':') $ fromMaybe err $ lookup "UNIT_PATH" env
      err = error "You must specify UNIT_PATH=..."
      [unit_file,bt_name] = args'
      bt_name' = mkIdent bt_name

  when (length args' /= 2) $
    error "Usage: mk_bundletype unit_file bt_name"

  xs <- readDefns unitDirs [unit_file]
  let
    f_env :: [(FlagsIdent, [String])]
    f_env = [ (x, evalFlags f_env f) | (x, Flags f) <- xs ]

    bt :: BundleType
    BundleT bt = fromMaybe err (lookup bt_name' xs)
    err = error $ "Can't find bundleIdent " ++ show (PP.bundleIdent bt_name')

  bt' <- bte "knit_c_parser" env f_env bt bt_name'
  print $ PP.bundleType bt'


evalFlags :: [(FlagsIdent, [String])] -> Flags -> [String]
evalFlags env flags = concatMap (evalFlag env) flags

evalFlag env (RawFlag str) = [str]
evalFlag env (UseFlags fi) = evalFI env fi

evalFI :: [(FlagsIdent, [String])] -> FlagsIdent -> [String]
evalFI env i = fromMaybe err (lookup i env)
 where
  err = error $ "Can't find flagsIdent " ++ show (PP.flagIdent i)

-- This extracts function prototypes from all the files included into
-- a bundletype using the flags as appropriate.
--
-- We use a gcc extension to achieve the effect of processing a single
-- C file containing:
--
--   #include "foo.h"
--   #include "bar.h"
--
-- by executing:
--
--   gcc -P -E -include foo.h -include bar.h -xc /dev/null
--
-- The more obvious: 
--
--   gcc -P -E foo.h bar.h
--
-- Processes each file separately so definitions in foo.h aren't available
-- in bar.h

bte :: String -> [(String,String)] -> [(FlagsIdent,[String])] -> BundleType -> BundleTIdent -> IO BundleType
bte cparser env fenv (bt,fs) x = do
  xs <- parseDefnsEx cparser env' (evalFlags fenv fs ++ mkIncs incs) "/dev/null" (show x)
  return ( map Include incs 
           ++ [ Ident (unQualify nm) | x <- xs, Just nm <- [nameOf x] ]
         , fs
         )
 where
  incs = [ x | Include x <- bt ]
  ms   = [ x | Ident   x <- bt ]
  env' = [ x ++ "=" ++ y | (x,y) <- env ]

  mkIncs fs = map ("-include " ++) fs ++ ["-xc"]

-- just good enough to find variable and function declarations/definitions
nameOf (C.Init ((nm,_),_)) = Just nm
nameOf (C.Fun  (nm,_) _ _) = Just nm
nameOf (C.Extern x)        = nameOf x
nameOf _                   = Nothing

getEnv :: [String] -> ([(String,String)],[String])
getEnv args = (map split env,rest)
 where
  (env, rest) = partition ('=' `elem`) args
  split x = (pre, tail post) where (pre,post) = break (== '=') x

----------------------------------------------------------------
-- End
----------------------------------------------------------------
