-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module Id
        ( Ident, fromIdent, mkIdent, hashIdent, freshIdent
        , Qualifier, fromQualifier, emptyQual, extendQual, concatQual, mkQual
        , QualId, toQualId, qualify, mkQualId, mkQualId2, unQualify, showQualifier, showQualId, qualifierOf, asQualifier, identOf
        , Hashable(..)
        , HashTable, hlookup, hinsert, mkHashTable, accumHashTable

        -- Pretty-print
        , ppIdent
        , ppQualifier,  qualifier'C
        , ppQualid, qualident'C
        ) where

import IOExts -- for Ident hackery
import IOExts( unsafePerformIO ) -- for Ident hackery
import PrettyUtils
import FiniteMap
import FastString

----------------------------------------------------------------
-- Pretty-print
----------------------------------------------------------------

ppIdent :: Ident -> Doc
ppIdent = text . show

ppQualifier :: Qualifier -> Doc
ppQualifier = text . show

ppQualid :: QualId -> Doc
ppQualid = text . show

-- a valid C identifier that is (hopefully) unlikely to occur in any program
-- it also has to be valid in shell scripts because of the way we generate
-- code at the moment - so don't use $
qualident'C :: QualId -> Doc
qualifier'C :: Qualifier -> Doc

qualident'C q = text (showQualId '_' q)
qualifier'C q = text (showQualifier '_' q)

{-
-- Use this if you want to use (an approximation of) the C++ name 
-- mangler.  May be more gdb friendly.
qualident'C (MkQ q) = ppPrefix (reverse quals) <> text (fromIdent nm)
 where
  (nm:quals) = fromQualifier q

qualifier'C q = ppPrefix (fromQualifierRev q)

ppPrefix [] = empty
ppPrefix quals = text "_Q_" <> int (length quals) <> char '_' 
               <> hcat (map ppQ quals)
               <> char '$' 
 where
  ppQ q = int (length q') <> text q'
   where q' = (fromIdent q)
-}

----------------------------------------------------------------
-- Qualified names
----------------------------------------------------------------

newtype QualId = MkQ Qualifier deriving (Eq, Ord)

instance Show QualId where showsPrec d (MkQ q) = showsPrec d q

toQualId :: Qualifier -> QualId
toQualId = MkQ

qualify :: Qualifier -> Ident -> QualId
qualify q i = MkQ (extendQual i q)

showQualId :: Char -> QualId -> String
showQualId c (MkQ q) = showQualifier c q

mkQualId :: String -> QualId
mkQualId = qualify EmptyQual . mkIdent

mkQualId2 :: Ident -> QualId
mkQualId2 = qualify EmptyQual

unQualify :: QualId -> Ident
unQualify (MkQ (ExtendQual _ x EmptyQual)) = x
unQualify x = error $ "Function arguments can't be qualified"

qualifierOf :: QualId -> Qualifier
qualifierOf (MkQ (ExtendQual _ x q)) = q
qualifierOf x = error $ "Function arguments can't be qualified"

identOf :: QualId -> Ident
identOf (MkQ (ExtendQual _ x q)) = x
identOf x = error $ "Function arguments can't be qualified"

asQualifier :: QualId -> Qualifier
asQualifier (MkQ q) = q

----------------------------------------------------------------
-- Identifiers
--
-- It's a safe bet that we're going to spend a lot of time
-- comparing identifiers so we play a low, underhand trick
-- to make this go a little faster than normal: we tag every
-- identifier with a unique integer and just compare the ints.
--
-- This is, of course, impossible in a purely functional language :-)
----------------------------------------------------------------

freshIdent :: IO Ident -- makes a unique anonymous identifier
mkIdent   :: String -> Ident
fromIdent :: Ident -> String
hashIdent :: Ident -> Int     -- unique!

data Ident = MkId !Int FastString

instance Show Ident where showsPrec d x r = fromIdent x ++ r
instance Eq   Ident where MkId x _ == MkId y _ = x == y
--instance Ord  Ident where MkId _ x <= MkId _ y = x <= y
instance Ord  Ident where MkId x _ <= MkId y _ = x <= y

mkIdent   s          = forceFS s' `seq` unsafePerformIO (insert mkIdent' idTree s')
 where 
  s' = toFastString s
fromIdent (MkId u s) = fromFastString s  -- ++ "$" ++ show u
hashIdent (MkId u s) = u

idTree :: MBTree FastString Ident
idTree = unsafePerformIO $ newIORef emptyT

uniqueId :: IORef Int
uniqueId = unsafePerformIO $ newIORef 0

freshIdent = do
  u <- readIORef uniqueId
  writeIORef uniqueId $! (u+1)
  let nm = toFastString $ "__fresh_" ++ show u
  forceFS nm `seq` return ()
  return (MkId u nm)

mkIdent' :: FastString -> IO Ident
mkIdent' y = do
  u <- readIORef uniqueId
  writeIORef uniqueId $! (u+1)
  return (MkId u y)

----------------------------------------------------------------
-- Qualifiers
-- 
-- Similarily, qualifiers are often compared so we represent them 
-- a little more efficiently.
----------------------------------------------------------------

data Qualifier
  = ExtendQual !Int Ident Qualifier
  | EmptyQual 

concatQual :: Qualifier -> Qualifier -> Qualifier
extendQual :: Ident -> Qualifier -> Qualifier
mkQual     :: [Ident] -> Qualifier

instance Eq Qualifier where 
  q1 == q2 = hashQual q1 == hashQual q2

instance Ord Qualifier where 
  compare q1 q2 = compare (hashQual q1) (hashQual q2)
{-
  compare EmptyQual EmptyQual          = EQ
  compare EmptyQual (ExtendQual _ _ _) = LT
  compare (ExtendQual _ _ _) EmptyQual = GT
  compare (ExtendQual u1 x1 q1) (ExtendQual u2 x2 q2) 
    | u1 == u2  = EQ
    | otherwise = primCompAux x1 x2 (compare q1 q2)

primCompAux      :: Ord a => a -> a -> Ordering -> Ordering
primCompAux x y o = case compare x y of EQ -> o; LT -> LT; GT -> GT
-}

instance Show Qualifier where
  showsPrec d q r = (punctuate' '.' q) r

showQualifier :: Char -> Qualifier -> String
showQualifier c q = punctuate' c q ""

punctuate' :: Char -> Qualifier -> ShowS
punctuate' c EmptyQual r    = r
punctuate' c (ExtendQual _ x EmptyQual) r = shows x r
punctuate' c (ExtendQual _ x q)         r = punctuate' c q (c : shows x r)

fromQualifier :: Qualifier -> [Ident]
fromQualifier q = aux q []
 where
  aux (ExtendQual u x q) r = aux q (x:r)
  aux EmptyQual          r = reverse r

fromQualifierRev :: Qualifier -> [Ident]
fromQualifierRev q = aux q []
 where
  aux (ExtendQual u x q) r = aux q (x:r)
  aux EmptyQual          r = r

hashQual :: Qualifier -> Int
hashQual (ExtendQual x _ _) = x
hashQual EmptyQual          = 0

mkQual = foldr extendQual EmptyQual . reverse

concatQual EmptyQual q = q
concatQual (ExtendQual _ i q1) q2 = extendQual i (concatQual q1 q2)

-- the seq is essential to ensure that lookups don't cause
-- further lookups
extendQual x q = q `seq` unsafePerformIO (insert extendQual' qTree (x,q))
emptyQual      = EmptyQual

qTree :: MBTree (Ident,Qualifier) Qualifier
qTree = unsafePerformIO $ newIORef emptyT

uniqueQ :: IORef Int
uniqueQ = unsafePerformIO $ newIORef 1 -- 0 = empty

extendQual' (x,q) = q `seq` do
  u <- readIORef uniqueQ
  writeIORef uniqueQ $! (u+1)
  return $! (ExtendQual u x q)
  

----------------------------------------------------------------
-- MBTree
--
-- It's all based on mutable binary trees
-- (or, maybe on finitemaps)
----------------------------------------------------------------

type MBTree a b = IORef (MBTree' a b)

emptyT :: MBTree' a b
insert :: Ord a => (a -> IO b) -> MBTree a b -> a -> IO b

type MBTree' a b = FiniteMap a b
emptyT = emptyFM
insert mk t y = do
  t' <- readIORef t
  case lookupFM t' y of
    Nothing -> do
      id <- mk y
      writeIORef t $! addToFM t' y id
      return id
    Just id ->
      return id

{-
-- Unbalanced implementation: terrible performance!
data MBTree' a b
  = Node a b (MBTree a b) (MBTree a b)
  | Leaf

emptyT = Leaf

insert mk t y = do
  t' <- readIORef t
  case t' of
   Node x i l r -> case compare x y of
                  LT -> insert mk l y
                  EQ -> return i
                  GT -> insert mk r y
   Leaf -> do
    id <- mk y
    l <- newIORef Leaf
    r <- newIORef Leaf
    writeIORef t $! (Node y id l r)
    return id
-}
----------------------------------------------------------------
-- Hashtables
--
-- We can also use the uniques as indices into hashtables
-- (also binary trees)
----------------------------------------------------------------

class Hashable a where hash :: a -> Int

instance Hashable Int       where hash = id
instance Hashable Ident     where hash = hashIdent
instance Hashable Qualifier where hash = hashQual
instance Hashable QualId    where hash (MkQ q) = hashQual q

data HashTable h a
  = EmptyHash
  | Hash Int a !(HashTable h a) !(HashTable h a)

hlookup :: Hashable h => h -> HashTable h a -> Maybe a
hlookup x t = hlookup' (hash x) t 

hlookup' :: Int -> HashTable h a -> Maybe a
hlookup' x EmptyHash = Nothing
hlookup' x (Hash x' y l r)
  | x < x'    = hlookup' x l
  | x > x'    = hlookup' x r
  | otherwise = Just y

-- doesn't overwrite old value
hinsert :: Hashable h => h -> a -> HashTable h a -> HashTable h a
hinsert x y t = case hlookup' h t of
  Nothing -> hinsert' h y id t 
  Just _  -> t
 where
  h = hash x

-- overwrites old value
hinsert' :: Int -> a -> (a -> a) -> HashTable h a -> HashTable h a
hinsert' x y f EmptyHash = Hash x y EmptyHash EmptyHash
hinsert' x y f (Hash x' y' l r)
  | x < x'    = Hash x' y' (hinsert' x y f l) r
  | x > x'    = Hash x' y' l (hinsert' x y f r)
  | otherwise = Hash x' (f y') l r

mkHashTable :: Hashable h => [(h, a)] -> HashTable h a
mkHashTable xs = foldl (\t (h,a) -> hinsert h a t) EmptyHash xs

accumHashTable :: (Hashable h, Eq a) => [(h,a)] -> HashTable h [a]
accumHashTable xs = foldl' (\ t (h,a) -> (hinsert' $! hash h) [a] (a:) t) EmptyHash xs
{-
 where 
  insert x xs
    | x `elem` xs = xs
    | otherwise   = x:xs
-}

-- copied from Hugs Prelude
foldl'           :: (a -> b -> a) -> a -> [b] -> a
foldl' f a []     = a
foldl' f a (x:xs) = (foldl' f $! f a x) xs


-- hixs :: HashTable h a -> [Int]
-- hixs h = hixs' h []
--  where
--   hixs' EmptyHash      is = is
--   hixs' (Hash i _ l r) is = hixs' l (hixs' r (i : is))

instance Functor (HashTable h) where
  fmap f EmptyHash = EmptyHash
  fmap f (Hash x y l r) = Hash x (f y) (fmap f l) (fmap f r)

----------------------------------------------------------------
-- End 
----------------------------------------------------------------
