-- Copyright (c) 2000, 2001 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module Main( main ) where

import System( getArgs )
import Monad( when )
import List( partition, intersect, union )
import ListUtils( chopList )
import Maybe( catMaybes, fromMaybe, fromJust)
import PrettyUtils

import qualified PPUnits as PP
import AbstractUnits
import ParseUnits( readDefns )
import Id(mkIdent)

usage = "Usage: knitGenBundles {var=value} unit_file unit_name -- imports -- exports -- non_uses -- files"

main = do
  args <- getArgs
  let arg_sections = chopList (== "--") args
      [knit_args,imports,exports,non_uses,files] = arg_sections
      (env,args') = getEnv knit_args
      unitsDirs   = chopList (== ':') $ fromMaybe err $ lookup "UNIT_PATH" env
      err = error "You must specify UNIT_PATH=..."
      [unitFile,unitName]  = args'

  when (length arg_sections /= 5) $ error usage
  when (length args' /= 2) $ error usage

  xs <- readDefns unitsDirs [unitFile]
  let
    bts :: [(BundleTIdent, BundleT')]
    bts = [ (x, evalBT bts bt) | (x, BundleT bt) <- xs ]

  let
    imports' = map mkIdent imports
    exports' = map mkIdent exports
    non_uses'   = map mkIdent non_uses

  print 
    $  text "unit" <+> text unitName <+> equals <+> lbrace
    $$ nest 2 
       (  text "imports" <> (lbrack <+> vcat (genImports bts imports') $$ rbrack <> semi)
       $$ text "exports" <> (lbrack <+> vcat (genExports bts exports' non_uses') $$ rbrack <> semi)
       $$ text "depends{ exports + inits + finis needs imports };"
       $$ text "files{"
       $$ nest 2 (vcat [ text f <> comma | f <- files ])
       $$ text "} with flags {};"
       )
    $$ rbrace

-- We generate export lists by looking for signatures in which every member
-- is a member of the export list.  On exports, it's not ok to export
-- things you don't define. 
genExports :: [(BundleTIdent,BundleT')] -> [MemberIdent] -> [MemberIdent] -> [Doc]
genExports bts exports non_uses
  =  map ppExport bts'' ++ 
  [  if null undefined
     then empty
     else text "out" <+> colon <+> braces (hlist PP.memberIdent undefined) <> comma
  ] ++ 
  [  if null unused
     then empty
     else text "unused" <+> colon <+> braces (hlist PP.memberIdent unused) <> comma
  ]
 where
  -- list all the candidates
  bts' = [ bt | bt@(x,ms) <- bts, ms `isSubBundleType` exports ]

  -- remove any which are strict subsets of others
  bts'' = [ bt 
          | bt@(x,ms) <- bts'
          , null [ () | bt'@(_,ms') <- bts'
                      , ms `isSubBundleType` ms'
                      , not (ms' `isSubBundleType` ms) ]
          ]

  -- figure out what's left
  defined   = foldr union [] (map snd bts')
  (unused,undefined) = partition (`elem` non_uses) 
                     $ filter (not . (`elem` defined)) 
                     $ exports

ppExport :: (BundleTIdent,BundleT') -> Doc
ppExport (b,ms) 
  =   PP.bundleTIdent b <+> colon <+> PP.bundleTIdent b <> comma
  <+> text "/*" <+> braces (hlist PP.memberIdent ms) <+> text "*/"


isSubBundleType :: BundleT' -> BundleT' -> Bool
isSubBundleType xs ys = all (`elem` ys) xs

-- We generate import lists by looking for signatures that cover 
-- the import set.  On imports, it's ok to import some things you 
-- don't use.
genImports :: [(BundleTIdent,BundleT')] -> [MemberIdent] -> [Doc]
genImports bts imports  
  =  map ppGroup groups ++
  [  if null undefined
     then empty
     else text "in" <+> colon <+> braces (hlist PP.memberIdent (map fst undefined)) <> comma
  ]
 where
  -- First generate a list of bundletypes that each symbol appears in
  definers :: [(MemberIdent, [BundleTIdent])]
  definers = [ (s, [ x | (x,ms) <- bts, s `elem` ms ]) | s <- imports ]
  
  -- Then collect together members defined in the same bundle
  -- The result will depend to some degree on the order members are listed
  groups :: [([MemberIdent],[BundleTIdent])]
  groups = foldr insertInGroup [] defined
  
  isDefined :: (MemberIdent, [BundleTIdent]) -> Bool
  isDefined (_,bts) = not (null bts)
  
  (defined,undefined) = partition isDefined definers

type Group = ([MemberIdent],[BundleTIdent])

ppGroup (ms, b:bs) 
  =   PP.bundleTIdent b <+> colon <+> PP.bundleTIdent b <> comma
  <+> text "/*" <+> hlist id alts <+> text "*/"
 where
  -- alternative answers
  alts :: [Doc]
  alts = braces (hlist PP.memberIdent ms) 
       : map PP.bundleTIdent bs

-- Insert member in any group that shares some signatures with member
insertInGroup :: (MemberIdent, [BundleTIdent]) -> [Group] -> [Group]
insertInGroup (m,is) [] = [([m],is)]
insertInGroup (m,is) (g@(ms,is'):gs)
  | null common
  = g : insertInGroup (m,is) gs
  | otherwise
  = (m:ms, common) : gs
 where
  common = intersect is is'


ppBundleT :: (BundleTIdent, BundleT') -> Doc
ppBundleT (x, bt) =  text "type" <+> PP.bundleTIdent x
                  $$ nest 2 (vmap PP.memberIdent bt)

type BundleT' = [MemberIdent]

evalBT :: [(BundleTIdent, BundleT')] -> BundleType -> BundleT'
evalBT env = concatMap (evalBTE env) . fst

evalBTE :: [(BundleTIdent, BundleT')] -> BTElement -> BundleT'
evalBTE env (Ident x)   = [x]
evalBTE env (Include f) = []
evalBTE env (Extends i) = evalBTI env i

evalBTI :: [(BundleTIdent, BundleT')] -> BundleTIdent -> BundleT'
evalBTI env i = fromMaybe err (lookup i env)
 where
  err = error $ "Can't find bundleIdent " ++ show (PP.bundleIdent i)


getEnv :: [String] -> ([(String,String)],[String])
getEnv args = (map split env,rest)
 where
  (env, rest) = partition ('=' `elem`) args
  split x = (pre, tail post) where (pre,post) = break (== '=') x
