-- Copyright (c) 2000, 2001 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module MakeMakefile
	( mmf
	) where

import EvalUnits
import AbstractUnits
import ParseUnits( readDefns )
import qualified PPUnits as PP

import Flatten( flattenBindings )
import Constraints( checkConstraints )
import Schedule( schedule )

import ZEncode
import PrettyUtils
import ListUtils( chopList )
import Monad( mapAndUnzipM, unless, when )
import Maybe( catMaybes, fromMaybe, fromJust)
import System( system, ExitCode(..) )
import IO

-- debugging
import Constraints( ppQConstraints )

-- -- experiment
-- import Fact 
-- import Constraints

-- ----------------------------------------------------------------
-- -- Random hackery to convert constraints and dependencies into
-- -- a uniform framework
-- ----------------------------------------------------------------
-- 
-- depToFact :: Depend' -> Fact
-- depToFact (xs, r, ys) = Fact.App (relToId r) [Idents xs, Idents ys]
-- 
-- relToId Calls    = needs_id
-- relToId Precedes = precs_id
-- 
-- needs_id = mkIdent "needs"
-- precs_id = mkIdent "<"
-- 
-- conToFact :: Constraint' -> Fact
-- conToFact (Eq' p1 p2) = Fact.App eq_id [propToFact p1, propToFact p2]
-- conToFact (Le' p1 p2) = Fact.App le_id [propToFact p1, propToFact p2]
-- 
-- le_id = mkIdent "<="
-- eq_id = mkIdent "=="
-- 
-- propToFact :: Property' -> Fact
-- propToFact (PropApp' (MkPred p) xs) = Fact.App p [Idents xs]
-- propToFact (PropClass (MkClass c _)) = Fact.Atom c

----------------------------------------------------------------
-- The main thing (makemakefile)
----------------------------------------------------------------

mmf :: FilePath -> String -> [(String,String)] -> Bool -> Bool -> Bool -> IO ()
mmf unitDefs unitName env flatten check no_codegen = do
  putStrLn "Parsing"
  xs <- readDefns unitsDirs [unitDefs]
  print $ pp' (length xs) <+> text "declarations"

  putStrLn "Evaluating"
  let u@(exps,cs,ds,is,fs,bs) = evalUnit xs unitName'
  return $! length cs -- force that puppy

--  putStrLn $ show (text "Dependencies:" <+> vmap ppFact (map depToFact ds))
--  putStrLn $ show (text "Constraints:"  <+> vmap ppFact (concatMap (map conToFact.snd) cs))

  when check $ do
    putStrLn "Checking Constraints"
    checkConstraints cs

  print $  pp' (length ds) <+> text "dependencies"
  putStrLn "Scheduling"
  let (init,fini) = schedule is fs exps ds
  print $ text "init" <> colon <+> vmap pp' init
  print $ text "fini" <> colon <+> vmap pp' fini

  bs1 <- if flatten 
         then do
           putStrLn "Flattening"
           -- don't forget that initialisers are "exported" too
           bs' <- flattenBindings env (init ++ fini ++ exps) bs
--           length bs' `seq` return bs'  -- force that puppy
           return bs'
         else return bs

  if no_codegen
    then do
      putStrLn "Not generating any code"
    else do
    
      let 
          -- The following look equivalent but the unzip version has
          -- a huge space leak.
    --      (bs',anonss) = unzip (map resolveBinding bs1)
          bs'    = map fst (map resolveBinding bs1)
          anonss = map snd (map resolveBinding bs1)
    
          (oss, rn_files, rns) = unzip3 (map (rename mv) bs')
          os = concat oss
    
          -- We'd better force this puppy soon 'cos anonss contains a pointer
          -- to the output of the flattener which tends to be BIG.
          anon_files = [ f | anons <- anonss, (f,c) <- anons ]
    
      seqList seqString anon_files $ return ()
    
      sum [ length fs' | (_,_,_,_,fs') <- bs' ] `seq` return ()  -- more forcing!
    
      putStrLn "Writing Anon Files"
      sequence_ [ mkFile mv f c | anons <- anonss, (f,c) <- anons ]
      putStrLn "Writing Files"
      sequence_ rn_files
      mkFile mv "knit_inits.c" $ map show
        [ ppSchedule "knit_init" init
        , ppSchedule "knit_fini" fini
        ] 
      -- We put the .o files into .a files so that ld can discard unneeded
      -- .o files.
      -- We use multiple .o files because building .a files takes quadratic
      -- time.  The choice of how many .a files to generate (i.e., how big
      -- each .a file should be) is completely random.
      -- We should use a hash function on the name so that the mapping of .o to .a
      -- is relatively stable under changes to the unit description.
      let oss = chunks 100 os
    
      -- We pair each set of .o files with a filename.
      let foss = zip (map mkFileName [0..]) oss
            where mkFileName x = "foo" ++ show x ++ ".a"
          libs = map fst foss
    
      putStrLn $ "Generating makefile insert: " ++ makeFile
      writeFile makeFile $ 
        showDoc (  text "# Machine generated makefile: do not modify"
                $$ vmap (\ (x,y) -> text (x ++ "=" ++ y)) env
                $$ text ("KNIT_OBJS = " ++ unwords os)
                $$ text ("KNIT_LIBS = " ++ unwords libs)
                $$ text ("RENAME_DOT_O_FILES = " ++ rename_dot_o_files)
                $$ vsep (map compile bs')
                $$ vsep rns
                $$ vcat [ makeRule fn os
                            (  command "$(RM)" [ fn ]
                            $$ echo_command "$(AR) csq" (fn : os)
                            )
                        | (fn, os) <- foss
                        ]
                $$ vcat [ text $ "-include " ++ basename oFile ++ ".d"
                        | oFile <- os
                        ]
                $$ text "# End of machine generated part of Makefile"
                $$ text ""
                )
      maybe (return ()) (\ mk -> do
          putStrLn $ "Copying makefile_template " ++ mk ++ " into makefile"
          systemEx "cp" [mk, "makefile"]
          return ()
        )
        (lookup "MAKEFILE" env)
      return ()

    -- end of codegen
  putStrLn "Done"
 where
  shellScript = "knit.sh"
  makeFile    = "knit_generated.mk"
  unitName'   = mkIdent unitName
  unitsDirs   = chopList (== ':') $ fromMaybe err $ lookup "UNIT_PATH" env
   where err = error "You must specify UNIT_PATH=..."
  bindir      = maybe "" (++"/") $ lookup "KNIT_TOOLS" env
  mv          = bindir ++ "knit_smartmv"
  rename_dot_o_files 
              = bindir ++ "rename_dot_o_files"

----------------------------------------------------------------
-- Replace relative filenames with "absolute" names.
--
-- Also generates a list of C files that need to be generated
-- because the C code was inlined in the main program.
----------------------------------------------------------------

type Binding'' = 
  ( Qualifier            -- prefix for local names
  , Bundle''              -- requires
  , Bundle''              -- provides
  , Bool                 -- does it need flattened (smooshed into giant C file)
  , [(FilePath,[Flag'])]
  )

resolveBinding :: Binding' -> (Binding'',[(FilePath,[String])])
resolveBinding (qual,rs,ps,flatten,fs) = 
  seqList (seq2 seqString seq) fs' $
  seqList (seqMb (seq2 seqString noseq)) ds $
  ((qual,rs,ps,flatten,fs'), catMaybes ds)
 where
  (fs',ds) = unzip (zipWith resolveFile fs fns)

  fns = [ show qual ++ "_anon_" ++ show n ++ ".c" | n <- [0..] ]

  resolveFile :: (RelativeFileName,[Flag']) -> FilePath -> ((FilePath,[Flag']), Maybe (FilePath,[String]))
  resolveFile (f@(SrcDir _ _), flags) fn = ((relativeName f, flags), Nothing)  
  resolveFile (AnonCode s, flags)     fn = ((fn,             flags), Just (fn,s))

----------------------------------------------------------------
-- Compiling
----------------------------------------------------------------

compile :: Binding'' -> Doc
compile (qual,_,_,flatten,files) = vsep (map compileFile files)
 where  
  compileFile :: (FilePath,[Flag']) -> Doc
  compileFile (cFile, flags) = makeC cFile oFile flags
   where
    prefix     = if null prefix' then prefix' else prefix' ++ "_"
    prefix'     = xEncode (show (qualifier'C qual))
    oFile      = cToO $ prefix ++ end ++ ".raw"
    end = xEncode cFile -- (base,end) = splitFileName cFile
    
makeC :: String -> String -> [String] -> Doc
makeC cFile oFile cmdLine 
  | last cFile == 'c'
  = makeRule oFile [cFile] 
--     (  echo_command "$(CC_CACHE)" $ 
     (  echo_command "$(CC)" (
                      [ "$(CFLAGS)"
                      , "-o", "$@"
                      , "-c", "$<"
                      ] ++ cmdLine
       )
     $$ text "if" <+> parens (command "test" ["-r", basename (notdir cFile) ++ ".d"]) <> semi <+> text "then" 
        <+> command "sed" (
            [ "-e"
            , "'s/^.*:/" ++ oFile ++ ":/'"
            , basename (notdir cFile) ++ ".d"
            , ">"
            , basename oFile ++ ".d"
            ]
            ) <> semi
        <+> text "fi"
     )
  | last cFile == 'S'
  = makeRule oFile [cFile] $
    echo_command "$(CC)" $ 
                      [ "$(CFLAGS)"
                      , "-o", "$@"
                      , "-c", "$<"
                      ] ++ cmdLine
  | last cFile == 'o' && null cmdLine
  = makeRule oFile [cFile] $
    echo_command "cp" $ [cFile, oFile]
  | otherwise
  = error $ "Can't make " ++ oFile ++ " from " ++ cFile ++ " using flags '" ++ unwords cmdLine ++ "'"

----------------------------------------------------------------
-- Renaming
----------------------------------------------------------------

rename :: FilePath -> Binding'' -> ([FilePath], IO ExitCode, Doc)
rename mv (qual, rs, ps, flatten, fs)
  = ( map (ofile . fst) fs
    , mkFile mv rnFileName (map show (map ppX (rs ++ ps)))
    , vmap rn fs
    )
 where
  ppX (x,y) = PP.ident x <> equals <> qualident'C y

  rn :: (FilePath,a) -> Doc
  rn (f,_) = makeRule (ofile f) [ifile f, rnFileName] $
             (  command "cp" [ ifile f, ofile f ]
                     $$ echo_command "$(RENAME_DOT_O_FILES)" 
                     [ "'" ++ pre ++ "'"
                     , rnFileName  
                     , ofile f
                     ]
             )

  -- when we flatten the C code, we use a null prefix 
  prefix      = xEncode prefix'
  prefix'     = if null prefix'' then prefix'' else prefix'' ++ "_"
  prefix''    = show (qualifier'C qual)

-- Use this normally
  pre = prefix'
--  Use this with C++ style name-mangling
--  pre = yEncode (show (qualifier'C qual))
--   where
--    yEncode = concatMap patch
--     where
--      patch '$' = "$$"  -- protect $ from being interpreted by Makefile
--      patch c   = [c]

  rnFileName = "rename_" ++ xEncode (show qual)   -- blech
  ifile f    = cToO $ prefix ++ end ++ ".raw"
   where
    end = xEncode f -- (base,end) = splitFileName f
  ofile f    = cToO $ prefix ++ end
   where
    end = xEncode f -- (base,end) = splitFileName f

cToO :: String -> String
cToO file = file ++ ".o"

----------------------------------------------------------------
-- Generating the initialisers
----------------------------------------------------------------

ppSchedule :: String -> [QualId] -> Doc
ppSchedule f xs
  =  vmap ppDecl xs
  $$ text "extern void knit_progress(const char*,int);"
  $$ text "int" <+> text f <+> parens (text "void")
  $$ lbrace
  $$ nest 2 (  blank
            $$ vmap ppCall xs
            $$ text "return 0;"
            )
  $$ rbrace

ppCall x =  text "{ int rc;"
         $$ text "rc =" <+> qualident'C x <> parens empty <> semi
         $$ text "knit_progress" <> parens 
                      (   doubleQuotes (ppQualid x)
                      <>  comma
                      <+> text "rc"
                      ) <> semi
         $$ text "if (rc) return rc;"
         $$ text "}"

-- Having taken great care that qualified names will never conflict with
-- valid C names (by using invalid C names, of course), we must make sure
-- we don't use them as C names!  
-- gcc's asm label extension is just the ticket.
ppDecl x =   text "extern" <+> text "int" <+> qualident'C x <> parens (text "void") 
--         <+> text "asm" <> parens (doubleQuotes (ppQualid x))
         <>  semi

----------------------------------------------------------------
-- Filename Utilities
----------------------------------------------------------------

splitFileName :: FilePath -> ([String],String)
splitFileName f = case break (=='/') f of
  (x,'/':y)  -> (x:xs,y') where (xs,y') = splitFileName y
  (x,"")     -> ([],x)
  (x,y)      -> error $ "splitFileName: " ++ y

----------------------------------------------------------------
-- Shell script utilities
----------------------------------------------------------------

command :: String -> [String] -> Doc
command x xs = text (unwords (x:xs))

echo_command :: String -> [String] -> Doc
echo_command x xs
  =  (text "$(KNIT_QUIET_MAKE_INFORM)" <+> text "Building $@")
  $$ (command x xs)

-- here documents are just a fancy way of writing strings which stretch over
-- many lines.           
here_document :: Doc -> Doc -> Doc
here_document l d 
  =  l <+> text "<< EOF"
  $$ d
  $$ text "EOF"

-- only create the file if we need to
mkFile :: FilePath -> String -> [String] -> IO ExitCode
mkFile mv nm contents = do
  really_mkFile "TMP" contents
--  systemEx "install" ["--preserve-timestamps", "TMP", nm]
  systemEx mv ["TMP", nm]

really_mkFile :: String -> [String] -> IO ()
really_mkFile nm contents = do
  h <- openFile nm WriteMode
  mapM_ (hPutStrLn h) contents
  hClose h

----------------------------------------------------------------
-- Makefile utilities
----------------------------------------------------------------

-- It takes wild hackery to generate correctly tabbed makefiles

makeRule :: FilePath -> [FilePath] -> Doc -> Doc
makeRule target dependencies command = makeRule' True target dependencies command

makeRule' :: Bool -> FilePath -> [FilePath] -> Doc -> Doc
makeRule' hard target dependencies command
  =  text target <+> (if hard then colon else text "::") <+> text (unwords dependencies)
  $$ vmap (text . ('\t' : )) (lines (showDoc command))

----------------------------------------------------------------
-- Misc utilities
----------------------------------------------------------------

pp' :: (Show a) => a -> Doc
pp' = text . show

-- split a list into convenient-sized sublists
chunks :: Int -> [a] -> [[a]]
chunks n [] = []
chunks n as = bs : chunks n cs where (bs,cs) = splitAt n as

-- Bugfix for ghc
my_system :: String -> IO ExitCode
my_system xs = do
  -- putStrLn xs -- for debugging
  rc <- System.system ("trap '' 26; " ++ xs)
  unless (rc == ExitSuccess) $ putStrLn xs
  return rc
    
systemEx :: String -> [String] -> IO ExitCode
systemEx cmd args = my_system (unwords (cmd:args))

fst3 :: (a,b,c) -> a
fst3    (a,b,c) =  a

snd3 :: (a,b,c) -> b
snd3    (a,b,c) =  b

thd3 :: (a,b,c) -> c
thd3    (a,b,c) =  c

-- filename functions to match the ones in GNU make
basename :: String -> String
notdir   :: String -> String
dir      :: String -> String

basename xs = reverse $ drop 1 $ dropWhile (/= '.') $ reverse xs
notdir   xs = reverse $ takeWhile (/= '/') $ reverse xs
dir      xs = reverse $ drop 1 $ dropWhile (/= '/') $ reverse xs

----------------------------------------------------------------
-- Forcing functions
----------------------------------------------------------------

type Seq a k = a -> k -> k

seqMb :: Seq a k -> Seq (Maybe a) k
seqMb sq Nothing  c = c
seqMb sq (Just x) c = sq x c

seqList :: Seq a k -> Seq [a] k
seqList sq [] c = c
seqList sq (x:xs) c = sq x (seqList sq xs c)

seqString :: Seq String k
seqString = seqList seq

noseq :: Seq a k
noseq _ k = k

seq2 :: Seq a k -> Seq b k -> Seq (a,b) k
seq2 s1 s2 (x,y) k = s1 x (s2 y k)

----------------------------------------------------------------
-- Test harnesses
--
-- These generate makefiles which are invoked using something like
-- this sequence (in an empty directory):
-- 
--   make CC=/usr/bin/gcc -f ~/knit/compiler/makefile Image
--
----------------------------------------------------------------

-- mt1 unit = mmf "../../knit/compiler/All.unit" unit "../oskit" "../build" "." True False False
-- 
-- -- the flag at the end controls scheduling
-- -- - if you turn it off, it's much faster (but you can't run the build)
-- at1  f unit = mmf f unit "/z/reid/oskit" "/z/reid/build_oskit" "../units/oskit" True   False True
-- fat1 f unit = mmf f unit "/z/reid/oskit" "/z/reid/build_oskit" "../units/oskit" False  False False

----------------------------------------------------------------


-- ToDo: 
-- o shouldn't need to uniquify if I use per-file renaming
-- o take care to uniquify both enum decl and enum use and all statics
--   maybe make list of things to uniquify in advance??
-- o could do better job of non-extern-inlines to avoid duplication
-- o worry about ordering of decls (since it matters so much!)
-- o mark non-exported code static to allow dead code elim 
--   - arguably easier than uniquifying
-- o handle anon files properly (should be the easy case!)

at2 unitName = do
  putStrLn "Parsing"
  xs <- readDefns ["../units/oskit"] ["All.unit"]
  putStrLn "Evaluating"
  let u@(exps,cs,ds,is,fs,bs) = evalUnit xs unitName'
--  print $ vmap ppBinding' bs
  putStrLn "Ignoring constraints"
  putStrLn "Mangling C code"
  ((_,_,_,_,[(AnonCode f,_)]):bs) <- flattenBindings env exps bs
--  print $
  putStrLn "Writing C code"
  writeFile "bar.c" $ unlines f
  putStrLn "Done"
  return ()
 where
  unitName' = mkIdent unitName
  env = [ ("OSKITDIR", "/z/reid/oskit") 
        , ("BUILDDIR", "/z/reid/knit_oskit") 
        ] 

----------------------------------------------------------------
-- End 
----------------------------------------------------------------
