-- Copyright (c) 2000, 2001 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

----------------------------------------------------------------
-- This module defines the following common monads:
-- 
--   SR     - state reader monad
--   State  - (strict) state transformer monad
--   IOS    - IO monad with state
--   Output - output monad
--   CPS    - continuation passing monad
--
-- Most of these monads can be found in Wadler's papers about monads.
----------------------------------------------------------------

module Monads(
	-- exports a Monad and Functor instance for each of these types:
        SR,     runSR,  fixSR,  getSR,                 
	State,  runS,   fixS,   getS,   setS,   modS,
	IOS,    runIOS, fixIOS, getIOS, setIOS, modIOS, liftIOS,
	Output, runO,   fixO,   outO,
	RO,     runRO,  fixRO,  outRO, withRO, withROEx, getRO, redirectRO,
	CPS,    runCPS, callcc, getcc
	) where

import IOExts( fixIO )

----------------------------------------------------------------
-- For each type defined in this module we provide:
--
-- o A Functor instance
-- o A Monad   instance
-- o A function 
--
--     run<M> :: <M> a -> T a
--
--   which executes a computation of type M. This function usually takes
--   extra parameters (eg an input state) and returns the value of the
--   computation and some output values (eg an output state)
--
-- o State monads (ie SR, State and IOS) also provide:
--
--     get<M> :: <M> s
--     set<M> :: s -> <M> ()        -- not SR
--     mod<M> :: (s -> s) -> <M> () -- not SR
--
--   which get the current state, set the state to a new value and apply
--   a function to the state, respectively.
--
-- o Some monads also provide a fixpoint operator:
--
--     fix<M> :: (a -> <M> a) -> <M> a
--
-- o and state reader monads may provide a way to temporarily
--   change the state
--   
--     with<M> :: r -> <M> r a -> <M> r' a
-- 
----------------------------------------------------------------


----------------------------------------------------------------
-- The state reader monad
----------------------------------------------------------------

newtype SR s a = SR (s -> a)

runSR  :: SR s a -> s -> a
getSR  :: SR s s
fixSR  :: (a -> SR s a) -> SR s a
	   
instance Functor (SR s) where fmap f xs = do { x <- xs; return (f x) }

instance Monad (SR s) where
  m >>= k  = SR (\s -> case (unSR m) s of a -> (unSR (k a)) s)
  m >>  k  = SR (\s -> case (unSR m) s of _ -> (unSR k) s)
  return a = SR (\s -> a)

runSR      = unSR 
getSR      = SR (\s -> s)
fixSR m    = SR (\s -> let a = unSR (m a) s in a)

-- left inverse of S (not exported)
unSR :: SR s a -> (s -> a)
unSR (SR m) = m

----------------------------------------------------------------
-- A strict state monad
----------------------------------------------------------------

newtype State s a = S (s -> (a,s))

runS    :: State s a -> s -> (a,s)
fixS    :: (a -> State s a) -> State s a
modS    :: (s -> s) -> State s ()
setS    :: s -> State s ()
getS    :: State s s
	   
instance Functor (State s) where fmap f xs = do { x <- xs; return (f x) }

instance Monad (State s) where
  m >>= k  = S (\s -> case (unS m) s of (a,s') -> (unS (k a)) s')
  m >>  k  = S (\s -> case (unS m) s of (_,s') -> (unS k) s')
  return a = S (\s -> (a,s))

runS        = unS 
fixS m      = S (\s -> let r = unS (m (fst r)) s in r)
modS f      = S (\s -> ((),f s))
setS s'     = S (\s -> ((),s'))  
getS        = S (\s -> (s,s))

-- left inverse of S (not exported)
unS :: State s a -> (s -> (a,s))
unS (S m) = m

----------------------------------------------------------------
-- A standard IO + state monad
----------------------------------------------------------------

newtype IOS s a = IOS (s -> IO (a, s))

runIOS  :: IOS s a -> s -> IO (a, s)
fixIOS  :: (a -> IOS s a) -> IOS s a
modIOS  :: (s -> s) -> IOS s s
setIOS  :: s -> IOS s ()
getIOS  :: IOS s s
liftIOS :: IO a -> IOS s a

instance Functor (IOS s) where fmap f xs = do { x <- xs; return (f x) }

instance Monad   (IOS s) where
 m >>= k  = IOS (\s -> unIOS m s >>= \(a,s') -> unIOS (k a) s')
 m >>  k  = IOS (\s -> unIOS m s >>= \(_,s') -> unIOS k s')
 return a = IOS (\s -> return (a,s))

runIOS        = unIOS 
fixIOS m      = IOS (\s -> fixIO (\r -> unIOS (m (fst r)) s))
modIOS f      = IOS (\s -> return (s, f s))
setIOS s'     = IOS (\s -> return ((),s'))
getIOS        = IOS (\s -> return (s,s))
liftIOS m     = IOS (\s -> m >>= \ a -> return (a,s))

-- left inverse of IOS (not exported)
unIOS :: IOS s a -> (s -> IO (a, s))
unIOS (IOS m) = m

----------------------------------------------------------------
-- An "output" monad - like that in Wadler's "Essence of Functional
-- Programming".
----------------------------------------------------------------

newtype Output s a = O (a, [s] -> [s])

runO    :: Output s a -> (a, [s])
outO    :: [s] -> Output s ()
fixO    :: (a -> Output s a) -> Output s a

instance Functor (Output s) where fmap f xs = do { x <- xs; return (f x) }

instance Monad (Output s) where
  m >>= k  = O (let (a,r) = unO m; (b,s) = unO (k a) in (b, r . s))
  m >>  k  = O (let (a,r) = unO m; (b,s) = unO k     in (b, r . s))
  return a = O (a, \s -> s)

runO (O (a,s)) = (a, s [])
outO s         = O ((), (s ++))
fixO m         = r where r@(O (a,s)) = m a

-- left inverse of O (not exported)
unO :: Output s a -> (a, [s] -> [s])
unO (O m) = m


----------------------------------------------------------------
-- An "output" monad - combined with a state reader
----------------------------------------------------------------

newtype RO r o a = RO (r -> (a, [o] -> [o]))

runRO   :: RO r o a -> r -> (a, [o])
getRO   :: RO r o r
withRO  :: r' -> RO r' o a -> RO r o a
withROEx:: (r -> r') -> RO r' o a -> RO r o a
outRO   :: [o] -> RO r o ()
fixRO   :: (a -> RO r o a) -> RO r o a
redirectRO :: RO r o a -> RO r o (a, [o])

instance Functor (RO r o) where fmap f xs = do { x <- xs; return (f x) }

instance Monad (RO r o) where
  m >>= k  = RO (\ r -> let (a,o1) = unRO m r; (b,o2) = unRO (k a) r in (b, o1 . o2))
  m >>  k  = RO (\ r -> let (a,o1) = unRO m r; (b,o2) = unRO k     r in (b, o1 . o2))
  return a = RO (\ r -> (a, id))

runRO f    = \ r -> case unRO f r of (a,o) -> (a, o [])
getRO      = RO (\r -> (r,  id))
withRO r m = RO (\_ -> unRO m r)
withROEx f m = RO (\r -> unRO m (f r))
outRO  o   = RO (\r -> ((), (o ++)))
fixRO  m   = RO (\r -> let x = unRO (m (fst x)) r in x)
redirectRO m   = RO (\r -> case unRO m r of (a,o) -> ((a, o []), id))

-- left inverse of RO (not exported)
unRO :: RO r o a -> (r -> (a, [o] -> [o]))
unRO (RO m) = m


----------------------------------------------------------------
-- A CPS monad - parameterised over the type of the continuation
----------------------------------------------------------------

newtype CPS r a = CPS ((a -> r) -> r)

runCPS :: CPS r a -> (a -> r) -> r
callcc :: r -> CPS r a
getcc  :: ((a -> r) -> CPS r a) -> CPS r a

instance Functor (CPS s) where fmap f xs = do { x <- xs; return (f x) }

instance Monad (CPS r) where
  m >>= k  = CPS (\c -> unCPS m (\a -> unCPS (k a) c)) 
  m >>  k  = CPS (\c -> unCPS m (\a -> unCPS k     c))
  return a = CPS (\c -> c a) 

runCPS        = unCPS 
callcc cc     = CPS (\c -> cc)
getcc m       = CPS (\c -> unCPS (m c) c)

unCPS :: CPS r a -> ((a -> r) -> r)
unCPS (CPS m) = m

----------------------------------------------------------------
-- End
----------------------------------------------------------------
