-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module PP1
	( module PP1
	, module AbstractC
	) where

import PrettyUtils
import AbstractC

--import Trace

----------------------------------------------------------------
-- Pretty printing 
----------------------------------------------------------------

ppCompound :: Compound -> Doc
ppCompound (ls,ds,ss) 
  =  braces'
       (  (if null ls then empty else text "__label__" <+> hlist ppIdent ls <> semi)
       $$ vmap ppInit ds 
       $$ vmap ppStatement ss
       )

-- if it's a compound statement, lay it out a la K&R
-- otherwise, just indent it

ppNested :: Statement -> Doc
ppNested (Compound c) = ppCompound c
ppNested s            = nest 8 (ppStatement s)

ppStatement :: Statement -> Doc
ppStatement (Return x) = text "return" <+> mb ppExpr x <> semi
ppStatement Break      = text "break" <> semi
ppStatement Continue   = text "continue" <> semi
ppStatement (Goto l)   = text "goto" <+> ppIdent l <> semi
ppStatement (If c t Empty) = text "if" <+> parens (ppExpr c) 
                         $$ ppNested t
ppStatement (If c t e) = text "if" <+> parens (ppExpr c) 
                         $$ ppNested t
			 $$ text "else"
                         $$ ppNested e
ppStatement (Switch e s) = text "switch" <+> parens (ppExpr e)
                         $$ ppStatement s
ppStatement (Label id s) = ppIdent id <> colon <+> ppStatement s
ppStatement (Case e s)   = text "case" <+> ppExpr e <> colon <+> ppStatement s
ppStatement (Default s)  = text "default" <> colon <+> ppStatement s
ppStatement (While e s) = text "while" <+> parens (ppExpr e) $$ ppNested s
ppStatement (Do s e)    = text "do" $$ ppNested s $$ text "while" <+> parens (ppExpr e) <> semi
ppStatement (For e1 e2 e3 s) 
  = text "for" 
  <> parens (   mb ppExpr e1 <> semi
            <+> mb ppExpr e2 <> semi
            <+> mb ppExpr e3
            ) 
  $$ ppNested s

ppStatement (Compound s) = ppCompound s
ppStatement (Expr e) = ppExpr e <> semi
ppStatement (Asm volatile asm a1 a2 a3)
  =   text "asm"
  <+> (if volatile then text "volatile" else empty)
  <+> parens (  text asm 
             $$ if ((null a1) && (null a2) && (null a3)) then empty else (colon <+> hlist ppAsmArg a1)
             $$ if ((null a2) && (null a3)) then empty else (colon <+> hlist ppAsmArg a2)
             $$ if (null a3) then empty else (colon <+> hlist ppAsmArg a3)
             )
  <> semi
ppStatement Empty = semi

ppAsmArg (s,e) = text s <+> mb (parens . ppExpr) e

ppExpr :: Expr -> Doc
ppExpr (Ident id) = qualident'C id
ppExpr (Constant c) = text c
ppExpr (String s) = text s
ppExpr (Constructor t i) = parens (ppType Nothing t) <> ppInitializer i
ppExpr (App f as) = ppExpr f <> parens (hlist ppExpr as)
ppExpr (Select id e) = ppExpr e <> char '.' <> ppIdent id
ppExpr (Arrow id e) = ppExpr e <> text "->" <> ppIdent id
ppExpr (Cast t e) = parens (ppType Nothing t) <> ppExpr e
ppExpr (SizeOf_E e) = text "sizeof" <+> ppExpr e                      
ppExpr (SizeOf_T t) = text "sizeof" <> parens (ppType Nothing t)
ppExpr (LabelAddr l) = text "&&" <> ppIdent l
ppExpr (Cond e1 e2 e3) = ppExpr e1 
                       <+> char '?' <+> ppExpr e2
                       <+> char ':' <+> ppExpr e3
ppExpr (Cond2 e1 e2  ) = ppExpr e1 
                       <+> text "?:" <+> ppExpr e2
ppExpr (UnOp "*" e)   = char '*' <> ppExpr e
ppExpr (UnOp "(_)" e) = parens (ppExpr e)
ppExpr (UnOp "--_" e) = text "--" <> ppExpr e
ppExpr (UnOp "_--" e) = ppExpr e  <> text "--"
ppExpr (UnOp "++_" e) = text "++" <> ppExpr e
ppExpr (UnOp "_++" e) = ppExpr e  <> text "++"
ppExpr (UnOp o e) = ppOp o <+> ppExpr e
ppExpr (BinOp "," e1 e2) = ppExpr e1 <> comma <+> ppExpr e2
ppExpr (BinOp "_[_]" e1 e2) = ppExpr e1 <> brackets (ppExpr e2)
ppExpr (BinOp o e1 e2) = ppExpr e1 <+> ppOp o <+> ppExpr e2
ppExpr (StatementExpr s) = parens (ppCompound s)

ppAttributes :: [Attribute] -> Doc
ppAttributes [] = empty
ppAttributes xs = text "__attribute__" <> parens (parens (hlist ppAttribute xs))

ppAttribute :: Attribute -> Doc
ppAttribute (Mode as)         = text "__mode__"   <> parens (hlist ppAttribute as)
ppAttribute (Format as)       = text "__format__" <> parens (hlist ppAttribute as)
ppAttribute (Attr_Constant x) = text x
ppAttribute (SimpleAttr a)    = text a
{-
ppAttribute Aligned           = text "__aligned__"
ppAttribute Unused            = text "__unused__"
ppAttribute Transparent_Union = text "__transparent_union__"
ppAttribute DI		      = text "__DI__"
ppAttribute Printf	      = text "__printf__"
ppAttribute Scanf	      = text "__scanf__"
ppAttribute NoReturn	      = text "__noreturn__"
ppAttribute Attr_Const	      = text "__const__"
ppAttribute StdCall           = text "__stdcall__"
ppAttribute COM               = text "__com__"
ppAttribute C                 = text "__c__"
ppAttribute CPU               = text "__cpu__"
-}

ppInitializer :: Initializer -> Doc
ppInitializer None = empty
ppInitializer (SimpleInit e) = ppExpr e
ppInitializer (FieldInit f e) = ppIdent f <> colon <+> ppInitializer e
ppInitializer (ArrayInit i e) = brackets (ppExpr i) <+> ppInitializer e
ppInitializer (ListInit is) = braces (hlist ppInitializer is)

ppInit :: Init -> Doc
ppInit (x, None) = ppDecl x <> semi
ppInit (x, y)  = ppDecl x <+> equals <+> ppInitializer y <> semi

ppDecl :: Decl -> Doc
ppDecl (x,t) = ppType3 t (qualident'C x) 

arglist :: Bool -> [Doc] -> Doc
arglist False as = hcat $ punctuate comma $ as
arglist True  as = hcat $ punctuate comma $ as ++ [text "..."]

isFunction (Function1 _ _ _ _) = True
isFunction (Function2 _ _ _ _) = True
isFunction _ = False

-- prints type in C syntax
ppType :: Maybe Ident -> Type -> Doc
ppType nm t = ppType3 t (mb ppIdent nm)
--ppType nm t = text (show (nm,t))

ppType3 :: Type -> Doc -> Doc
ppType3 (Function1 as var r attrs) d = ppType3 r (d <> parens (arglist var [ ppIdent a | a <- as ]) <+> ppAttributes attrs)
ppType3 (Function2 as var r attrs) d = ppType3 r (d <> parens (arglist var (map ppArg2' as)) <+> ppAttributes attrs)
ppType3 (SimpleTy t)         d = qualident'C t <+> d
ppType3 (TypeOf e)           d = text "typeof" <+> ppExpr e <+> d
ppType3 (Ptr t) d
  | isFunction t = ppType3 t (parens (char '*' <> d))
  | otherwise    = ppType3 t (char '*' <> d)
ppType3 (Array sz t)         d = ppType3 t (d <> brackets (mb ppExpr sz))
ppType3 (Const t)            d = text "const" <+> ppType3 t d
ppType3 (TyAttribute a t)    d = ppType3 t (ppAttributes a <+> d)
ppType3 (TyApp x y)          d = ppType3 x (ppType3 y d)
ppType3 (TyEnum id)          d = text "enum"   <+> qualident'C id <+> d
ppType3 (TyStruct id)        d = text "struct" <+> qualident'C id <+> d
ppType3 (TyUnion id)         d = text "union"  <+> qualident'C id <+> d
ppType3 (ParensType ty)      d = parens (ppType3 ty d)



ppDefn :: Defn -> Doc
ppDefn (Enum id xs)  = 
  (   text "enum"
  <+> qualident'C id
  $$  ppEnumerators xs
  <>  semi
  ) 
ppDefn (Struct id xs)  =
  (   text "struct"
  <+> qualident'C id
  $$  ppFields xs
  <>  semi
  )
ppDefn (Union id xs)  =
  (   text "union"
  <+> qualident'C id
  $$  ppFields xs
  <>  semi
  )
ppDefn (Typedef x t) = 
  (   text "typedef" 
  <+> ppType3 t (qualident'C x)
  <>  semi
  )
ppDefn (Extern d) =
      (   text "extern"
      <+> ppDefn d
      )
ppDefn (Fun d ds c) =
       (  ppDecl d
       $$ vmap ppInit (map (\ d -> (d, None)) ds)
       $$ ppCompound c
       )
ppDefn (Init x) =
     ( ppInit x
     )
ppDefn (TopAsm volatile asm a1 a2 a3) =
     ( ppStatement (Asm volatile asm a1 a2 a3)
     )

ppFields [] = empty
ppFields xs = braces' (vmap ppField xs)

ppField (nm, sz, ty) 
  = ppType nm ty <+> (mb ppSize sz) <> semi
ppSize sz = colon <+> ppExpr sz

ppEnumerators [] = empty
ppEnumerators xs = braces' (vcat (punctuate comma (map ppEnumerator xs)))

ppArg2' (x,y) = ppType x y

ppEnumerator (x, Nothing) = qualident'C	x
ppEnumerator (x, Just y)  = qualident'C	x <+> equals <+> ppExpr y

ppExtern :: Init -> Doc
ppExtern ((x,t),None) = text "extern" <+> ppType3 t (qualident'C x) <> semi

braces' d = lbrace $$ nest 8 (blank $$ d) $$ rbrace

---------------------------------------------------------------
-- Print Asm with renaming macros
---------------------------------------------------------------

ppAsm :: String -> [(Ident,QualId)] -> Doc
ppAsm asm renamings =
  (
  if (asm == "") then empty else (text "\nSTART-FILE" 
                                  $$ vmap startMacro renamings
                                  $$ text "START-ASM\n")
  <+> text asm
  $$ if (asm == "") then empty else text "\nEND-FILE\n"
  )
 where
  startMacro (x,y) = ppIdent x <+> qualident'C y
  endMacro   (x,y) = ppIdent x <+> qualident'C y

----------------------------------------------------------------
-- End
----------------------------------------------------------------
