-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module ParseC
	( parseDefns, parseDefnsEx
	) where

import AbstractC
import PP1
import SExpr
import ZEncode

import Prelude hiding(init)
import PrettyUtils
import System( system, ExitCode(..) )

import MonadUtils
	( MonadPlus(..)
	, guard, unless
	, liftM0, liftM1, liftM2, liftM3, liftM4 
	, concatMapM
	)

parseDefns :: FilePath -> IO [Defn]
parseDefns fn = parseDefnsEx "knit_c_parser" [] ["-I../includes", "-I../sigs", "-I../../z/build_oskit"] fn fn

parseDefnsEx :: String -> [String] -> [String] -> FilePath -> FilePath -> IO [Defn]
parseDefnsEx cparser env flags fn tgt_fn = do
  { my_system $ "rm -f " ++ xxx ++ " " ++ yyy
  ; rc <- my_system $ unwords $ "env " : env ++ [shell ("gcc" : ["-P","-E","$KNIT_CPPFLAGS"] ++ flags ++ [fn,"> " ++ yyy])]
  ; unless (rc == ExitSuccess) (ioError $ userError $ "Can't preprocess " ++ fn)
  ; rc <- my_system (cparser ++ " < " ++ yyy ++ " > " ++ xxx)
  ; unless (rc == ExitSuccess) (ioError $ userError $ "Can't parse " ++ fn)
  ; cs <- readFile xxx
  ; let (_,ys) = runM $ mapM' def' $ parse $ forceSpine $ tokenize cs
  ; seq (forceSpine ys) $ return ys
  }
 where
  xxx = (zEncode tgt_fn ++ "xxx")
  yyy = (zEncode tgt_fn ++ "yyy")

  shell xs = "sh -c '" ++ unwords xs ++ "'"

my_system xs = do
  -- putStrLn xs -- for debugging
  rc <- System.system ("trap '' 26; " ++ xs)
  unless (rc == ExitSuccess) $ putStrLn xs
  return rc
    
-- When processing many files at once, we want to force tokenizing
-- to finish before we start parsing and the like in the hope that
-- Hugs will not run out of file handles.
forceSpine :: [a] -> [a]
forceSpine xs = length xs `seq` xs

systemEx :: String -> [String] -> IO ExitCode
systemEx cmd args = my_system (unwords (cmd:args))
-- where system x = print x >> return ExitSuccess

----------------------------------------------------------------
-- The Parser monad
--
-- As we parse C declarations, we have to lift out declarations of
-- structs, unions, and enums and replace them with references
-- to the declarations.  This monad provides labels (with which
-- to label the structs) and collects the declarations in an ordered
-- list.
--
-- To help debugging, we add in a read-only "context" field which
-- can be printed to generate better error messages.
----------------------------------------------------------------

newtype M a = M (Doc -> [Ident] -> [Defn] -> (a,[Ident],[Defn]))

unM :: M a -> (Doc -> [Ident] -> [Defn] -> (a,[Ident],[Defn]))
unM (M xs) = xs

runM :: M a -> (a,[Defn])
runM xs = case unM xs empty labs [] of (a,_,ss) -> (a, reverse ss)
 where
  labs = [ mkIdent $ "__t" ++ show i | i <- [0..] ]

nestM :: M a -> M (a,[Defn])
nestM xs = M (\ ctx ls ss -> case unM xs ctx ls [] of 
                         (a,ls',ss') -> ((a, reverse ss'),ls',ss))

newTag :: M Ident
newTag = M (\ ctx (l:ls) ss -> (l,ls,ss))

addDefn :: Defn -> M ()
addDefn s = M (\ ctx ls ss -> ((),ls,s:ss))

instance Functor M where
  fmap f xs = do { x <- xs; return (f x) }

instance Monad M where
  m >>= k = M (\ ctx ls ss -> case unM m ctx ls ss of
                          (a,ls',ss') -> unM (k a) ctx ls' ss')
  m >>  k = M (\ ctx ls ss -> case unM m ctx ls ss of
                          (_,ls',ss') -> unM k     ctx ls' ss')
  return x = M (\ ctx ls ss -> (x,ls,ss))
  fail s = M (\ ctx ls ss -> error ("fail_M: " ++ s ++ " in context:\n" ++ show ctx))

withContextEx :: (Doc -> Doc) -> (M a -> M a)
withContextEx f m = M (\ ctx ls ss -> unM m (f ctx) ls ss)

getContext :: M Doc
getContext = M (\ ctx ls ss -> (ctx,ls,ss))

withContext :: Doc -> (M a -> M a)
withContext d m = withContextEx (f d) m
 where
  f d old = d $$ (text "in: " <> old)

----------------------------------------------------------------
-- Declarations
----------------------------------------------------------------

mbq :: SExpr -> M QualId
mbq Nil = do { t <- newTag; return (toQualId (mkQual [t])) }
mbq x   = return (qualident x)


ident :: SExpr -> Ident
ident (Atom x) = mkIdent x
ident e = error $ "ParseC.ident: " ++ show e

qualident :: SExpr -> QualId
qualident = toQualId . mkQual . qualifier

qualifier :: SExpr -> [Ident]
qualifier (Thing "::" (Cons x y)) = ident x : qualifier y
qualifier (Atom x)                = [mkIdent x]




def' :: SExpr -> M ()
def     :: SExpr -> M [Defn]
decl'   :: Type  -> SExpr -> M Decl
abstract' :: Type -> SExpr -> M (Maybe QualId, Type)
abstract :: SExpr -> M [(Maybe QualId, Type)]

def' x = do
  { ds <- def x
  ; mapM_ addDefn ds
  }

def x = -- withContext (text "def " <> pp x) $
        def2 x

def2 (Cons (Cons (Atom "typedef") lhs) rhs) = do
  { t <- tySpec lhs
  ; mapM' (tydef t) (toList rhs)
  }

def2 (Cons (Cons (Atom "extern") lhs) rhs) = do
  { (d,ds) <- nestM (def2 (Cons lhs rhs))
  ; return (ds ++ map Extern d)
  }

def2 (Thing "fun" (Cons lhs (Cons rhs (Cons ds (Thing "{}" body))))) = do
  { t   <- tySpec lhs
  ; d   <- decl' t rhs
  ; ds' <- concatMapM' init (toList ds)
  ; c   <- -- withContext (text "funbody " <> pp body) $
           compound body
  ; return [Fun d (map fst ds') c]
  }

def2 x@(Thing "asm" v) = do
  { (Asm a b c d e) <- statement' x
  ; return [TopAsm a b c d e]
  }

def2 x@(Cons rty f) = do
  { x' <- init x
  ; return (map Init x')
  }


--def x = text "DEF" <+> pp x

tydef :: Type -> SExpr -> M Defn
tydef t e = do
  { ((x,t'),None) <- init' t e
  ; return (Typedef x t')
  }

decls :: SExpr -> M [Decl]
decls (Cons rty xs) = do 
  { rty' <- tySpec rty
  ; mapM' (decl' rty') (toList xs)
  }

isVarArgs :: SExpr -> Bool
isVarArgs Nil = False
isVarArgs _   = True

decl' rty (Atom x) = return (mkQualId x, rty)
decl' rty x@(Thing "::" _) = return (qualident x, rty)
-- decl' rty (S [A "fun1", S (A fn:xs)]) = (fn, Function1 ... rty)

decl' rty (Thing "fun" (Cons f (Cons (Thing "pidlist" (Cons xs ellipsis)) attrs))) = do
  { let xs' = map ident (toList xs)
  ; attrs' <- attributes attrs
  ; decl' (Function1 xs' (isVarArgs ellipsis) rty attrs') f 
  }
decl' rty (Thing "fun" (Cons f (Cons (Thing "ptylist" (Cons xs ellipsis)) attrs))) = do
  { xs' <- concatMapM' abstract2 (toList xs)
  ; attrs' <- attributes attrs
  ; decl' (Function2 xs' (isVarArgs ellipsis) rty attrs') f 
  }
decl' rty (Thing "ptr" (Cons p x)) = do
  { p' <- pointer rty p
  ; decl' p' x
  }
decl' rty (Thing "[]" (Cons x sz)) = do
  { sz' <- mbe sz
  ; decl' (Array sz' rty) x
  }
decl' rty (Thing "attrs" (Cons attrs y)) = do
  { attrs' <- attributes attrs
  ; decl' (TyAttribute attrs' rty) y
  }
decl' rty x = error $ "decl' " ++ show x


-- ToDo: why does this always return a singleton list?
-- abstract declarator?
--abstract x = [(Nothing, SimpleTy $ show (braces $ pp x))]

abstract (Thing "x" (Cons rtys Nil))  = do
  { rtys' <- tySpecs rtys
  ; return [(Nothing, rtys')]
  }
abstract (Thing "x" (Cons rtys xs))   = do
  { rtys' <- tySpecs rtys
  ; (x,y) <- decl' rtys' xs
  ; return [(Just x, y)]
  }
abstract (Thing "y" (Cons x y))       = do
  { x' <- tySpecs x
  ; y' <- abstract' x' y
  ; return [y']
  }

abstract2 :: SExpr -> M [(Maybe Ident, Type)]
abstract2 e = do
  { xts <- abstract e
  ; return [ (fmap unQualify x, t) | (x,t) <- xts ]
  }

-- Sigh - almost identical to decl'
abstract' rty Nil      = return (Nothing, rty)
abstract' rty (Atom x) = return (Just (mkQualId x), rty)
abstract' rty x@(Thing "::" _) = return (Just (qualident x), rty)
abstract' rty (Thing "[]" (Cons x sz)) = do
  { sz' <- mbe sz
  ; abstract' (Array sz' rty) x
  }
abstract' rty (Thing "fun" (Cons f (Cons (Cons xs ellipsis) attrs))) = do
  { xs' <- concatMapM' abstract2 (toList xs)
  ; attrs' <- attributes attrs
  ; abstract' (Function2 xs' (isVarArgs ellipsis) rty attrs') f
  }
abstract' rty (Thing "(fun')" (Cons f (Cons xs ellipsis))) = do
  { xs'        <- concatMapM' abstract2 (toList xs)
  ; abstract' (Function2 xs' (isVarArgs ellipsis) rty []) f
  }
abstract' rty (Thing "abstract" (Cons y z)) = do
  { p <- pointer rty y
  ; abstract' p z
  }
abstract' rty x = error $ "abstract' " ++ show (ppType Nothing rty, x)

pointer :: Type -> SExpr -> M Type
pointer rty Nil             = return rty
pointer rty (Atom "*1")     = return (Ptr rty)
pointer rty (Thing "*2" ts) = tySpecs' (Ptr rty) ts
pointer rty (Thing "*3" ts) = pointer (Ptr rty) ts
pointer rty (Thing "*4" (Cons ts p)) = do
  { rty' <- tySpecs' (Ptr rty) ts
  ; pointer rty' p
  }

tySpec' :: SExpr -> M Type
tySpec' (Cons (Atom "const") x) = do{ ty <- tySpec' x; return (Const ty) }
tySpec' (Cons x y) = do
  { x' <- tySpecs x
  ; (Nothing, y') <- abstract' x' y
  ; return y'
  }

-- if there's no id here, make one up (nombid = no-maybe identifier)
nombid :: SExpr -> M Ident
nombid Nil = newTag
nombid (Atom x) = return $! mkIdent x

tySpec :: SExpr -> M Type

tySpec (Atom x) = liftM0 (SimpleTy (mkQualId x))

tySpec (Cons x (Thing "abstract" (Cons y z))) = do
  { t0 <- tySpec x
  ; p <- pointer t0 y
  ; (_, t) <- (abstract' p z)
  ; return t
  }

tySpec (Thing "typeof" e) = liftM1 TypeOf (expr e)

tySpec (Thing "enum"   (Cons nm xs)) = do
  { nm' <- mbq nm
  ; xs' <- mapM' enumerator (toList xs)
--  ; unless (null xs') $ addDefn (Enum nm' xs')
  ; addDefn (Enum nm' xs')
  ; return (TyEnum nm')
  }
tySpec (Cons (Atom "union") (Cons nm xs))  = do
  { nm' <- mbq nm
  ; xs' <- concatMapM' field (toList xs)
--  ; unless (null xs') $ addDefn (Union nm' xs')
  ; addDefn (Union nm' xs')
  ; return (TyUnion nm')
  }
tySpec (Cons (Atom "struct") (Cons nm xs)) = do
  { nm' <- mbq nm
  ; xs' <- concatMapM' field (toList xs)
-- no matter what, we declare the struct because gcc issues a warning if it
-- sees a struct it's never heard of before
--  ; unless (null xs') $ addDefn (Struct nm' xs')
  ; addDefn (Struct nm' xs')
  ; return (TyStruct nm')
  }
tySpec (Cons x Nil) = tySpec x
--tySpec (Cons (Atom "const") y) = liftM1 Const (tySpec y)
tySpec (Cons x y)   = liftM2 TyApp (tySpec x) (tySpec y)
tySpec (Thing "attrs" (Cons x as)) = liftM2 TyAttribute (attributes as) (tySpec x)
--tySpec x = SimpleTy $ "tyspec" ++ (show (pp x)) -- hack
tySpec Nil          = liftM0 intType -- default - blech!
tySpec x            = fail $ "tySpec " ++ show x

tySpecs :: SExpr -> M Type
--tySpecs x = SimpleTy $ show (braces $ pp x)
tySpecs (Cons (Atom "const") x) = do{ ty <- tySpec x; return (Const ty) }
tySpecs (Cons x y) = do
  { x' <- tySpec x
  ; tySpecs' x' y
  }

tySpecs' :: Type -> SExpr -> M Type
tySpecs' ty Nil                     = liftM0 ty
tySpecs' ty (Atom x)                = liftM0 (TyApp ty (SimpleTy (mkQualId x)))
--tySpecs' ty (Cons (Atom "const") x) = tySpecs' (Const ty) x  -- no
tySpecs' ty (Cons x y)              = do
  { x' <- tySpecs' ty x
  ; tySpecs' x' y
  }
-- tySpecs' ty (Cons (Atom "*") x)  = tySpecs' (Ptr ty) x
-- tySpecs' ty (Cons (Cons (Atom "*") Nil) x) = tySpecs' (Ptr ty) x
--tySpecs' ty x = SimpleTy $ (show (ppType Nothing ty, pp x)) -- hack

----------------------------------------------------------------
-- Misc
----------------------------------------------------------------

init :: SExpr -> M [Init]
init (Cons rty xs) = do
  { rty' <- tySpec rty
  ; mapM' (init' rty') (toList xs)
  }

init' :: Type -> SExpr -> M Init
init' rty (Cons x y)   = liftM2 (,) (decl' rty x) (initializer y)

initializer :: SExpr -> M Initializer
initializer Nil             = liftM0 None
initializer (Thing "{_}" y) = liftM1 ListInit   (mapM' initializer (toList y))
initializer (Thing ":=" (Cons (Thing "var" x) y)) = liftM1 (FieldInit (ident x)) (initializer y)
initializer (Thing "[]=" (Cons x y)) = liftM2 ArrayInit (expr x) (initializer y)
initializer e               = liftM1 SimpleInit (expr e)

mbid :: SExpr -> Maybe Ident
mbid Nil      = Nothing
mbid (Atom x) = Just $! mkIdent x

enumerator :: SExpr -> M (QualId, Maybe Expr)
enumerator (Cons x y) = liftM2 (,) (liftM0 (qualident x)) (mbe y)

decl2 :: Type -> SExpr -> M (Ident, Type)
decl2 t e = do
  { (x,t) <- decl' t e
  ; return (unQualify x, t)
  }

field :: SExpr -> M [(Maybe Ident, Maybe Expr, Type)]
field (Cons ty (Cons as ds)) = do
  { ty' <- tySpec ty
  ; mapM'  (field' ty') (toList ds)
  }
 where
  field' t (Cons Nil sz) = do
    { e <- mbe sz
    ; return (Nothing, e, t)
    }
  field' t (Cons d sz)   = do
    { (d',ty') <- decl2 t d
    ; e <- mbe sz
    ; return (Just d', e, ty')
    }


----------------------------------------------------------------
-- Attributes
----------------------------------------------------------------

attributes :: SExpr -> M [Attribute]
attributes = mapM' attribute . toList

attribute :: SExpr -> M Attribute
attribute (Thing "app" (Cons (Atom "mode") as))   = liftM1 Mode (attributes as)
attribute (Thing "app" (Cons (Atom "__mode__") as))   = liftM1 Mode (attributes as)
attribute (Thing "app" (Cons (Atom "format") as)) = liftM1 Format (attributes as)
attribute (Thing "app" (Cons (Atom "__format__") as)) = liftM1 Format (attributes as)
attribute (Thing "const" (Atom x))   = liftM0 (Attr_Constant x)
attribute (Atom x)                   = liftM0 (SimpleAttr x)
attribute x                          = error $ "ParseC.attribute: " ++ show x
{-
attribute (Atom "aligned")           = liftM0 Aligned
attribute (Atom "unused")            = liftM0 Unused
attribute (Atom "transparent_union") = liftM0 Transparent_Union
attribute (Atom "di")                = liftM0 DI
attribute (Atom "printf")            = liftM0 Printf
attribute (Atom "scanf")             = liftM0 Scanf
attribute (Atom "noreturn")          = liftM0 NoReturn
attribute (Atom "const")             = liftM0 Attr_Const
attribute (Atom "stdcall")           = liftM0 StdCall
-}

----------------------------------------------------------------
-- Statements
----------------------------------------------------------------

compound :: SExpr -> M Compound
compound (Cons w (Cons x y)) = do
  { ls <- mapM' label' (toList w)
  ; ds <- concatMapM' init (toList x)
  ; ss <- -- withContext (text "statements " <> pp y) $
          mapM' statement' (toList y)
  ; ls `seq` ds `seq` ss `seq` return (ls,ds,ss)
  }

label' :: SExpr -> M Label
label' (Atom x) = return $! mkIdent x

statement :: SExpr -> M Statement
statement x = withContext (text "statement " <> pp x) $
              statement' x

statement' :: SExpr -> M Statement

statement' (Thing "return" e)                 = liftM1 Return (mbe e)
statement' (Atom "break")                     = liftM0 Break
statement' (Atom "continue")                  = liftM0 Continue
statement' (Thing "goto" (Atom l))            = liftM0 (Goto (mkIdent l))
statement' (Atom "empty")                     = liftM0 Empty
statement' (Thing "if" (Cons c (Cons t e)))   = liftM3 If (expr c) (statement t) (statement e)
statement' (Thing "switch" (Cons e s))        = liftM2 Switch (expr e) (statement s)
statement' (Thing "label" (Cons (Atom id) s)) = liftM1 (Label (mkIdent id)) (statement s)
statement' (Thing "case" (Cons e s))          = liftM2 Case (expr e) (statement s)
statement' (Thing "default" s)                = liftM1 Default (statement s)
statement' (Thing "while" (Cons c s))         = liftM2 While (expr c) (statement s)
statement' (Thing "do" (Cons s c))            = liftM2 Do (statement s) (expr c)
statement' (Thing "for" (Cons (Cons e1 e2) (Cons e3 s))) 
         				      = liftM4 For (mbe e1) (mbe e2) (mbe e3) (statement s)
statement' (Thing "{}" s)                     = liftM1 Compound (compound s)
statement' (Thing "asm" (Cons Nil (Cons (Cons (Atom op) a1) (Cons a2 a3)))) 
  = liftM3 (Asm False op) (mapM' asm_arg (toList a1)) (mapM' asm_arg (toList a2)) (mapM' asm_arg (toList a3))
statement' (Thing "asm" (Cons _   (Cons (Cons (Atom op) a1) (Cons a2 a3)))) 
  = liftM3 (Asm True op) (mapM' asm_arg (toList a1)) (mapM' asm_arg (toList a2)) (mapM' asm_arg (toList a3))
statement' (Thing "expr" x)                   = liftM1 Expr (expr x)

asm_arg (Cons (Atom x) y) = do
  { y' <- mbe y
  ; return (x, y')
  }

----------------------------------------------------------------
-- Expr
----------------------------------------------------------------

expr :: SExpr -> M Expr

expr (Thing "var" v)                   = q `seq` liftM0 (Ident q) where q = qualident v
expr (Thing "const" (Atom c))          = liftM0 (Constant c)
expr (Thing "string" (Atom s))         = liftM0 (String s)
expr (Thing "app" (Cons f as))         = liftM2 App (expr f) (mapM' expr (toList as))
expr (Thing "index" (Cons x i))        = liftM2 (BinOp "_[_]") (expr x) (expr i)
expr (Thing "++_" x)  		       = liftM1 (UnOp "++_") (expr x)
expr (Thing "_++" x)  		       = liftM1 (UnOp "_++") (expr x)
expr (Thing "--_" x)  		       = liftM1 (UnOp "--_") (expr x)
expr (Thing "_--" x)  		       = liftM1 (UnOp "_--") (expr x)
expr (Thing "," (Cons x y)) 	       = liftM2 (BinOp ",") (expr x) (expr y)
expr (Thing "cast" (Cons x y))         = liftM2 Cast (tySpec' x) (expr y)
expr (Thing "constr" (Cons x y))       = liftM2 Constructor (tySpec' x) 
                                                (liftM1 ListInit   (mapM' initializer (toList y)))
expr (Thing "unop" (Cons (Atom op) x)) = liftM1 (UnOp op) (expr x)
expr (Thing "->" (Cons x (Atom y)))    = y `seq` liftM1 (Arrow  y')  (expr x) where y' = mkIdent y
expr (Thing "."  (Cons x (Atom y)))    = y `seq` liftM1 (Select y') (expr x) where y' = mkIdent y
expr (Thing "*"  (Cons x y))           = liftM2 (BinOp "*")  (expr x) (expr y)
expr (Thing "/"  (Cons x y))           = liftM2 (BinOp "/")  (expr x) (expr y)
expr (Thing "%"  (Cons x y))           = liftM2 (BinOp "%")  (expr x) (expr y)
expr (Thing "+"  (Cons x y))           = liftM2 (BinOp "+")  (expr x) (expr y)
expr (Thing "-"  (Cons x y))           = liftM2 (BinOp "-")  (expr x) (expr y)
expr (Thing "<<" (Cons x y))           = liftM2 (BinOp "<<") (expr x) (expr y)
expr (Thing ">>" (Cons x y))           = liftM2 (BinOp ">>") (expr x) (expr y)
expr (Thing "<"  (Cons x y))           = liftM2 (BinOp "<")  (expr x) (expr y)
expr (Thing ">"  (Cons x y))           = liftM2 (BinOp ">")  (expr x) (expr y)
expr (Thing "<=" (Cons x y))           = liftM2 (BinOp "<=") (expr x) (expr y)
expr (Thing ">=" (Cons x y))           = liftM2 (BinOp ">=") (expr x) (expr y)
expr (Thing "==" (Cons x y))           = liftM2 (BinOp "==") (expr x) (expr y)
expr (Thing "!=" (Cons x y))           = liftM2 (BinOp "!=") (expr x) (expr y)
expr (Thing "&"  (Cons x y))           = liftM2 (BinOp "&")  (expr x) (expr y)
expr (Thing "|"  (Cons x y))           = liftM2 (BinOp "|")  (expr x) (expr y)
expr (Thing "^"  (Cons x y))           = liftM2 (BinOp "^")  (expr x) (expr y)
expr (Thing "&&" (Cons x y))           = liftM2 (BinOp "&&") (expr x) (expr y)
expr (Thing "||" (Cons x y))           = liftM2 (BinOp "||") (expr x) (expr y)
expr (Thing "?:"  (Cons x (Cons y z))) = liftM3 Cond         (expr x) (expr y) (expr z)
expr (Thing "?nil:"  (Cons x y))       = liftM2 Cond2        (expr x) (expr y)
expr (Thing "="  (Cons x (Cons (Atom y) z))) = liftM2 (BinOp y) (expr x) (expr z)
expr (Thing "sizeof_t" ty)             = liftM1 SizeOf_T (tySpec ty)
expr (Thing "sizeof" e)                = liftM1 SizeOf_E (expr e)
expr (Thing "label_addr" (Atom l))     = liftM0 (LabelAddr (mkIdent l))
expr (Thing "(_)" e)                   = liftM1 parensExpr (expr e)
expr (Thing "stat_expr" (Thing "{}" c))= liftM1 StatementExpr (compound c)
expr e = error $ "ParseC.expr: " ++ show e

mbe :: SExpr -> M (Maybe Expr)
mbe Nil = liftM0 Nothing
mbe x   = liftM1 Just (expr x)

----------------------------------------------------------------
-- Utils
----------------------------------------------------------------

mapM' :: Monad m => (a -> m b) -> [a] -> m [b]
mapM' f xs = do
  ys <- mapM f xs
  seqList seq ys $ return ys

concatMapM' :: (Functor m, Monad m) => (a -> m [b]) -> [a] -> m [b]
concatMapM' f xs = do
  ys <- concatMapM f xs
  seqList seq ys $ return ys

----------------------------------------------------------------
-- Forcing functions
----------------------------------------------------------------

type Seq a k = a -> k -> k

seqMb :: Seq a k -> Seq (Maybe a) k
seqMb sq Nothing  c = c
seqMb sq (Just x) c = sq x c

seqList :: Seq a k -> Seq [a] k
seqList sq [] c = c
seqList sq (x:xs) c = sq x (seqList sq xs c)

seqString :: Seq String k
seqString = seqList seq

noseq :: Seq a k
noseq _ k = k

seq2 :: Seq a k -> Seq b k -> Seq (a,b) k
seq2 s1 s2 (x,y) k = s1 x (s2 y k)

----------------------------------------------------------------
-- Tests
----------------------------------------------------------------

t3 fn = do
  { cs <- readFile fn
  ; let (_,ys) = runM $ mapM def' $ parse $ tokenize cs
  ; print $ vmap ppDefn ys
  }

t4 fn = do
  { cs <- readFile fn
  ; let (_,ys) = runM $ mapM def' $ parse $ tokenize cs
  ; writeFile "x.c" $ show $ vmap ppDefn ys
  }

t5 fn = do
  { ys <- parseDefns fn
  ; print $ vmap ppDefn ys
--  ; writeFile "x.c" $ show $ vmap ppDefn ys
  }

