-- Copyright (c) 2000, 2001 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

----------------------------------------------------------------
-- The simplified units implementation
--
-- (This is the parser)
----------------------------------------------------------------

module ParseUnits
	( readDefns
	) where

import ParserUtils
        ( Parser, runP, tok, token, errorP, force
        , (|||)
        , optional, oneOf
        , zeroOrMore, oneOrMore
        , sepdBy, sepdBy1
        , sloppy_sepdBy, sloppy_sepdBy1
        )
import ListUtils( chopList )
import Char(isAlphaNum)
import Maybe( fromMaybe, isJust )
import List( nub, findIndex )
import IO( try )
import Monad
import Char( isSpace, isAlpha, isDigit, lexLitChar )
import AbstractUnits
import qualified PPUnits as PP -- for testing purposes
import PrettyUtils(empty)
import qualified PrettyUtils as P
import System(exitFailure)     -- for my_error
import IOExts(unsafePerformIO) -- for my_error
--import Trace

----------------------------------------------------------------
-- The main export
----------------------------------------------------------------

readDefns :: [FilePath] -> [FileName] -> IO [(Ident,Defn)]
readDefns path fs = do { xss <- readDefns' [] fs; return (concat xss) }
 where
  readDefns' :: [FileName] -> [FileName] -> IO [[(Ident,Defn)]]
  readDefns' seen [] = return []
  readDefns' seen (f:fs)
    | notDir f `elem` seen
    = readDefns' seen fs
    | otherwise
    = do
        (f',s) <- readFileOnPath path f
        let (fs',xs) = parseDefns' f' s
        xss <- readDefns' (notDir f:seen) (fs++fs')
        return (xs:xss)
    
-- Try every filename-path combination until one works.
-- Return the filename and the contents
readFileOnPath :: [FilePath] -> FileName -> IO (FilePath,String)
readFileOnPath path f = do
  let fs = case f of
           -- If `f' is an absolute pathname, look for `f' only.
           ('/':_) -> [f]
           -- Otherwise, expand `f' by prepending every dir in `path'.
           _       -> [ dir ++ "/" ++ f | dir <- path ]
  ss <- mapM (\ filename -> try (readFile' filename)) fs
  case [ s | Right s <- ss ] of
   []  -> my_error (Just (f,-1)) $ "Unable to find file '" ++ f ++ "' in path " ++ show path
   [s] -> return s
   _   -> my_error (Just (f,-1)) $ "Ambiguous file '" ++ f ++ "' in path " ++ show path
 where
  readFile' f = do{ s <- readFile f; return (f,s) }

parseDefns' :: FileName -> String -> ([FileName],[(Ident,Defn)])
parseDefns' fn s = ( [ x | Left x <- xys ], [ y | Right y <- xys ] )
 where
  xys = run defns (tokenise' fn s)

notDir :: FileName -> FileName
notDir [] = []
notDir fn = last (chopList (== '/') fn)

----------------------------------------------------------------
-- Parsing units, bundle types, etc
----------------------------------------------------------------

defns         :: P [Either FileName (Ident,Defn)]
include       :: P FileName
defn          :: P (Ident,Defn)   
unit          :: P Unit        
modunit       :: P ModUnit        
addunit       :: P AddUnit        
remunit       :: P RemUnit        
import'       :: P Import 
export'       :: P Export   
bundleType    :: P BundleType
btElement     :: P BTElement
constraint    :: P Constraint
property      :: P Property
clause        :: P Clause
formula       :: P Formula
atomic        :: P Atomic
aProp         :: P AProp
flags         :: P Flags        
binding       :: P Binding     
unitInstance  :: P UnitInstance         
bundle        :: P Bundle
bundleElement :: P BundleElement
depend        :: P Depend
objectSet     :: P ObjectSet
atomicObjectSet :: P ObjectSet
glue          :: P Glue
qualIdent     :: P QualIdent 
bundleIdent   :: P BundleIdent  
unitIdent     :: P UnitIdent   
bundleTIdent  :: P BundleTIdent    
memberIdent   :: P MemberIdent    
flagsIdent    :: P FlagsIdent    
propIdent     :: P PropIdent    
typeIdent     :: P TypeIdent    
ruleIdent     :: P RuleIdent    
ident         :: P Ident       
fileName      :: P FileName    

defns = do
  ys <- cantFail "top level definitions" $ 
          zeroOrMore (oneOf deflt (oneOf directory (oneOf include defn)))
  let 
      defs = [ f | Left  f <- ys ]
      dirs = [ f | Right (Left  f) <- ys ]
      xs   = [ f | Right (Right f) <- ys ]

      xs1 = case dirs of
            []  -> xs
            [d] -> map (setDir d) xs
            _   -> my_error Nothing "Multiple directories not allowed"

  return (map (setDefaults defs) xs1)
         
setDir :: FileName -> Either FileName (Ident, Defn) -> Either FileName (Ident, Defn)
setDir d (Right (x, Unit (rs,ps,rules,cs,is,fs,ds,f,b))) = Right (x, Unit (rs,ps,rules,cs,is,fs,ds,f,b'))
 where
  b' = case b of
        Left (fs,rs) -> Left (map set fs,rs)
         where
          set (SrcDir Nothing x, y) = (SrcDir (Just d) x, y)
          set x = x
        Right r -> Right r
setDir d x = x

-- add default rules to all atomic units
setDefaults :: [RuleIdent] -> Either FileName (Ident, Defn) -> Either FileName (Ident, Defn)
setDefaults d (Right (x, Unit (rs,ps,rules,cs,is,fs,ds,f,Left b))) = Right (x, Unit (rs,ps,rules++d,cs,is,fs,ds,f,Left b))
setDefaults d x = x

include   = do{ lit "include";   fileName }
directory = do{ lit "directory"; fileName }
deflt     = do{ lit "default";   ruleIdent }

defn = eqn (lit "bundletype") bundleTIdent (fmap BundleT bundleType)
   ||| eqn (lit "unit")       unitIdent    (fmap Unit  unit ||| fmap ModUnit modunit)
   ||| eqn (lit "flags")      flagsIdent   (fmap Flags flags)
   ||| eqn (lit "package")    packageIdent (fmap Package package)
   ||| do{ lit "property"; x <- propIdent; return (x, Prop) }
   ||| do
         lit "type"
         x <- typeIdent
         ss <- (punc "<=" >> supertypes) ||| (return [])
         return (x, Type ss)
   ||| eqn (lit "rule") ruleIdent (fmap Rule (braces (zeroOrMore clause)))
   ||| do{ s <- doccomment; return (emptyIdent, Doc s) } -- hack

emptyIdent = mkIdent ""

supertypes
  =   do{ super <- typeIdent; return [super] }
  ||| parens (sloppy_list typeIdent)

eqn tag x y = do { tag;  a <- x;  equals; b <- y; return (a,b) }

modunit = do
  lit "modify"
  x <- unitIdent
  lit "deleting"
  rs <- braces' "deleting" $ remunit
  lit "adding"
  as <- braces' "adding" $ addunit
  return (x,rs,as)

remunit = do
  iss <- zeroOrMore $ do{ lit "imports"; is <- sloppy_btuple bundleIdent; semi; return is }
  ess <- zeroOrMore $ do{ lit "exports"; es <- sloppy_btuple bundleIdent; semi; return es }
  css <- zeroOrMore $ do{ lit "constraints"; cs <- cantFail "constraints" $ braces (sloppy_semi_list constraint); semi; return cs }
  deps <- zeroOrMore $ do { lit "depends"; ds <- cantFail "depend" $ braces (sloppy_semi_list depend); semi; return ds }
  gls <- zeroOrMore globals
  gss <- zeroOrMore glue'
  defss <- zeroOrMore defaults
  bss <- zeroOrMore $ do{ lit "link"; bs <- braces (sloppy_list bundleIdent); semi; return bs }
  return (concat iss, concat ess, concat css, concat deps, concat gls, concat gss, concat defss, concat bss)

addunit = do
  iss <- zeroOrMore $ do{ lit "imports"; is <- cantFail "imports" $ sloppy_btuple import'; semi; return is }
  ess <- zeroOrMore $ do{ lit "exports"; es <- cantFail "exports" $ sloppy_btuple export'; semi; return es }
  css <- zeroOrMore $ do{ lit "constraints"; cs <- cantFail "constraints" $ braces (sloppy_semi_list constraint); semi; return cs }
  deps <- zeroOrMore $ do { lit "depends"; ds <- cantFail "depend" $ braces (sloppy_semi_list depend); semi; return ds }
  gls <- zeroOrMore globals
  gss <- zeroOrMore glue'
  defss <- zeroOrMore defaults
  bss <- zeroOrMore $ do{ lit "link"; cantFail "bindings" $ x_sloppy_list "binding" lbrace rbrace semi binding }
  return (concat iss, concat ess, concat css, concat deps, concat gls, concat gss, concat defss, concat bss)

unit = braces' "unit" $ do 
  rs <- cantFail "imports" $ do
    lit "imports"
    x <- sloppy_btuple import'
    semi
    return x
  ps <- cantFail "exports" $ do
    lit "exports"
    x <- sloppy_btuple1 export'
    semi
    return x
  rules <- zeroOrMore $ do
    lit "rules"
    cs <- cantFail "rules" $ braces (sloppy_list ruleIdent)
    semi
    return cs
  css <- zeroOrMore $ do
    lit "constraints"
    cs <- cantFail "constraints" $ braces (sloppy_semi_list constraint)
    semi
    return cs
  is <- zeroOrMore initializer
  fs <- zeroOrMore finalizer
  ds <- zeroOrMore $ do { lit "depends"; ds <- cantFail "depend" $ braces (sloppy_semi_list depend); semi; return ds }
  mbflatten <- optional (do{ f <- flatten; semi; return f}) 
  body <- cantFail "unit body" $ oneOf atomicUnit bindings
  optional semi
  either (const $ if null ds then errorP yoiksEmptyDeps else return ()) 
         (const $ return ()) 
         body
  return (rs, ps, concat rules, concat css, is, fs, concat ds, mbflatten, body)
 where
  yoiksEmptyDeps s = "Empty dependency list not allowed in atomic units"
                   ++ "\nnear input: " ++ unwords (map show $ take 100 s)

  flatten =   do{ lit "flatten";   return True }
          ||| do{ lit "noflatten"; return False }

-- Concrete initialisers/finalisers - and the things they initialise/finalise
initializer, finalizer :: P (QualIdent,ObjectSet)
initializer = do
  lit "initializer"
  x <- qualIdent
  lit "for"
  ds <- objectSet
  semi
  return (x, ds)

finalizer = do
  lit "finalizer"
  x <- qualIdent
  lit "for"
  ds <- objectSet
  semi
  return (x, ds)

atomicUnit :: P AtomicUnit
atomicUnit = do
  fss <- withFlags fileSet `sloppy_sepdBy1` semi
  rns <- zeroOrMore rename
  return ([ (f,flags) | (fs,flags) <- fss, f <- fs ], concat rns)
 where
  fileSet
    =   do{ code <- ccode; return [AnonCode code] }
    ||| do
          lit "files"
          dir <- optional fileName
          fileList <- braces' "fileList" (sloppy_list fileName)
          return (map (SrcDir dir) fileList)

  rename = do
    lit "rename"
    renameList <- braces' "renameList"
			  (sloppy_semi_list (renameTo ||| renameFooFix))
    semi
    return renameList

  renameTo = do
    x <- bundleIdent
    punc "."
    y <- ident
    lit "to"
    z <- ident
    return (RenameTo (x,y) z)

  renameFooFix = do
    x <- bundleIdent
    lit "with"
    which <- oneOf (lit "prefix") (lit "suffix")
    y <- ident
    return (either (const (RenamePrefix x y)) (const (RenameSuffix x y)) which)

withFlags :: P a -> P (a, Flags)
withFlags xs = do
  x <- xs
  flags <- optional $ do 
    lit "with"
    lit "flags"
    fs <- flags
    return fs
  return (x, fromMaybe [] flags)

bindings = do
  gss <- zeroOrMore globals
  uss <- zeroOrMore glue'
  dss <- zeroOrMore defaults
  lit "link"
  bs <- cantFail "bindings" $ x_sloppy_list "binding" lbrace rbrace semi binding
  return (concat gss, concat uss, concat dss, bs)

globals = do
  lit "global"
  ds <- braces (sloppy_list global)
  semi
  return ds

global = bundleIdent

glue' = do
  lit "glue"
  ds <- braces (sloppy_list glue)
  semi
  return ds

defaults = do
  lit "defaults"
  ds <- braces (sloppy_list bundleIdent)
  semi
  return ds


import' 
  = do i <- bundleIdent
       colon                    -- no longer optional to avoid an ambiguity
       s <- bundleType
       return (i, s)

export' = import'

bundleType = withFlags bundleType'
bundleType'
  =   do{ bt <- bundleTIdent; return [Extends bt] }
  ||| braces' "bundletype" (sloppy_list btElement)

-- the order of these is kinda important - if you put Ident too early,
-- you can't backtrack enough
btElement
  =   do { lit "extends"; x <- bundleTIdent; return (Extends x) }
  ||| do { lit "include"; x <- fileName;     return (Include x) }
  ||| do {                x <- memberIdent;  return (Ident   x) }

constraint = do{ x <- property; op <- rel; y <- property; return (op x y) }
 where
  rel =   do{ punc "=";  return Eq }
      ||| do{ punc "<="; return Le }
      ||| do{ punc ">="; return (flip Le) }

property 
  =   do{ x <- ident; (fmap (PropApp x) atomicObjectSet) ||| return (PropConst x) }

clause = do
  lit "var"
  vs <- sloppy_list ident
  semi
  ((|||) (do{ f <- formula; punc "=>"; a <- atomic; return (vs, Just f, a) })
         (do{ a <- atomic; return (vs, Nothing, a) }))
         
formula = do
  f <- formula2
  (    do{ fs <- oneOrMore(punc "&" >> formula2); return (foldl (:&:) f fs) }
   ||| do{ fs <- oneOrMore(punc "|" >> formula2); return (foldl (:|:) f fs) }
   ||| return f
   )

formula2
  =   do{ a <- atomic; return (Atomic a) }
  ||| do{ punc "!"; f <- formula2; return (Not f) }
  ||| do{ f <- parens formula; return (Parens f) }

atomic 
  =   do{ lit "import"; v <- ident; return (Import v) }
  ||| do{ lit "export"; v <- ident; return (Export v) }
  ||| do{ lit "init";   v <- ident; return (Init   v) }
  ||| do{ lit "fini";   v <- ident; return (Fini   v) }
  ||| do{ p1 <- aProp; punc "<="; p2 <- aProp; return (ALe p1 p2) }

aProp 
  =   do{ p <- propIdent; x <- parens ident; return (APropApp p x) }
  ||| do{ c <- typeIdent; return (APropConst c) }

package = braces' "package" (sloppy_list glue)

glue =   do{ lit "package"; p <- packageIdent; return (GluePackage p) }
     ||| do{                u <- unitIdent;    return (GlueIdent   u) }

binding = multiBind ||| singleBind

multiBind = do
  xs <- sloppy_btuple1 bundleIdent
  punc "<-"
  r <- unitInstance
  return (MultiBind xs r)

singleBind = do
  x <- bundleIdent
  punc "="
  b <- bundle
  return (SingleBind x b)

unitInstance_easy
  =   do{ lit "unit";      u <- unit; return (RhsUnit u) }
  ||| do{ lit "flatten";   u <- unitInstance;  return (RhsFlatten True  u) }
  ||| do{ lit "noflatten"; u <- unitInstance;  return (RhsFlatten False u) }
  ||| do{ code <- ccode; return (RhsUnit (codeToUnit code)) }

unitInstance 
  =   unitInstance_easy
  ||| do
        u <- unitIdent
        punc "<-"
        args <- oneOf posArgs nameArgs
        return (App u args)

posArgs = sloppy_btuple posArg
posArg  = arg

nameArgs = braces' "nameArgs"
  (   do{ punc "..."; return ([], True) }
  ||| do{ as <- list nameArg
 	; ellipsis <- optional $ do{ comma; punc "..." }
 	; optional comma
 	; return (as, isJust ellipsis)  
        }
  )
 	
nameArg = do
  x <- bundleIdent
  i <- (do{ equals; arg } ||| return (Left $ BundleIdent x))
  return (x,i)

arg = oneOf bundle (parens unitInstance) 

-- This is something of a hack: we "parse" the code looking for
-- sequences of the form 
--
--   <anything> ident = <anything> ;
--
codeToUnit :: [String] -> Unit
codeToUnit ls = ([],[p],[],[],[],[],[],Just True,Left ([(AnonCode ls, [])],[]))
 where
  code = unlines ls
  
  -- things separated by semis (braces and strings are ignored)
  defns :: [String] 
  defns = chopList (== ';') code

  vars = [ dropWhile (== '*') $ takeWhile (/= '[') $ last (chopList (== ' ') lhs)
         | defn <- defns
         , [lhs,init] <- [chopList (== '=') defn]
         ]
  bt = ([ Ident (mkIdent v) | v <- vars ], [])
  p = (mkIdent "__anon", bt)

bundle = do
  bs <- atomic_bundle `sloppy_sepdBy1` punc "+"
  case bs of
   [b] -> return b
   _   -> return (BundleUnion bs)

atomic_bundle
  =   do{ lit "find"; s <- bundleType; return (BundleFind s) }
  ||| do{ r <- bundleBits;  return (BundleBits r) }
  ||| do{ x <- bundleIdent; return (BundleIdent x) }
  ||| parens bundle

bundleBits = braces' "renaming" (sloppy_list1 bundleElement)

bundleElement = do
  lhs <- ident
  x <- optional $ do
    equals
    qualIdent
  return (lhs, fromMaybe (Nothing, lhs) x)

depend = do
  ls <- objectSet
  r <- depRel
  rs <- objectSet
  return (ls, r, rs)

objectSet = do
  x <- atomicObjectSet
  (    do
         punc "+"
         xs <- atomicObjectSet `sloppy_sepdBy1` punc "+"
         return (foldl ObjUnion x xs)
   ||| do
         punc "-"
         y <- atomicObjectSet
         return (ObjDiff x y)
   ||| return x
   )

atomicObjectSet
  =   do{ lit "imports"; return ObjImports }
  ||| do{ lit "exports"; return ObjExports }
  ||| do{ lit "inits";   return ObjInits   }
  ||| do{ lit "finis";   return ObjFinis   }
  ||| do{ x <- bundleIdent; return (ObjIdent x) }
  ||| do{ qs <- braces' "objectSet" (sloppy_list qualIdent); return (ObjSet qs)}
  ||| do{ p  <- parens objectSet; return (ObjParens p) }

depRel 
  =   (lit  "needs" >> return Calls)
  ||| (punc "<"     >> return Precedes)

qualIdent = do
  x <- ident
  (    do { punc "."; y <- ident; return (Just x,  y) }
   ||| do {                       return (Nothing, x) }
   )

flags
  =   do{ f <- flagsIdent; return [UseFlags f] }
  ||| braces' "flags" (sloppy_list flag)

flag
  =   do { lit "flags"; x <- flagsIdent; return (UseFlags x) }
  ||| do {              x <- string;     return (RawFlag x) }

fileName = string

bundleIdent  = ident
unitIdent    = ident
bundleTIdent = ident
memberIdent  = ident
packageIdent = ident
flagsIdent   = ident
propIdent    = ident
typeIdent    = ident
ruleIdent    = ident

ident = do { Var i <- tok' isIdent; return (mkIdent i) }
 where
  isIdent (Var xs) = True
  isIdent _        = False

lit :: String -> P ()
lit s = do
  t <- token'
  case t of
   Var s' -> guard (s==s')
   _      -> mzero

punc :: String -> P ()
punc s = do
  t <- token'
  case t of
   Punc s' -> guard (s==s')
   _       -> mzero

string :: P String
string = do
  t <- token'
  case t of
   Str s -> return s
   _     -> mzero

ccode :: P [String]
ccode = do
  t <- token'
  case t of
   CCode s -> return s
   _       -> mzero

doccomment :: P String
doccomment = do
  t <- token'
  case t of
   DocComment ss -> return (concat ss)
   _             -> mzero

semi, comma, colon,equals :: P ()
semi   = punc ";"
comma  = punc ","
colon  = punc ":"
equals = punc "="

lparen, rparen, lbrace, rbrace, lbrack, rbrack :: P ()
lparen = punc "("
rparen = punc ")"
lbrace = punc "{"
rbrace = punc "}"
lbrack = punc "["
rbrack = punc "]"

parens :: P a -> P a
parens p = do
  lparen
  a <- p
  rparen
  return a
  
parens' :: String -> P a -> P a
parens' what p = do
  lparen
  cantFail what $ do 
    a <- p
    cantFail what $ rparen
    return a
  
braces :: P a -> P a
braces p = do
  lbrace
  a <- p
  rbrace
  return a

braces' :: String -> P a -> P a
braces' what p = do
  lbrace
  cantFail what $ do
    a <- p
    cantFail what $ rbrace
    return a
  
bracket	:: P a -> P a
bracket	p = do
  lbrack
  a <- p
  rbrack
  return a

bracket' :: String -> P a -> P a
bracket' what p = do
  lbrack
  cantFail what $ do
    a <- p
    cantFail what $ rbrack
    return a
  
list, list1 :: P a -> P [a]
list p  = p `sepdBy` comma
list1 p = p `sepdBy1` comma

-- Rename to `sloppy_comma_list'?
sloppy_list, sloppy_list1 :: P a -> P [a]
sloppy_list p  = p `sloppy_sepdBy`  comma
sloppy_list1 p = p `sloppy_sepdBy1` comma

sloppy_semi_list, sloppy_semi_list1 :: P a -> P [a]
sloppy_semi_list p  = p `sloppy_sepdBy`  semi
sloppy_semi_list1 p = p `sloppy_sepdBy1` semi

tuple, tuple1 :: P a -> P [a]
tuple  = parens . list
tuple1 = parens . list1

sloppy_tuple, sloppy_tuple1 :: P a -> P [a]
sloppy_tuple  = parens . sloppy_list
sloppy_tuple1 = parens . sloppy_list1

sloppy_btuple, sloppy_btuple1 :: P a -> P [a]
sloppy_btuple  = bracket . sloppy_list
sloppy_btuple1 = bracket . sloppy_list1

-- temporary hack: allow braces as well
sloppy_tuple', sloppy_tuple1' :: P a -> P [a]
sloppy_tuple'  p = parens (sloppy_list p)  ||| braces (sloppy_list p)
sloppy_tuple1' p = parens (sloppy_list1 p) ||| braces (sloppy_list1 p)


-- ToDo: This bit has been subject to severe hackery including random
-- insertion and deletion of calls to force.  Not sure it works...

x_sloppy_list, x_sloppy_list1 :: String -> P l -> P r -> P punc -> P a -> P [a]
x_sloppy_list  what l r sep p = do { l; x_sepdBy  what p sep r }
x_sloppy_list1 what l r sep p = do { l; x_sepdBy1 what p sep r }


x_sepdBy :: String -> P a -> P b -> P c -> P [a]
x_sepdBy what p punc r = {-force-} (x_sepdBy1 what p punc r ||| do{ optional punc; r; return []})

x_sepdBy1 :: String -> P a -> P b -> P c -> P [a]
x_sepdBy1 what p punc r = {-force $-} do 
  a <- p
  as <- x_sepdBy' what p punc r
  return (a:as)

x_sepdBy' :: String -> P a -> P b -> P c -> P [a]
x_sepdBy' what p punc r = {-force-} (
  do { r; return [] }
  |||
  do { punc
     ; (   x_sepdBy1 what p punc r
       ||| (cantFail what $ do { r; return [] })
       )
     }
  )

----------------------------------------------------------------
-- Lexer
----------------------------------------------------------------

run :: P a -> [Token'] -> a
run p s = case runP p s of
  Nothing     -> my_error Nothing "Parse failed"
  Just (a,[]) -> a
  Just (a,s') -> my_error (toksToLoc s') $ "Parse failed"

cantFail :: String -> P a -> P a
cantFail what p = p ||| errorP genMsg
 where
  genMsg s = my_error (toksToLoc s) $ "Syntax error: '" ++ what ++ "' construct expected"

type Loc = (FileName, Line)

toksToLoc :: [Token'] -> Maybe Loc
toksToLoc []    = Nothing 
toksToLoc (t:_) = Just (tokToLoc t)

tokToLoc :: Token' -> Loc
tokToLoc (Token' f l t) = (f,l)

showLoc :: Loc -> String
showLoc (f,l) = f ++ ":" ++ show l ++ ": "


-- Late addition: we tag tokens with file positions in the usual way 
data Token' = Token' FileName Line Token

instance Show Token' where showsPrec d (Token' _ _ t) = showsPrec d t

type Line = Int

token' :: P Token
token' = do{ Token' _ _ t <- token; return t }

tok' :: (Token -> Bool) -> P Token
tok' p = do{ Token' _ _ t <- token; guard (p t); return t }

data Token 
  = Var   String
  | Punc  String  
  | Num   String
  | Str   String
  | CCode [String]
  | DocComment [String]
  | EOF

instance Show Token where
  showsPrec p (Var   s) = showString s
  showsPrec p (Punc  s) = showString s
  showsPrec p (Num   s) = showString s
  showsPrec p (Str   s) = showChar '"' . showString s . showChar '"'
  showsPrec p (CCode s) = showString "%{\n" . showString (unlines s) . showString "\n%}"
  showsPrec p (DocComment ss) = showString "/*# " . showString (unlines ss) . showString " #*/"
  showsPrec p EOF       = showString "EOF"

type P a = Parser [Token'] a

tokenise :: (Line -> ReadS (Line,Token)) -> FileName -> Line -> String -> [Token']
tokenise f fn l [] = []
tokenise f fn l xs = 
  case f l xs of
  [((_,EOF),[])] -> []
  [((l',a),xs')] -> Token' fn l' a : tokenise f fn l' xs'
  [] 
    | all isSpace xs -> []
    | otherwise -> my_error (Just (fn, l)) $ "Invalid token: " ++ take 80 xs
  _  -> my_error (Just (fn, l)) $ "Ambiguous token: " ++ take 80 xs

len :: Token -> Line
len (Str s)    = length (filter (== '\n') s)
len (CCode ss) = length ss - 1
len (DocComment ss) = length ss - 1
len _ = 0

-- sleazily use the Haskell lexer - close enough for now
tokenise' :: FileName -> String -> [Token']
tokenise' fn = tokenise lex' fn 1
 where

  lex' :: Line -> ReadS (Line,Token)
  lex' l ('\n':s)          = (lex' $! (l+1)) $! s
  lex' l (c:s) | isSpace c = lex' l s
  lex' l ('/':'*':s)       = lexComment l s
   where  
    lexComment l ('*':'/':r) = lex' l r
    lexComment l ('/':'*':r) = my_error loc "Nested comment" where loc = Just (fn,l)
    lexComment l ('\n':r)    = (lexComment $! (l + 1)) $! r
    lexComment l (c:r)       = lexComment l r
    lexComment l []          = my_error loc "File ends before end of comment" where loc = Just (fn,l)
  lex' l ('/':'/':s)         = lex' l (dropWhile (/= '\n') s)
  lex' l s = [ l' `seq` ((l', t), s') | (t, s') <- lex s, let l' = l + len t ]
   where
    
    -- almost the standard Haskell lexer - except idents can start with __
    -- (changes to isSingle and extra case for "_...")
    -- also added support for comments
    lex                    :: ReadS Token
  
    lex ""                  = [(EOF,"")]
    lex ('/':'*':'#':s)     = lexDocComment [] [] s
    lex ('*':'/':s)         = my_error loc "Dangling */" where loc = Just (fn,l)
    lex ('%':'}':s)         = my_error loc "Dangling %}" where loc = Just (fn,l)
    lex ('%':'{':s)         = lexCCode s [] []
     where
      lexCCode ('%':'}':r) x xs = [(CCode (reverse (reverse x:xs)), r)]
      lexCCode ('\n':r)    x xs = lexCCode r [] (reverse x:xs)
      lexCCode (c:r)       x xs = lexCCode r (c:x) xs
      lexCCode []          x xs = my_error loc "File ends before end of %{ ..." where loc = Just (fn,l)
  -- Moved this functionality out of lex because it complicates 
  -- counting how many lines a token spans
  --  lex (c:s) | isSpace c   = lex (dropWhile isSpace s)
    lex ('"':s)             = [(Str str, t)      | (str,t) <- lexString s]
                              where
                              lexString ('"':s)    = [("",s)]
                              lexString ('\\':c:s) = [(c:str, u)
                                                  | (str,u) <- lexString s ]
                              lexString (c:s)   = [(c:str, u)
                                                  | (str,u) <- lexString s ]
    lex ('\'':s)            = [(Var nam,t) | (nam, '\'':t) <- [span (/= '\'') s] ]
    lex ('<':'=':s)         = [(Punc "<=",s)]
    lex ('>':'=':s)         = [(Punc ">=",s)]
    lex ('<':'-':s)         = [(Punc "<-",s)]
    lex ('.':'.':'.':s)     = [(Punc "...",s)]
    lex (c:s) | isSingle c  = [(Punc [c],s)]
              | isIdChar c  = [(Var (c:nam),t)         | (nam,t) <- [span isIdChar s]]
              | otherwise   = []    -- bad character
                    where
                    isSingle c  =  c `elem` ".,:;()[]{}=+-<"
                    isIdChar c  =  isAlphaNum c || c == '_'
      
    lexDocComment ss s ('#':'*':'/':r) = [(DocComment (reverse s : ss), r)]
    lexDocComment ss s ('/':'*':r) = my_error loc "Nested comment in doc comment" where loc = Just (fn,l)
    lexDocComment ss s ('\n':r)    = lexDocComment (reverse s : ss) [] r
    lexDocComment ss s (c:r)       = lexDocComment ss (c:s) r
    lexDocComment ss s []          = my_error loc "File ends before end of doc comment" where loc = Just (fn,l)
  
my_error :: Maybe Loc -> String -> a
my_error loc msg = unsafePerformIO $ do
  maybe (return ()) (putStr . showLoc) loc
  putStrLn msg
  exitFailure

----------------------------------------------------------------
-- Tests
----------------------------------------------------------------

t1 :: IO ()
t1 = do
  xs <- readDefns ["."] ["test"]
  print $ PP.defns xs

t2 :: IO ()
t2 = do
  xs <- readDefns ["."] ["mtest1"]
  print $ PP.defns xs

t3 :: String -> IO ()
t3 x = do
  xs <- readDefns ["."] ["mtest1"]
  print $ PP.defns $ filter ((== mkIdent x).fst) xs

----------------------------------------------------------------
-- End
----------------------------------------------------------------
