-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

----------------------------------------------------------------
-- Some parsing utilities
----------------------------------------------------------------

module ParserUtils
	( module ParserUtils
	, module Parser
	, mplus
	) where

import Parser
import Char(isSpace)
import Monad

----------------------------------------------------------------
-- General Purpose
----------------------------------------------------------------

-- get the next token
token :: Parser [t] t
token = tokenP unCons
 where
  unCons :: [a] -> Maybe (a,[a])
  unCons (x:xs) = Just (x,xs)
  unCons []     = Nothing

-- test that next token is what we wanted
tok :: (token -> Bool) -> Parser [token] token
tok p = do 
  t <- token
  guard (p t)
  return t

-- test that next token is what we wanted
lit :: (Eq token) => token -> Parser [token] ()
lit t = tok (==t) >> return ()

optional :: Parser s a -> Parser s (Maybe a)
optional p = 
  do { a <- p; return (Just a) }
  `mplus` 
  do { return Nothing }

sepdBy :: Parser s a -> Parser s b -> Parser s [a]
p `sepdBy` punc = force ((p `sepdBy1` punc) `mplus` return [])

sepdBy1 :: Parser s a -> Parser s b -> Parser s [a]
p `sepdBy1` punc = do 
  a <- p
  as <- p `sepdBy'` punc
  return (a:as)

sepdBy' :: Parser s a -> Parser s b -> Parser s [a]
p `sepdBy'` punc = force (
  do { punc; p `sepdBy1` punc }
  `mplus` 
  do { return [] }
  )
 
-- just like sepdBy except you can optionally put an extra separator at the end
sloppy_sepdBy  :: Parser s a -> Parser s b -> Parser s [a]
sloppy_sepdBy1 :: Parser s a -> Parser s b -> Parser s [a]
p `sloppy_sepdBy` punc = do { as <- p `sepdBy` punc; optional punc; return as }
p `sloppy_sepdBy1` punc = do { as <- p `sepdBy1` punc; optional punc; return as }

oneOf :: Parser s a -> Parser s b -> Parser s (Either a b)
oneOf p1 p2 = do{ a <- p1; return (Left a) } `mplus` do{ b <- p2; return (Right b) }

(|||) :: MonadPlus m => m a -> m a -> m a
(|||) = mplus

----------------------------------------------------------------
-- Special purpose: Token == String
----------------------------------------------------------------

run :: Parser [String] a -> [String] -> a
run p s = case runP p s of
  Nothing     -> error "Parse failed"
  Just (a,[]) -> a
  Just (a,s') -> error $ "Didn't use all input " ++ unwords (take 10 s')

semi, comma, colon,equals :: Parser [String] ()
semi = lit ";"
comma = lit ","
colon = lit ":"
equals = lit "="

lparen, rparen, lbrace, rbrace, lbrack, rbrack :: Parser [String] ()
lparen = lit "("
rparen = lit ")"
lbrace = lit "{"
rbrace = lit "}"
lbrack = lit "["
rbrack = lit "]"

string :: Parser [String] String
string = do 
  s <- tok isString
  return (tail (init s))
 where
  isString ('"':_) = True
  isString _       = False

parens :: Parser [String] a -> Parser [String] a 
parens p = do
  lparen
  a <- p
  rparen
  return a

braces :: Parser [String] a -> Parser [String] a 
braces p = do
  lbrace
  a <- p
  rbrace
  return a

bracket	:: Parser [String] a -> Parser [String] a 
bracket	p = do
  lbrack
  a <- p
  rbrack
  return a

list, list1 :: Parser [String] a -> Parser [String] [a]
list p  = p `sepdBy` comma
list1 p = p `sepdBy1` comma

sloppy_list, sloppy_list1 :: Parser [String] a -> Parser [String] [a]
sloppy_list p  = p `sloppy_sepdBy`  comma
sloppy_list1 p = p `sloppy_sepdBy1` comma

tuple, tuple1 :: Parser [String] a -> Parser [String] [a]
tuple  = parens . list
tuple1 = parens . list1

sloppy_tuple, sloppy_tuple1 :: Parser [String] a -> Parser [String] [a]
sloppy_tuple  = parens . sloppy_list
sloppy_tuple1 = parens . sloppy_list1


tokenise :: ReadS String -> String -> [String]
tokenise f [] = []
tokenise f xs = 
  case f xs of
  [("",[])] -> []
  [(a,xs')] -> a : tokenise f xs'
  [] 
    | all isSpace xs -> []
    | otherwise -> error $ "Invalid token: " ++ take 80 xs
  _  -> error $ "Ambiguous token: " ++ take 80 xs
 
----------------------------------------------------------------
-- End
----------------------------------------------------------------
