-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module RN
	( rnDefn, mkEnv, Env
        , builtin_vars
	) where

import SExpr
import Prelude hiding (init)
import Maybe( fromMaybe, listToMaybe )

import AbstractC as C
import ParseC( parseDefns ) -- testing only
import PrettyUtils

----------------------------------------------------------------
-- Renaming
--
-- A conventional renamer.
----------------------------------------------------------------

rnQualId      :: Env -> QualId -> QualId
rnQualFIdent  :: Env -> QualId -> QualId
rnTag         :: Env -> QualId -> QualId
rnTypeName    :: Env -> QualId -> QualId
mkEnv         :: [(Ident,QualId)] -> Qualifier -> Qualifier -> [QualId] -> Env
addLocalIds   :: [Ident] -> Env -> Env
addLocalTags  :: [Ident] -> Env -> Env

rnDefn        :: Env -> Defn -> Defn
rnType        :: Env -> Type -> Type
rnExpr        :: Env -> Expr -> Expr
rnCompound    :: Env -> Compound -> Compound
rnStatement   :: Env -> Statement -> Statement
rnDecl        :: Env -> Decl -> Decl
rnInit        :: Env -> Init -> Init
rnInitializer :: Env -> Initializer -> Initializer

-- vars/typedefs are in a different namespace from tags
type Env =
	( NameSpace -- vars & typedefs
	, NameSpace -- tags
	, Qualifier -- per module 
        , [QualId]
        , Qualifier -- per file (gets concatenated with per-module)
	) 

mkEnv xs q fq is = ((builtin_vars,xs,builtinable_vars),([],[],[]),q,is,fq)

builtin_vars = map mkIdent
  [ "__builtin_constant_p"
  , "__builtin_memcmp"
  , "__constant_memcpy"
  , "__builtin_memcpy"
  , "__builtin_strlen"
  , "__builtin_alloca"
  , "__memcpy"
  , "__constant_c_and_count_memset"
  , "__constant_c_memset"
  , "__memset_generic"
  , "__FUNCTION__"
  , "__PRETTY_FUNCTION__"
  , "__builtin_return_address"
  ]
-- The following can be converted to __builtin_ form by gcc,
--  so the unit specification doesn't necessarily require an
--  import for them. Instead, force the use of the builtin version.
builtinable_vars = map mkIdent
  [ "memcpy"
  , "strcpy"
  , "memcmp"
  , "strlen"
  , "alloca"
  ]

type NameSpace =
	( [Ident] -- locals
	, [(Ident,QualId)]
        , [Ident] -- builtinable
	)

rn :: NameSpace -> Qualifier -> QualId -> QualId
rn (l,g,b) m qx
  | unQualify qx `elem` l = qx

  | x `elem` b = qx -- not remotely sure this is right - ADR

  | otherwise 
  = case lookup x g of
   Just q     -> q
   Nothing    -> qualify m x
 where
  x = unQualify qx

--rnQualId (a,b,q,fq) x = rn a q x
rnQualId = rnQualFIdent

rnQualFIdent (a,b,q,is,fq) x = rn a q' x where q' = if x `elem` is then fq `concatQual` q else q
rnTag        (a,b,q,is,fq) x = qualify (fq `concatQual` q) (unQualify x)

rnTypeName  (a,b,q,is,fq) qx
  | qx `elem` builtin_types
  = qx
rnTypeName  env x = rnQualFIdent env x

builtin_types = map mkQualId
    [ "volatile"
    , "static"
    , "const"
    , "unsigned"
    , "signed"
    , "__signed__"
    , "__inline"
    , "inline"
    , "extern"
    , "struct"
    , "register"
    , "void"
    , "int"
    , "short"
    , "long"
    , "char"
    , "float"
    , "double"
    ]


-- local names don't get renamed in any way
addLocalIds  ids ((l,g,b),ys,q,is,fq) = ((ids++l,g,b),ys,q,is,fq)
addLocalTags ids (xs,(l,g,b),q,is,fq) = (xs,(ids++l,g,b),q,is,fq)

-- separated out so we can add _scc_ easily
rnDefn env d = rnDefn' env d

rnDefn' env (Enum   x es) = Enum (rnTag env x) (map (rnEnumerator env) es)
rnDefn' env (Union  x fs) = Union (rnTag env x) (map (rnField env) fs)
rnDefn' env (Struct x fs) = Struct (rnTag env x) (map (rnField env) fs)
rnDefn' env (Typedef x t) = Typedef (rnQualFIdent env x) (rnType env t)
rnDefn' env (Extern d)    = Extern (rnDefn' env d)
rnDefn' env (Fun d ds c)  = Fun (rnDecl env d) (map (rnArgDecl env) ds) (rnCompound env' c)
 where
  env' = addLocalIds xs env
  xs   = case d of
         (_, Function1 as _ _ _) -> as
         (_, Function2 as _ _ _) -> [ x | (Just x,t) <- as ]
rnDefn' env x@(TopAsm a b c d e) = x
rnDefn' env (Init i)      = Init (rnInit env i)
--rnDefn' env x = x -- ToDo: until I fill in the missing stuff above

-- ToDo: rename dependencies properly
rnBody :: Env -> ([QualId], Compound) -> ([QualId], Compound)
rnBody env (x,c) = (map (rnQualId env) x, rnCompound env c)

rnEnumerator :: Env -> (QualId, Maybe Expr) -> (QualId, Maybe Expr)
rnEnumerator env (x,e) = (rnQualFIdent env x, rnmbe env e)

rnField :: Env -> (Maybe Ident, Maybe Expr, Type) -> (Maybe Ident, Maybe Expr, Type)
rnField env (x,e,t) = (x, rnmbe env e, rnType env t)

rnDecl env (v,t) = (rnQualId env v, rnType env t)
rnArgDecl env (v,t) = (v, rnType env t)
rnInit env (d,i) = (rnDecl env d, rnInitializer env i)

rnInitializer env None = None
rnInitializer env (SimpleInit e)  = SimpleInit (rnExpr env e)
rnInitializer env (FieldInit f e) = FieldInit f (rnInitializer env e)
rnInitializer env (ArrayInit i e) = ArrayInit (rnExpr env i) (rnInitializer env e)
rnInitializer env (ListInit is)   = ListInit (map (rnInitializer env) is)

-- Note: the rule for SimpleTy potentially zaps some things which we're
-- not parsing properly like "volatile" and certainly zaps "immutable types"
-- such as "char" - but surely no-one would try to rename "volatile" (which
-- is a C keyword) and if they rename "char", they probably know what they're
-- doing.
rnType env (Function1 ts b t as)   = Function1 ts b (rnType env t) as
rnType env (Function2 args b t as) = Function2 (map (rnArg env) args) b (rnType env t) as
rnType env (SimpleTy s)        	   = SimpleTy (rnTypeName env s)
rnType env (Ptr t)             	   = Ptr (rnType env t)
rnType env (Array e t)         	   = Array (rnmbe env e) (rnType env t)
rnType env (TyEnum x)          	   = TyEnum (rnTag env x)
rnType env (TyUnion x)         	   = TyUnion (rnTag env x)
rnType env (TyStruct x)        	   = TyStruct (rnTag env x)
rnType env (Const t)           	   = Const (rnType env t)
rnType env (TyAttribute as t)  	   = TyAttribute as (rnType env t)
rnType env (TyApp t1 t2)       	   = TyApp (rnType env t1) (rnType env t2)
rnType env (TypeOf e)      	   = TypeOf (rnExpr env e)
rnType env (ParensType t)      	   = ParensType (rnType env t)

rnArg :: Env -> (Maybe Ident,Type) -> (Maybe Ident,Type)
rnArg env (v,t) = (v, rnType env t)

rnExpr env (Ident x)          = Ident (rnQualId env x)
rnExpr env (Constant s)       = Constant s
rnExpr env (String s)         = String s
rnExpr env (Constructor t i)  = Constructor (rnType env t) (rnInitializer env i)
rnExpr env (App e es)	      = App (rnExpr env e) (map (rnExpr env) es)
rnExpr env (Select x e)	      = Select x (rnExpr env e)
rnExpr env (Arrow  x e)	      = Arrow  x (rnExpr env e)
rnExpr env (Cast t e)	      = Cast (rnType env t) (rnExpr env e)
rnExpr env (SizeOf_E e)	      = SizeOf_E (rnExpr env e)
rnExpr env (SizeOf_T t)	      = SizeOf_T (rnType env t)
rnExpr env (LabelAddr l)      = LabelAddr l
rnExpr env (Cond e1 e2 e3)    = Cond (rnExpr env e1) (rnExpr env e2) (rnExpr env e3)
rnExpr env (Cond2 e1 e2  )    = Cond2 (rnExpr env e1) (rnExpr env e2) 
rnExpr env (UnOp op e)	      = UnOp op (rnExpr env e)
rnExpr env (BinOp op e1 e2)   = BinOp op (rnExpr env e1) (rnExpr env e2)
rnExpr env (StatementExpr c)  = StatementExpr (rnCompound env c)

rnmbe env e = fmap (rnExpr env) e

-- FIXME: extern declarations can appear in `ls'
rnCompound  env (ls,is,ss) = (ls,map (rnInit env') is, map (rnStatement env') ss)
 where
  env' = addLocalIds [ unQualify x | ((x,ty),_) <- is, not (isExternType ty) ] env

isExternType ty = ty `isQualifiedBy` externQualid
externQualid = mkQualId "extern"
isQualifiedBy :: C.Type -> QualId -> Bool
isQualifiedBy (C.Function1 _ _ ty _)  q = ty `isQualifiedBy` q
isQualifiedBy (C.Function2 _ _ ty _)  q = ty `isQualifiedBy` q
isQualifiedBy (C.SimpleTy t)          q = t == q
isQualifiedBy (C.TyApp t1 t2)         q = isQualifiedBy t1 q || isQualifiedBy t2 q
isQualifiedBy (C.Ptr     ty)          q = ty `isQualifiedBy` q
isQualifiedBy (C.Array _ ty)          q = ty `isQualifiedBy` q
isQualifiedBy (C.Const   ty)          q = ty `isQualifiedBy` q
isQualifiedBy (C.TyAttribute _ ty)    q = ty `isQualifiedBy` q
isQualifiedBy (C.ParensType ty)       q = ty `isQualifiedBy` q
isQualifiedBy _                       q = False


rnStatement env (Return e)       = Return (rnmbe env e)
rnStatement env Break            = Break
rnStatement env Continue         = Continue
rnStatement env (Goto l)         = Goto l
rnStatement env (If e s1 s2)     = If (rnExpr env e) (rnStatement env s1) (rnStatement env s2)
rnStatement env (Switch e s)     = Switch (rnExpr env e) (rnStatement env s)
rnStatement env (Label l s)      = Label l (rnStatement env s)
rnStatement env (Case e s)       = Case (rnExpr env e) (rnStatement env s)
rnStatement env (Default s)      = Default (rnStatement env s)
rnStatement env (While e s)      = While (rnExpr env e) (rnStatement env s)
rnStatement env (Do s e)         = Do (rnStatement env s) (rnExpr env e)
rnStatement env (For e1 e2 e3 s) = For (rnmbe env e1) (rnmbe env e2) (rnmbe env e3) (rnStatement env s)
rnStatement env (Compound c)     = Compound (rnCompound env c)
rnStatement env (Expr e)         = Expr (rnExpr env e)
rnStatement env (Asm v op a1 a2 a3) = Asm v op
                               	    (rnAsmArgs env a1)
                               	    (rnAsmArgs env a2)
                               	    (rnAsmArgs env a3)
rnStatement env Empty            = Empty

rnAsmArgs :: Env -> [AsmArg] -> [AsmArg]
rnAsmArgs env = map (rnAsmArg env)

rnAsmArg :: Env -> AsmArg -> AsmArg
rnAsmArg env (x,y) = (x, rnmbe env y)

----------------------------------------------------------------
-- Tests
----------------------------------------------------------------

-- t6 pre fn = do
--   { ys <- parseDefns fn
--   ; let ns = names ys
--   ; let ys' = map (rnDefn (mkEnv pre ns)) ys
--   ; print $ vmap ppDefn ys'
-- --  ; writeFile "x.c" $ show $ vmap ppDefn ys'
--   }
