-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module SExpr
	( SExpr(..)
        , tokenize
        , parse
        , pp
        , toList
	) where

import AbstractC

import Prelude hiding (init)
import PrettyUtils
import Parser
import Maybe( fromMaybe, listToMaybe )
import Char( ord, chr, isAlphaNum, isSpace )
--import Trace
import Monad( MonadPlus(..), guard, unless, liftM, liftM2, liftM3, liftM4 )

tokenize :: String -> [String]
tokenize s = tokenize' [] (dropWhile isSpace s)
 where
  tokenize' [] []     = []
  tokenize' a  []     = a' `seq` [a']
       where a' = reverse a
  tokenize' a  (c:cs)
   | isSpace c = forceString a' `seq` (a' : tokenize' [] (dropWhile isSpace cs))
   | otherwise = tokenize' (c:a) cs
       where a' = reverse a

forceString :: String -> ()
forceString (c:cs) = c `seq` forceString cs
forceString [] = ()

t1 fn = do
  { cs <- readFile fn
  ; putStr $ unlines $ tokenize cs
  }

data SExpr
  = Nil
  | Cons SExpr SExpr
  | Thing String SExpr
  | Atom  String
 deriving Show

toList :: SExpr -> [SExpr]
toList Nil = []
toList (Cons x y) = x : toList y
toList x = error $ "toList " ++ show (pp x)

infixr 5 |||

(|||) :: MonadPlus m => m a -> m a -> m a
(|||) = mplus

{-
token :: Parser [String] String
token = tokenP myHead
 where  myHead []     = Nothing
        myHead (c:cs) = Just (c,cs)

lit :: String -> Parser [String] ()
lit x = do
  { tok <- token
  ; guard (tok == x)
  }

parse :: [String] -> [SExpr]
parse = fromMaybe (error "parse error") . parse'

parse' :: [String] -> Maybe [SExpr]
parse' = fmap fst . runP sexprs

sexprs :: Parser [String] [SExpr]
sexprs = oneOrMore sexpr

sexpr :: Parser [String] SExpr
sexpr 
  =   do 
        { lit "{nil}"
        ; return Nil
        }
  ||| do 
        { lit "("
        ; xs <- oneOrMore $ do 
                  { x <- sexpr
		  ; lit ","
		  ; return x
                  }
	; y <- sexpr
        ; lit ")"
        ; return (foldr Cons y xs)
        }
  ||| do 
        { lit "["
        ; l <- token
        ; x <- sexpr
        ; lit "]"
        ; return (Thing l x)
        }
  ||| do 
        { l <- token
        ; return (Atom l)
        }


-}

-- this hand-coded parser is something like 10 times faster than
-- the parser-combinator version above!

parse :: [String] -> [SExpr]
parse = sexprs

sexprs :: [String] -> [SExpr]
sexprs [] = []
-- sexprs xs = e : sexprs xs' where (e,xs') = sexpr xs
sexprs xs = case sexpr xs of { (e,xs') -> e : sexprs xs' }

sexpr :: [String] -> (SExpr,[String])
sexpr ("(":xs)     = (e,xs') where (e,xs') = sexpr_List xs []
sexpr ("[":l:xs)   = (Thing l x,xs') where (x,"]":xs') = sexpr xs
sexpr ("{nil}":xs) = (Nil,    xs)
sexpr (x:xs)       = (Atom x, xs)
sexpr []           = error "sexpr: End of input"

sexpr_List :: [String] -> [SExpr] -> (SExpr,[String])
--sexpr_List xs       r = sexpr_List' xs' (e:r) where (e,xs') = sexpr xs 
sexpr_List xs r = case sexpr xs of { (e,xs') -> sexpr_List' xs' (e:r) }

sexpr_List' :: [String] -> [SExpr] -> (SExpr,[String])
sexpr_List' (",":xs) r = sexpr_List xs r
sexpr_List' (")":xs) r = (foldr1 Cons (reverse r),xs)

pp :: SExpr -> Doc
--pp _ = undefined  -- use to test that all calls to pp are gone
pp Nil = text "nil"
pp (Cons x y) = parens (pp x <> comma <> pp y)
pp (Thing t x) = lbrack <> text t <+> pp x <> rbrack
pp (Atom x)   = text x

size :: Int -> SExpr -> Int
size i Nil         = i
size i (Cons x y)  = (size $! ((size $! (i+1)) x)) y 
size i (Thing t x) = (size $! (i+1)) x          
size i (Atom x)    = i + 1

t2 fn = do
  { cs <- readFile fn
--  ; print $ vmap pp $ parse $ tokenize cs -- leaks space
  ; mapM_ (print . pp) $ parse $ tokenize cs
  }

t2' fn = do
  { cs <- readFile fn
  ; print $ map (size 0) $ parse $ tokenize cs
  }

----------------------------------------------------------------
-- End
----------------------------------------------------------------
