-- Copyright (c) 2000 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module Schedule(
        schedule
        ) where

import AbstractUnits
import qualified PPUnits as PP
import EvalUnits

import PrettyUtils
import List(nub,find)
import ListUtils( chopList )
import Maybe(fromMaybe)

import IOExts(unsafePerformIO)
import qualified Graph as G

----------------------------------------------------------------
-- The scheduler
----------------------------------------------------------------

-- Improved version of cross to avoid combinatorial explosion
crossLe :: Int -> Depend' -> [(QualId,DependRelation,QualId)]
crossLe i ([],       r, bs      ) = []
crossLe i (as,       r, []      ) = []
crossLe i (as@(a:_), r, bs@(b:_)) = [ (a,r,x) | a <- as ] ++ [ (x,r,b) | b <- bs ]
 where
  x = mkQualId ("$$dep_" ++ show i)

schedule :: [Initializer'] -> [Finalizer'] -> [QualId] -> [Depend'] -> ([QualId],[QualId])
schedule is fs exports ds 
  | cyclic
  = error $ show
          $  text "Cycle exists:"
          $$ braces (hlist (text . show . unhash) $ filter (`elem` live) $ r)
          $$ explain (take 2 $ filter (`elem` live) $ r)
  | otherwise
  = (init,fini)
 where

  patch_ds = zipWith crossLe [0..] ds
  ds' = [ (hash x, r, hash y) | xs <- patch_ds, (x,r,y) <- xs ]
  is' = [ (hash x, map hash y) | (x,y) <- is ]
  fs' = [ (hash x, map hash y) | (x,y) <- fs ]

  exports'     = map hash exports
  export_node  = mkQualId "$$exports"
  export_node' = hash export_node

  live = calc_live is' fs' export_node' exports' ds'

  callGraph :: G.Graph
  callGraph = calc_cg export_node' exports' ds'

  g2 :: G.Graph
  g2 = calc_g2 is' fs' live ds' callGraph export_node' exports'

  -- now schedule the live code according to dependencies between
  -- main, initialisers and finalisers
  (cyclic,r) = topSort2 g2 live
  sched = filter (`elem` live) r

  (init',_:fini') = break (== export_node') sched
  init = map unhash init'
  fini = map unhash fini'

  unhash :: Data -> QualId
  unhash x = head [ y | y <- ys, hash y == x ]
   where
    is' = map fst is
    fs' = map fst fs
    ys  = export_node : is' ++ fs'

  -- inefficiently generate an explanation of the cycle if we can
  explain :: [Data] -> Doc
  explain _ = empty
{-  No explanation for now - the code doesn't actually work.
  explain (x:y:_) = path x y
   where
    path x y = text "Chains of dependencies:" 
                        <+> vmap (hlist (text . show . unhash)) (paths x y [])
    path x y = case paths x y [] of
               []    -> empty
               (p:_) -> text "Chain of dependencies:" 
                        <+> hlist (text . show . unhash) p
    paths :: Data -> Data -> [Data] -> [[Data]]
    paths x y seen 
      | x == y        = [[y]]
      | x `elem` seen = [[]] -- failure
      | otherwise     = concat [ map (x:) (paths m y (x:seen)) | m <- next x ]

    next :: Data -> [Data]
    next a =  [ x | (x,ys) <- is', a `elem` ys ]
           ++ [ x | (x,ys) <- fs', a `elem` ys ]
           ++ [ x | (x,Precedes,y) <- ds', y == a ]
           ++ [ y | (x,Calls,y) <- ds', x == a ]
-}

-- we only consider live code: main plus initialisers and finalisers
-- reachable from live code.
calc_live is fs export_node exports ds = export_node : really_live
 where
  really_live = filter isCool $ reachable2 g1 [export_node]

  isCool x = isInit x || isFini x

  is' = map fst is
  fs' = map fst fs

  isInit, isFini :: Data -> Bool
  isInit x = x `elem` is'
  isFini x = x `elem` fs'

  -- use this graph to find live code
  -- note that x < y dependencies do not keep code alive
  es1 =  [ (x,    i) | (i,xs) <- is, x <- xs ]
      ++ [ (x,    f) | (f,xs) <- fs, x <- xs ]
      ++ [ (export_node, export') | export' <- exports ]
      ++ [ (x,    y) | (x,Calls,y) <- ds ]

  g1 :: G.Graph
  g1 = buildG2 es1

calc_cg export_node exports ds = buildG2 callGraph'
 where
  callGraph' = scc1 ++ scc2 ++ precs ++ calls
  -- for purposes of the callgraph, we pretend that the single giant node
  -- calls the real nodes it represents.
  scc1 = [ (export_node, export') | export' <- exports ] 
  scc2 = [ (export', export_node) | export' <- exports ] 
  calls  = [ (x, y) | (x,Calls,y) <- ds ]
  -- and we treat x<y as y `Calls` x
  precs  = [ (y, x) | (x,Precedes,y) <- ds ]
    
calc_g2 is fs live ds callGraph export_node exports = buildG2 deps
 where
  -- use this graph to calculate schedule
  deps :: [(Data,Data)]
  deps =  concat [deps_i,deps_f,deps_p,deps_scc]
 
  deps_i, deps_f, deps_p, deps_scc :: [(Data,Data)]
  deps_i = [ (i, x) | x <- live, i <- inits x ]
  deps_f = [ (x, f) | x <- live, f <- finis x ]
 
  deps_p = nub [ (x', y') 
           | (x,Precedes,y) <- ds
           , x' <- filter (`elem` live) $ reachable2 callGraph [x]
           , y' <- live
           , y `elem` reachable2 callGraph [y']
           ]
 
  -- all exports are treated as a single node (by forming them into an scc)
  deps_scc = [ (export_node, r) | r <- exports ] 
          ++ [ (r, export_node) | r <- exports ] 
  
  inits, finis :: Data -> [Data]
  inits x = nub $ concatMap (initOf !) $ reachable2 callGraph [x]
  finis x = nub $ concatMap (finiOf !) $ reachable2 callGraph [x]

  initOf, finiOf :: HashTable Data [Data]
  initOf = accumHashTable $ ids ++ [ (x, i) | (i,xs) <- is, x <- xs ]
  finiOf = accumHashTable $ fds ++ [ (x, i) | (i,xs) <- fs, x <- xs ]
  
  ids, fds :: [(Data,Data)]
  ids = [ (x, i) | (i,Precedes,x) <- ds, isInit i ]
     ++ [ (x, i) | (x,Calls,i)    <- ds, isInit i ]
  fds = [ (x, f) | (x,Precedes,f) <- ds, isFini f ]
     ++ [ (x, f) | (f,Calls,x)    <- ds, isFini f ]
  
  isInit, isFini :: Data -> Bool
  isInit x = x `elem` is'
  isFini x = x `elem` fs'

  is' = map fst is
  fs' = map fst fs


----------------------------------------------------------------
-- Graph support
----------------------------------------------------------------

type Data = G.Data

reachable2 :: G.Graph -> [Data] -> [Data]
reachable2 g vs = unsafePerformIO $ do
  vs' <- toList vs
  G.fromList $ G.reachableBFS g $ vs'

toList :: [Data] -> IO G.List
toList xs = do
  l <- G.emptyList
  mapM_ (G.insert l) xs    -- not sure I want to use insert here
  return l

topSort2 :: G.Graph -> [Data] -> (Bool,[Data])
topSort2 g l = unsafePerformIO $ do
  l' <- toList l
  let (cyclic,r) = G.topSort2 g l' l'
  r' <- G.fromList r
  return (cyclic, r')

buildG2 :: [(Data,Data)] -> G.Graph
buildG2 es = unsafePerformIO $ do
  g <- G.emptyGraph
  mapM_ (\(x,y) -> G.addEdge g x y) es
  return g

g ! x = fromMaybe [] (hlookup x g)

----------------------------------------------------------------
-- End
----------------------------------------------------------------
