-- This file is copied from the Hugs98 distribution which is distributed
-- under this license:
-- 
-- The Hugs 98 system is Copyright (c) Mark P Jones, Alastair Reid, the
-- Yale Haskell Group, and the Oregon Graduate Institute of Science and
-- Technology, 1994-1999, All rights reserved, and is distributed as free
-- software under the following license.
-- 
-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions
-- are met:
-- 
-- - Redistributions of source code must retain the above copyright notice,
-- this list of conditions and the following disclaimer.
-- 
-- - Redistributions in binary form must reproduce the above copyright
-- notice, this list of conditions and the following disclaimer in the
-- documentation and/or other materials provided with the distribution.
-- 
-- - Neither name of the copyright holders nor the names of its
-- contributors may be used to endorse or promote products derived from
-- this software without specific prior written permission.
-- 
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
-- "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-- LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-- A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-- HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
-- INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
-- BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
-- OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-- ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
-- TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
-- USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-- 

module Sequence(
	Sequence( fromList, toList ),
	-- instances for [], Maybe and List
	List -- same instances as for []
	) where

import Maybe(maybeToList, listToMaybe)
import Monad

class (Functor m, MonadPlus m) => Sequence m where
  fromList :: [a] -> m a
  toList   :: m a -> [a]

  fromList = msum . fmap return

----------------------------------------------------------------
-- []: fast access to head but slow append
----------------------------------------------------------------

instance Sequence [] where
  fromList = id
  toList   = id

----------------------------------------------------------------
-- Maybe: single element lists - sort of
----------------------------------------------------------------

instance Sequence Maybe where
  toList   = maybeToList
  fromList = listToMaybe

----------------------------------------------------------------
-- List: lists with fast append (but slower indexing!)
----------------------------------------------------------------

-- Instead of providing Cons as a constructor, we provide Append.

data List a = Empty
             | Singleton a
             | Append (List a) (List a)

-- We define all the same instances that are defined for [].

-- The following definitions are independent of the choice of
-- representation since there's very little benefit in writing
-- representation-dependent versions.

instance Eq a => Eq (List a) where
  xs == ys = toList xs == toList ys

instance Ord a => Ord (List a) where
  compare xs ys = compare (toList xs) (toList ys)

instance Read a => Read (List a) where
  readsPrec p s = [ (fromList xs, r)      | (xs,  r) <- readsPrec p s ]
  readList    s = [ (map fromList xss, r) | (xss, r) <- readList    s ] 

instance Show a => Show (List a) where
  showsPrec p xs  = showsPrec p (toList xs)
  showList    xss = showList    (map toList xss)

-- The following operations are representation dependent and ought
-- to go much faster than any alternative way of writing them.
--
-- For example, the monadic operators preserve the structure of their
-- input.

instance Functor List where
  fmap f Empty          = Empty
  fmap f (Singleton x)  = Singleton (f x)
  fmap f (Append xs ys) = Append (fmap f xs) (fmap f ys)

instance Monad List where
  Empty        >>= k = Empty
  Singleton x  >>= k = k x
  Append xs ys >>= k = Append (xs >>= k) (ys >>= k)

  return = Singleton

instance MonadPlus List where
  mzero = Empty
  mplus = Append

instance Sequence List where
  fromList = foldr (\ x xs -> Singleton x `Append` xs) Empty

  toList xs = flatten xs []
   where
    -- flatten uses the standard technique of a "work list" yss
    -- flatten xs yss = xs ++ concatMap toList yss
    -- flatten :: List a -> [List a] -> [a]
    flatten Empty          []       = []
    flatten Empty          (ys:yss) = flatten ys yss
    flatten (Singleton x)  []       = [x]
    flatten (Singleton x)  (ys:yss) = x : flatten ys yss

    -- special cases for extra speed
    flatten (Append (Singleton x) ys) yss = x:flatten ys yss
    flatten (Append xs Empty) yss   = flatten xs yss
    flatten (Append Empty ys) yss   = flatten ys yss

    flatten (Append xs ys) yss      = flatten xs (ys:yss)
