-- Copyright (c) 2000, 2001 University of Utah and the Flux Group.
-- All rights reserved.
-- 
-- This file is part of the Knit component composition software.
-- 
-- Permission to use, copy, modify, distribute, and sell this software and
-- its documentation is hereby granted without fee, provided that the above
-- copyright notice and this permission/disclaimer notice is retained in all
-- copies or modified versions, and that both notices appear in supporting
-- documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
-- WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
-- ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
-- 
-- Users are requested, but not required, to send to csl-dist@cs.utah.edu any
-- improvements that they make and grant Univ. of Utah redistribution rights.

module Main( main ) where

import System( getArgs )
import Monad( when )
import List( sort, partition )

-- import PrettyUtils
import ParseUnits( readDefns )
import qualified PPUnits as PP
import AbstractUnits
import Html

import ParseC
import qualified PP1 as PPC
import qualified AbstractC as C
import Maybe(fromMaybe)
import MonadUtils( concatMapM )
import ListUtils( chopList )

main = do
  args <- getArgs
  let (env,args') = getEnv args
      unitDirs    = chopList (== ':') $ fromMaybe err $ lookup "UNIT_PATH" env
      err = error "You must specify UNIT_PATH=..."
      (dst_dir:unit_files) = args'
  when (length args' < 2) $
    error "Usage: knitdoc dst_dir unit_files"

  xs <- readDefns unitDirs unit_files
  let
    xs' = regroup xs
    units = [ (x, t, u) | (x, t, Unit u) <- xs' ]
    bts  = [ (x, t, s) | (x, t, BundleT s) <- xs' ]

    bts' :: [(BundleTIdent, BundleT')]
    bts' = [ (x, evalBT bts' bt) | (x, BundleT bt) <- xs ]

    fenv :: [(FlagsIdent, [String])]
    fenv = [ (x, evalFlags fenv f) | (x, Flags f) <- xs ]

  writeFile (dst_dir ++ "/index.html")$ renderHtml $ top (unwords unit_files)
  writeFile (dst_dir ++ "/blank.html")$ renderHtml $ blank
  mapM_ (write_unit dst_dir)    units
  mapM_ (write_bt  bts' fenv env dst_dir xs) bts
  writeFile (dst_dir ++ "/u_ix.html") $ renderHtml $ frameIt2 "Units"        $ punctuate br $ map unitIdent' $ sort $ map (fromIdent . fst3) units
  writeFile (dst_dir ++ "/s_ix.html") $ renderHtml $ frameIt2 "Bundle Types" $ punctuate br $ map bundleTIdent'  $ sort $ map (fromIdent . fst3) bts

write_unit dst_dir (nm, t, u) = do
  writeFile (dst_dir ++ "/unit_" ++ fromIdent nm ++ ".html") $ 
    renderHtml $ frameIt "unit" (fromIdent nm) $ unit (nm, t, u)

write_bt bts' fenv env dst_dir xs (nm, t, s) = do
  ds <- bte "knit_c_parser" env fenv s nm
  writeFile (dst_dir ++ "/bt_" ++ fromIdent nm ++ ".html") $ 
    renderHtml $ frameIt "bundletype" (fromIdent nm) 
               $   bt xs (nm, t, s) ds

regroup :: [(Ident, Defn)] -> [(Ident, [String], Defn)]
regroup = regroup' []
 where
  regroup' :: [String] -> [(Ident, Defn)] -> [(Ident, [String], Defn)]
  regroup' r [] = []
  regroup' r ((x,Doc s):xs) = regroup' (s:r) xs
  regroup' r ((x,t):xs) = (x, concatMap (chop []) $ reverse r, t) : regroup' [] xs

  chop s  ('\n':'\n':r) = reverse s : chop [] r
  chop s  (c:r) = chop (c:s) r
  chop [] [] = []
  chop s  [] = [reverse s]

top nm = header << thetitle << nm
  +++ frameset ! [cols "20%, *"] <<
      (   frameset ! [rows "50%, *"] <<
          (   frame ! [src "s_ix.html"] << ""
          +++ frame ! [src "u_ix.html" ] << ""
          )
      +++ frame ! [name "main_window",  src "blank.html" ] << ""
      )

blank
  =   header << thetitle << "Knit"
  +++ body ! [bgcolor white] << 
        bold << font![size "40", face "Helvetica"] << ""

frameIt :: (HTML a) => String -> String -> a -> Html
frameIt ty nm theBody
  =   header << thetitle << nm
  +++ body ! [bgcolor white] <<
        (   hdr ty nm
        +++ br
--      +++ hr
--      +++ hlinks pages
--      +++ hr
        +++ theBody
        +++ hr
--        +++ hlinks pages
--        +++ hr
        +++ table![width "100%"] <<
--              ( (td![align "left"] << 
--                 font![size "1"]   <<
--                 ("Copyright " +++ copyright +++ " 1999 " +++ alastair))
--              `beside`
                (
                (td![align "right"] <<
                 anchor ! [href url_htmllib] <<
                 font![size "1", color "008888"] <<
                 "Rendered using Haskell HTML Combinators"
               )
              )
         )

frameIt2 :: (HTML a) => String -> a -> Html
frameIt2 nm theBody
  =   header << thetitle << nm
  +++ body ! [bgcolor white] <<
        (   bold << font![size "15", face "Helvetica"] << nm
        +++ br
        +++ theBody
        )

hdr ty nm = table ! [cellpadding 0, cellspacing 0, border 0, bgcolor bluish] <<
        (td ! [align "left", width "0%"]  << 
           (table ! [cellpadding 5, cellspacing 0, border 0, bgcolor greenish] <<
            (td << bold << font![size "15", color white, face "Helvetica"] << ty)))
        `beside`
        (td ! [align "left", width "100%"]  << 
           (table ! [cellpadding 5, cellspacing 0, border 0, bgcolor bluish] <<
            (td << bold << font![size "15", face "Helvetica", color greenish] << nm)))

url_htmllib = "http://www.cse.ogi.edu/~andy/html/"
beige :: String
beige = "#FFF4E1"

-- colors used in the Hugs home pages
orange = "#ffaa88"
bluish = "#eeeeff"
greenish = "#638494"

-- ToDo: links to uses of units in compound units

unit (nm, t, (rs,ps,_,cs,is,fs,ds,mbflatten,body)) = p <<
    (   anchor ! [name ("unit_" ++ nm' ++ ".html")] << ""
    +++ map (paragraph.toHtml) t
    +++ br
    +++ bold << "unit " +++ big << toHtml nm' +++ " = {"
    +++ table << td << 
          (   bold << "imports" +++ "["
          +++ impsexps rs
          +++ "];"
          +++ br
          +++ bold << "exports" +++ "["
          +++ impsexps ps
          +++ "];"
          +++ br
          +++ "..."
        )
    +++ "}"
    )
 where
  nm' = fromIdent nm

bt xs (nm, t, s) ds = p <<
    (   anchor ! [name ("bt_" ++ nm' ++ ".html")] << ""
    +++ map (paragraph.toHtml) t
    +++ br
    +++ table << 
           (        (td ! [valign "top"] << (bold << "bundletype " +++ big << toHtml nm' +++ " = "))
           `beside` (td ! [valign "top"] << bundleType s)
           )
    +++ br +++ bold << "C types: " 
    +++ pre (toHtmlFromList [ show (PPC.ppDefn d) ++ "\n" | d <- ds])
    +++ br +++ bold << "Extended by: " 
    +++ table << td << punctuate ", " (map bundleTIdent' exts)
    +++ br +++ bold << "Exported by: " 
    +++ table << td << punctuate ", " (map unitIdent' exps)
    +++ br +++ bold << "Imported by: "        
    +++ table << td << punctuate ", " (map unitIdent' imps)
    )
 where
  exts = sort 
         [ fromIdent y
         | (y, BundleT es) <- xs
         , Extends e <- fst es
         , nm == e
         ]
  imps = sort 
         [ fromIdent y
         | (y, Unit (rs,ps,_,cs,is,fs,ds,mbflatten,body)) <- xs
         , (_, ([Extends x],_)) <- rs 
         , nm == x
         ]
  exps = sort 
         [ fromIdent y
         | (y, Unit (rs,ps,_,cs,is,fs,ds,mbflatten,body)) <- xs
         , (_, ([Extends x],_)) <- ps 
         , nm == x
         ]
  nm' = fromIdent nm

impsexps [] = []
impsexps xs = [table ! [cellpadding 0, border 0] << aboves (map impexp xs)]

impexp (i, s) =        (td ! [valign "top"] << bundleIdent i)
              `beside` (td ! [valign "top"] << " : ")
              `beside` (td ! [valign "top"] << bundleType s)

unitIdent  x = refident "unit" (show x)
bundleTIdent   x = refident "bt"  (show x)
unitIdent' x = refident "unit" x
bundleTIdent'  x = refident "bt"  x
bundleIdent x = thecode << show x
cIdent     x = thecode << show x
fileName   x = thecode << show x
refident what x = anchor ! [href (what ++ "_" ++ x ++ ".html"), target "main_window"] << x
ident    x   = show x
quotes x = "\"" ++ x ++ "\""

bundleType :: BundleType -> Html
bundleType = bundleType' . fst 

bundleType' [Extends x] = tag "span" (bold << bundleTIdent x)
bundleType' xs          = "{ " +++ punctuate br (map btElement xs) +++ " }"

btElement :: BTElement -> Html
btElement (Ident   x) = cIdent x
btElement (Include f) = "include " +++ fileName f
btElement (Extends s) = tag "span" (bold << "extends " +++ bundleTIdent s)

punctuate :: (HTML a) => a -> [Html] -> Html
punctuate p []     = noHtml
punctuate p [x]    = x
punctuate p (x:xs) = x +++ p +++ punctuate p xs

fst3 :: (a,b,c) -> a
fst3 (a,b,c) = a

type BundleT' = [MemberIdent]

evalBT :: [(BundleTIdent, BundleT')] -> BundleType -> BundleT'
evalBT env (bt,_) = concatMap (evalBTE env) bt

evalBTE :: [(BundleTIdent, BundleT')] -> BTElement -> BundleT'
evalBTE env (Ident x)   = [x]
evalBTE env (Include f) = []
evalBTE env (Extends i) = evalBTI env i

evalBTI :: [(BundleTIdent, BundleT')] -> BundleTIdent -> BundleT'
evalBTI env i = fromMaybe err (lookup i env)
 where
  err = error $ "Can't find bundleTIdent " ++ show (PP.bundleTIdent i)

evalFlags :: [(FlagsIdent, [String])] -> Flags -> [String]
evalFlags env flags = concatMap (evalFlag env) flags

evalFlag env (RawFlag str) = [str]
evalFlag env (UseFlags fi) = evalFI env fi

evalFI :: [(FlagsIdent, [String])] -> FlagsIdent -> [String]
evalFI env i = fromMaybe err (lookup i env)
 where
  err = error $ "Can't find flagsIdent " ++ show (PP.flagIdent i)

-- This extracts function prototypes from all the files included into
-- a bundletype using the flags as appropriate.
--
-- We use a gcc extension to achieve the effect of processing a single
-- C file containing:
--
--   #include "foo.h"
--   #include "bar.h"
--
-- by executing:
--
--   gcc -P -E -include foo.h -include bar.h -xc /dev/null
--
-- The more obvious: 
--
--   gcc -P -E foo.h bar.h
--
-- Processes each file separately so definitions in foo.h aren't available
-- in bar.h

bte :: String -> [(String,String)] -> [(FlagsIdent,[String])] -> BundleType -> BundleTIdent -> IO [C.Defn]
bte cparser env fenv (bt,fs) x
  | null incs
  = return []
  | otherwise
  = do
      xs <- parseDefnsEx cparser env' (evalFlags fenv fs ++ mkIncs incs) "/dev/null" (show x)
      return [ x | x <- xs, Just nm <- [nameOf x], unQualify nm `elem` ms ]
 where
  incs = [ x | Include x <- bt ]
  ms   = [ x | Ident   x <- bt ]
  env' = [ x ++ "=" ++ y | (x,y) <- env ]

  mkIncs fs = map ("-include " ++) fs ++ ["-xc"]

-- just good enough to find variable and function declarations/definitions
nameOf (C.Init ((nm,_),_)) = Just nm
nameOf (C.Fun  (nm,_) _ _) = Just nm
nameOf (C.Extern x)        = nameOf x
nameOf _                   = Nothing

getEnv :: [String] -> ([(String,String)],[String])
getEnv args = (map split env,rest)
 where
  (env, rest) = partition ('=' `elem`) args
  split x = (pre, tail post) where (pre,post) = break (== '=') x

----------------------------------------------------------------
-- End
----------------------------------------------------------------
