/*
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Knit component composition software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
/* 
 * This is based on Jeff Lee's grammar and lexer for the April 1985
 * draft of the ANSI C standard as posted to net.sources by
 * tps@sdchem.UUCP (Tom Stockfisch) on 3 Mar 87.
 * The original README is in gram.README.
 *
 * This has since been heavily modified to handle typedefs
 * differently, to support features of the current ANSI standard, to
 * support many gcc extensions and to generate an "sexpr" tree for
 * communicating with Knit.  
 */

%{
extern void registerTypename(char* nm);
static int in_typedef = 0; /* are we in a typedef? */

#include "sexpr.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
 
void yyerror(const char *); 
int yylex();
 
extern char *yytext;
extern int column;
extern int line;

/*
 * To my great surprise, adjacent string literals are not merged
 * in (gcc's) preprocessor - so we do it in the grammar instead.
 */
int lengthString(sexpr x)
{
	if (is_string(x)) {
		return strlen(stringOf(x));
	} else if (is_cons(x)) {
		return lengthString(car(x)) + lengthString(cdr(x));
	} else {
		fflush(stdout);
		printf("\nlengthString: invalid string object\n");
		exit(2);
	}
}

// just like strcpy except it returns the end of the string
// instead of totally useless information.
char* strcopy2(char *dst, const char*src)
{
	while ((*dst++ = *src++)) {
	}
	return dst-1;
}


char* copyString(char *dst, sexpr x)
{
	if (is_string(x)) {
		return strcopy2(dst,stringOf(x));
	} else if (is_cons(x)) {
		return copyString(copyString(dst,car(x)),cdr(x));
	} else {
		fflush(stdout);
		printf("\ncopyString: invalid string object\n");
		exit(2);
	}
}


sexpr concatStrings(sexpr x)
{
	int l = lengthString(x);
	char* s = (char*)malloc(l+1);
	assert(s);
	copyString(s,x);
	return fast_string(s);
}


#define YYSTYPE sexpr /* not sure about this */
%}

%token IDENTIFIER CONSTANT STRING_LITERAL SIZEOF
%token PTR_OP INC_OP DEC_OP LEFT_OP RIGHT_OP LE_OP GE_OP EQ_OP NE_OP
%token AND_OP OR_OP MUL_ASSIGN DIV_ASSIGN MOD_ASSIGN ADD_ASSIGN
%token SUB_ASSIGN LEFT_ASSIGN RIGHT_ASSIGN AND_ASSIGN
%token XOR_ASSIGN OR_ASSIGN TYPE_NAME

%token TYPEDEF EXTERN STATIC AUTO REGISTER
%token CHAR SHORT INT LONG SIGNED UNSIGNED FLOAT DOUBLE CONST VOLATILE VOID
%token STRUCT UNION ENUM ELIPSIS RANGE

%token CASE DEFAULT IF ELSE SWITCH WHILE DO FOR GOTO CONTINUE BREAK RETURN

%token ATTRIBUTE UNUSED ALIGNED TRANSPARENT_UNION MODE DI FORMAT PRINTF SCANF
%token   NORETURN CONST STDCALL
%token INLINE TYPEOF
%token ASM LABEL
%token EXTENSION

%start file
%%

primary_expr
        : identifier      { $$ = thing("var",$1); }
	| CONSTANT        { $$ = thing("const",other(yytext)); }
	| string  { $$ = thing("string",$1); }
	| '(' expr ')'    { $$ = thing("(_)",$2); } 
	| EXTENSION '(' compound_statement ')' { $$ = thing("stat_expr",$3); }
	| '(' compound_statement ')' { $$ = thing("stat_expr",$2); }
	;

postfix_expr
	: primary_expr    { $$ = $1; }
	| postfix_expr '[' expr ']' { $$ = thing("index",pair($1,$3)); }
	| postfix_expr '(' ')'      { $$ = thing("app",pair($1,nil)); }
	| postfix_expr '(' argument_expr_list ')'
                                    { $$ = thing("app",pair($1,reverse($3))); }
	| postfix_expr '.' identifier { $$ = thing(".",pair($1,$3)); }
	| postfix_expr PTR_OP identifier { $$ = thing("->",pair($1,$3)); }
	| postfix_expr INC_OP { $$ = thing("_++",$1); }
	| postfix_expr DEC_OP { $$ = thing("_--",$1); }
	;

argument_expr_list
	: assignment_expr { $$ = singleton($1); }
	| argument_expr_list ',' assignment_expr { $$ = cons($3,$1); }
	;

unary_expr
	: postfix_expr      
	| INC_OP unary_expr { $$ = thing("++_",$2); }
	| DEC_OP unary_expr { $$ = thing("--_",$2); }
	| unary_operator cast_expr { $$ = thing("unop",pair($1,$2)); }
	| SIZEOF unary_expr { $$ = thing("sizeof",$2); }
	| SIZEOF '(' type_name ')' { $$ = thing("sizeof_t",$3); }
	| AND_OP identifier { $$ = thing("label_addr",$2); }
	;

unary_operator
	: '&' { $$ = other("&"); }
	| '*' { $$ = other("*"); }
	| '+' { $$ = other("+"); }
	| '-' { $$ = other("-"); }
	| '~' { $$ = other("~"); }
	| '!' { $$ = other("!"); }
	;

cast_expr
	: unary_expr
	| cast cast_expr  { $$ = thing("cast",pair($1,$2)); }
        | cast '{' initializer_list '}'
                          { $$ = thing("constr",pair($1,reverse($3))); }
	;

cast : '(' type_name ')' { $$ = $2; }

multiplicative_expr
	: cast_expr
	| multiplicative_expr '*' cast_expr { $$ = thing("*",pair($1,$3)); }
	| multiplicative_expr '/' cast_expr { $$ = thing("/",pair($1,$3)); }
	| multiplicative_expr '%' cast_expr { $$ = thing("%",pair($1,$3)); }
	;

additive_expr
	: multiplicative_expr 
	| additive_expr '+' multiplicative_expr { $$ = thing("+",pair($1,$3)); }
	| additive_expr '-' multiplicative_expr { $$ = thing("-",pair($1,$3)); }
	;

shift_expr
	: additive_expr
	| shift_expr LEFT_OP additive_expr { $$ = thing("<<",pair($1,$3)); }
	| shift_expr RIGHT_OP additive_expr { $$ = thing(">>",pair($1,$3)); }
	;

relational_expr
	: shift_expr
	| relational_expr '<' shift_expr { $$ = thing("<",pair($1,$3)); }
	| relational_expr '>' shift_expr { $$ = thing(">",pair($1,$3)); }
	| relational_expr LE_OP shift_expr { $$ = thing("<=",pair($1,$3)); }
	| relational_expr GE_OP shift_expr { $$ = thing(">=",pair($1,$3)); }
	;

equality_expr
	: relational_expr
	| equality_expr EQ_OP relational_expr { $$ = thing("==",pair($1,$3)); }
	| equality_expr NE_OP relational_expr { $$ = thing("!=",pair($1,$3)); }
	;

and_expr
	: equality_expr
	| and_expr '&' equality_expr { $$ = thing("&",pair($1,$3)); }
	;

exclusive_or_expr
	: and_expr
	| exclusive_or_expr '^' and_expr { $$ = thing("^",pair($1,$3)); }
	;

inclusive_or_expr
	: exclusive_or_expr
	| inclusive_or_expr '|' exclusive_or_expr { $$ = thing("|",pair($1,$3)); }
	;

logical_and_expr
	: inclusive_or_expr
	| logical_and_expr AND_OP inclusive_or_expr { $$ = thing("&&",pair($1,$3)); }
	;

logical_or_expr
	: logical_and_expr
	| logical_or_expr OR_OP logical_and_expr { $$ = thing("||",pair($1,$3)); }
	;

conditional_expr
	: logical_or_expr
	| logical_or_expr '?' logical_or_expr ':' conditional_expr { $$ = thing("?:",triple($1,$3,$5)); }
	| logical_or_expr '?' ':' conditional_expr { $$ = thing("?nil:",pair($1,$4)); }
	;

assignment_expr
	: conditional_expr { $$ = $1; } 
	| unary_expr assignment_operator assignment_expr { $$ = thing("=",triple($1,$2,$3)); }
	;

assignment_operator
	: '='          { $$ = other("="); }
	| MUL_ASSIGN   { $$ = other("*="); }
	| DIV_ASSIGN   { $$ = other("/="); }
	| MOD_ASSIGN   { $$ = other("%="); }
	| ADD_ASSIGN   { $$ = other("+="); }
	| SUB_ASSIGN   { $$ = other("-="); }
	| LEFT_ASSIGN  { $$ = other("<<="); }
	| RIGHT_ASSIGN { $$ = other(">>="); }
	| AND_ASSIGN   { $$ = other("&="); }
	| XOR_ASSIGN   { $$ = other("^="); }
	| OR_ASSIGN    { $$ = other("|="); }
	;

expr
	: assignment_expr          { $$ = $1; }
	| expr ',' assignment_expr { $$ = thing(",",pair($1,$3)); }
	;

constant_expr
	: conditional_expr { $$ = $1; }
	;

declaration
	: declaration_specifiers ';'                      { $$ = pair($1,nil);        }
	| declaration_specifiers init_declarator_list attributes ';' 
          { if ($3 == nil) $$ = pair($1,reverse($2)); else $$ = pair(thing("attrs",cons($1,$3)),reverse($2)); }
	;

declaration_specifiers
	: storage_class_specifier    { $$ = singleton($1); }
	| storage_class_specifier declaration_specifiers { $$ = cons($1,$2); }
	| type_specifier attributes  { if ($2 == nil) $$ = $1; else $$ = thing("attrs",cons($1,$2)); }
	| type_specifier declaration_specifiers { $$ = cons($1,$2); }
	;

init_declarator_list
	: init_declarator                           { $$ = singleton($1); }
	| init_declarator_list ',' init_declarator  { $$ = cons($3,$1); }
	;

init_declarator
	: declarator                 { $$ = pair($1,nil); }
	| declarator '=' initializer { $$ = pair($1,$3); }
	;

storage_class_specifier
        : TYPEDEF   { in_typedef = 1; $$ = other("typedef"); }
	| EXTERN    { $$ = other("extern"); }
	| STATIC    { $$ = other("static"); }
	| AUTO      { $$ = other("auto"); }
	| REGISTER  { $$ = other("register"); }
	| INLINE    { $$ = other("__inline"); }
	;

type_specifier
        : CHAR     { $$ = other("char"); }
	| SHORT    { $$ = other("short"); }
	| INT      { $$ = other("int"); }
	| LONG     { $$ = other("long"); }
	| SIGNED   { $$ = other("signed"); }
	| UNSIGNED { $$ = other("unsigned"); }
	| FLOAT    { $$ = other("float"); }
	| DOUBLE   { $$ = other("double"); }
	| CONST    { $$ = other("const"); }
	| VOLATILE { $$ = other("volatile"); }
	| VOID     { $$ = other("void"); }
	| struct_or_union_specifier
	| enum_specifier
	| TYPEOF unary_expr        { $$ = thing("typeof",$2); }
 	| TYPE_NAME 
	;

attributes
        :                                          { $$ = nil; }
        | ATTRIBUTE '(' '(' attribute_list ')' ')' { $$ = reverse($4); }
        ;

attribute_list
        : attribute                    { $$ = singleton($1); }
        | attribute_list ',' attribute { $$ = cons($3,$1); }
        ;

attribute
        : IDENTIFIER '(' attribute_list ')' { $$ = thing("app",pair($1,reverse($3))); }
        | IDENTIFIER                     { $$ = $1; }
        | CONST                          { $$ = other("const"); }
	| CONSTANT                       { $$ = thing("const",other(yytext)); }
        ;

struct_or_union_specifier
	: struct_or_union identifier '{' struct_declaration_list '}'  { in_typedef++; $$ = triple($1,$2,reverse($4)); }
	| struct_or_union '{' struct_declaration_list '}'             { in_typedef++; $$ = triple($1,nil,reverse($3)); }
	| struct_or_union identifier                                  { in_typedef++; $$ = triple($1,$2,nil); }
	;

struct_or_union
	: STRUCT { in_typedef--; $$ = other("struct"); }
	| UNION  { in_typedef--; $$ = other("union"); }
	;

struct_declaration_list
	:                                            { $$ = nil; }
	| struct_declaration                         { $$ = singleton($1); }
	| struct_declaration_list struct_declaration { $$ = cons($2,$1);   }
	| struct_declaration_list ';' struct_declaration { $$ = cons($3,$1);   }
	;

struct_declaration
	: type_specifier_list attributes struct_declarator_list ';' { $$ = triple(reverse($1),reverse($2),reverse($3)); }
	;

struct_declarator_list
	: struct_declarator                             { $$ = singleton($1); }
	| struct_declarator_list ',' struct_declarator  { $$ = cons($3,$1);   }
	;

struct_declarator
	: declarator                       { $$ = pair($1,nil); }
	| ':' constant_expr                { $$ = pair(nil,$2); }
	| declarator ':' constant_expr     { $$ = pair($1,$3);  }
	;

enum_specifier
	: ENUM '{' enumerator_list optcomma '}'            { $$ = thing("enum",pair(nil,reverse($3))); }
	| ENUM identifier '{' enumerator_list optcomma '}' { $$ = thing("enum",pair($2,reverse($4))); }
	| ENUM identifier                         { $$ = thing("enum",pair($2,nil)); }
	;

enumerator_list
	: enumerator                     { $$ = singleton($1); }
	| enumerator_list ',' enumerator { $$ = cons($3,$1);   }
	;

optcomma
        : ',' { $$ = nil; } 
        |     { $$ = nil; }
        ;

enumerator
	: identifier                     { $$ = pair($1,nil); }
	| identifier '=' constant_expr   { $$ = pair($1,$3);  }
	;

declarator
	: declarator2          { $$ = $1; }
        | pointer attributes declarator2  { sexpr xx = $2 == nil ? $3 : thing("attrs",cons($2,$3));
                                            $$ = thing("ptr",pair($1,xx)); }
	;

declarator2
        : identifier                                               { if (in_typedef > 0) { registerTypename(otherOf($1)); } $$ = $1; }
	| '(' declarator ')'                  			   { $$ = $2; }
	| declarator2 '[' ']'                 			   { $$ = thing("[]",pair($1,nil)); }
	| declarator2 '[' constant_expr ']'   			   { $$ = thing("[]",pair($1,$3)); }
	| declarator2 '(' { in_typedef--; } parameters ')' attributes { in_typedef++; $$ = thing("fun",triple($1,$4,$6)); }

parameters
        :                           { $$ = thing("ptylist",cons(nil,nil)); }
        | parameter_type_list       { $$ = thing("ptylist",$1); }
	| parameter_identifier_list { $$ = thing("pidlist",$1); }
	;

pointer
	: '*'                               { $$ = other("*1"); }
	| '*' type_specifier_list           { $$ = thing("*2",reverse($2)); }
	| '*' pointer                       { $$ = thing("*3",$2); }
	| '*' type_specifier_list pointer   { $$ = thing("*4",pair(reverse($2),$3)); }
	;

type_specifier_list
	: type_specifier                      { $$ = singleton($1); }
	| type_specifier_list type_specifier  { $$ = cons($2,$1);   }
	;

parameter_identifier_list
	: identifier_list                     { $$ = pair(reverse($1),nil); }
	| identifier_list ',' ELIPSIS         { $$ = pair(reverse($1),other("...")); }
	;

identifier_list
	: identifier                          { $$ = singleton($1); }
	| identifier_list ',' identifier      { $$ = cons($3,$1);   }
	;

parameter_type_list 
	: parameter_list                      { $$ = pair(reverse($1),nil); }
	| parameter_list ',' ELIPSIS          { $$ = pair(reverse($1),other("...")); }
	;

parameter_list
	: parameter_declaration                    { $$ = singleton($1); }
	| REGISTER parameter_declaration           { $$ = singleton($2); /* hack */ }
	| parameter_list ',' parameter_declaration { $$ = cons($3,$1);   }
	| parameter_list ',' REGISTER parameter_declaration { $$ = cons($4,$1); /* hack */  }
	;

parameter_declaration
	: type_specifier_list declarator     { $$ = thing("x",pair(reverse($1),$2)); }
	| type_name                          { $$ = thing("y",$1); }
	;

type_name
	: type_specifier_list                     { $$ = pair(reverse($1),nil); }
	| type_specifier_list abstract_declarator { $$ = pair(reverse($1),$2);  }
	;

abstract_declarator
	: pointer                          { $$ = thing("abstract",pair($1,nil)); }
	| abstract_declarator2             { $$ = thing("abstract",pair(nil,$1)); }
	| pointer abstract_declarator2     { $$ = thing("abstract",pair($1,$2));  }
/*
        | pointer attributes abstract_declarator2  { sexpr xx = $2 == nil ? $3 : thing("attrs",cons($2,$3));
                                            $$ = thing("abstract",pair($1,xx)); }
*/
	;

abstract_declarator2
	: '(' abstract_declarator ')'      { $$ = $2; }
	| '[' ']'                          { $$ = thing("[]",pair(nil,nil)); }
	| '[' constant_expr ']'            { $$ = thing("[]",pair(nil,$2)); }
	| abstract_declarator2 '[' ']'     { $$ = thing("[]",pair($1,nil)); }
	| abstract_declarator2 '[' constant_expr ']' { $$ = thing("[]",pair($1,$3)); }
	| '(' ')'                          { $$ = thing("(fun')", pair(nil,pair(nil,nil))); }
	| '(' parameter_type_list ')'      { $$ = thing("(fun')", pair(nil,$2)); }
	| abstract_declarator2 '(' ')'     { $$ = thing("(fun')", pair($1,pair(nil,nil))); }
	| abstract_declarator2 '(' parameter_type_list ')' { $$ = thing("(fun')", pair($1,$3)); }
	;

initializer
	: '[' index ']' initializer        { $$ = thing("[]=", cons($2,$4)); }
/* first field should be label */
	| assignment_expr ':' initializer            { $$ = thing(":=",  cons($1,$3)); }
	| '.' identifier '=' initializer        { $$ = thing(":=",  cons($2,$4)); }
	| assignment_expr                  { $$ = $1; }
	| '{' '}'                          { $$ = thing("{_}", nil); }
	| '{' initializer_list '}'         { $$ = thing("{_}", reverse($2)); }
	| '{' initializer_list ',' '}'     { $$ = thing("{_}", reverse($2)); }
	;

/* It's not clear what's allowed here */
index
        : constant_expr { $$ = $1; }
        ;

initializer_list
	: initializer                      { $$ = singleton($1); }
	| initializer_list ',' initializer { $$ = cons($3,$1);   }
	;

statement
	: labeled_statement
	| compound_statement
	| expression_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	| ASM asm ';'             { $$ = thing("asm",cons(nil,$2)); }
	| ASM VOLATILE asm ';'	  { $$ = thing("asm",cons(other("T"),$3)); }
	;

/* four flavours with input args, output args, etc */
asm
	: '(' string ')'	       
                 	          { $$ = quad($2,nil,nil,nil); }
	| '(' string ':' asm_args ')'	       
                     	          { $$ = quad($2,$4,nil,nil); }
	| '(' string ':' asm_args ':' asm_args ')'	       
                                  { $$ = quad($2,$4,$6,nil); }
	| '(' string ':' asm_args ':' asm_args ':' asm_args ')'	       
                                  { $$ = quad($2,$4,$6,$8); }
	;

asm_args
	: asm_args2 { $$ = reverse($1); }
	|           { $$ = nil; }
	;
	
asm_args2
	: asm_args2 ',' asm_arg			   { $$ = cons($3,$1);   }
	| asm_arg				   { $$ = singleton($1); }
	;

asm_arg
	: string '(' expr ')'	{ $$ = pair($1,$3); }
        | string	        { $$ = pair($1,nil); }
	;

labeled_statement
	: identifier ':' statement          { $$ = thing("label", pair($1,$3)); }
	| identifier ':'                    { $$ = thing("label", pair($1,other("empty"))); }
	| CASE constant_expr ':' statement  { $$ = thing("case",  pair($2,$4)); }
	| CASE constant_expr ':'            { $$ = thing("case",  pair($2,other("empty"))); }
	| DEFAULT ':' statement             { $$ = thing("default", $3); }
	| DEFAULT ':'                       { $$ = thing("default", other("empty")); }
	;

/* gcc allows local labels before normal declarations */
compound_statement
	: '{' '}'                            	       	     { $$ = thing("{}",triple(nil,nil,nil)); }
	| '{' statement_list '}'            	             { $$ = thing("{}",triple(nil,nil,reverse($2))); }
	| '{' label_list statement_list '}'            	     { $$ = thing("{}",triple($2,nil,reverse($3))); }
	| '{' declaration_list '}'           	       	     { $$ = thing("{}",triple(nil,reverse($2),nil)); }
	| '{' declaration_list statement_list '}'            { $$ = thing("{}",triple(nil,reverse($2),reverse($3))); }
	| '{' label_list declaration_list statement_list '}' { $$ = thing("{}",triple($2,reverse($3),reverse($4))); }
	;

label_list
        : LABEL identifier_list ';' { $$ = reverse($2); }
        ;

declaration_list
	: declaration                             { $$ = singleton($1); } 
	| declaration_list declaration            { $$ = cons($2,$1); }   
	;

statement_list
	: statement                             { $$ = singleton($1); } 
	| statement_list statement              { $$ = cons($2,$1); }   
	;

expression_statement
	: ';'                    { $$ = other("empty"); }
	| expr ';'               { $$ = thing("expr",$1); }
	;

selection_statement
	: IF '(' expr ')' statement                { $$ = thing("if",triple($3,$5,other("empty"))); }
	| IF '(' expr ')' statement ELSE statement { $$ = thing("if",triple($3,$5,$7)); }
	| SWITCH '(' expr ')' statement            { $$ = thing("switch",pair($3,$5)); }
	;

iteration_statement
	: WHILE '(' expr ')' statement                 { $$ = thing("while",pair($3,$5)); }
	| DO statement WHILE '(' expr ')' ';'          { $$ = thing("do",pair($2,$5)); }
	| FOR '(' ';' ';' ')' statement                { $$ = thing("for",quad(nil,nil,nil,$6)); }
	| FOR '(' ';' ';' expr ')' statement           { $$ = thing("for",quad(nil,nil,$5,$7)); }
	| FOR '(' ';' expr ';' ')' statement           { $$ = thing("for",quad(nil,$4,nil,$7)); }
	| FOR '(' ';' expr ';' expr ')' statement      { $$ = thing("for",quad(nil,$4,$6,$8)); }
	| FOR '(' expr ';' ';' ')' statement           { $$ = thing("for",quad($3,nil,nil,$7)); }
	| FOR '(' expr ';' ';' expr ')' statement      { $$ = thing("for",quad($3,nil,$6,$8)); }
	| FOR '(' expr ';' expr ';' ')' statement      { $$ = thing("for",quad($3,$5,nil,$8)); }
	| FOR '(' expr ';' expr ';' expr ')' statement { $$ = thing("for",quad($3,$5,$7,$9)); }
	;

jump_statement
	: GOTO identifier ';'  { $$ = thing("goto",$2); }
	| CONTINUE ';'         { $$ = other("continue"); }
	| BREAK ';'            { $$ = other("break"); }
	| RETURN ';'           { $$ = thing("return",nil); }
	| RETURN expr ';'      { $$ = thing("return",$2); }
	;

file
	:                              { }
	| external_definition          { printf("\n"); up($1); printf("\n"); }
	| file external_definition     { printf("\n"); up($2); printf("\n"); }
	| file ';' external_definition { printf("\n"); up($3); printf("\n"); }
	;

external_definition
	: function_definition
	| declaration  { in_typedef = 0; }
        | ASM asm      { $$ = thing("asm",cons(nil,$2)); }
	;

function_definition
	: declarator function_body                        { $$ = thing("fun",triple(nil,$1,$2)); }
	| declaration_specifiers declarator function_body { $$ = thing("fun",triple($1,$2,$3)); }
	;

function_body
	: compound_statement                  { $$ = pair(nil,$1); }
        | declaration_list compound_statement { $$ = pair(reverse($1),$2); }
	;

identifier
        : IDENTIFIER
        | TYPE_NAME  
	;

string  : string1 { $$ = concatStrings($1); }
        ;

string1
        : string1 STRING_LITERAL { $$ = cons($1,$2); }
        | STRING_LITERAL         { $$ = $1; }
	;

%%

void yyerror(const char *s)
{
	fflush(stdout);
	printf("\n%*s\n%*s\non line %d", column, "^", column, s, line);
	exit(1);
}
