/*
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Knit component composition software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

/* 
 * Simple graph algorithms 
 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "graph.h"

/*
 * Automatically resizing vector/matrix representations
 */

struct vector {
        int  magic; // serious bug-hunting effort...
        int  refs;  // how many pointers are there to vector
        int  size;
        data *data;
};

#define v_magic 0x12345678

int isVector(vector v)
{
        return v != NULL && v->magic == v_magic && v->refs > 0;
}

void freeV(vector v)
{
        assert(isVector(v));
//        printf("freeV: %p\n",v);

        if (--v->refs) return;
        v->magic = 0;
        if (v->data) {
                free(v->data);
        }
        free(v);
}

void addRefV(vector v)
{
        assert(isVector(v));
        ++v->refs;
}

vector emptyV(void)
{
        vector v = (vector) malloc(sizeof(struct vector));
        assert(v != NULL);
        v->magic = v_magic;
        v->refs = 1;
        v->size = 0;
        v->data = NULL;
//        printf("emptyV: %p\n",v);
        return v;
}

void setV(vector v, int i, data d)
{
        assert(isVector(v));
        assert(i >= 0);
        if (i >= v->size) {
                int newsize = i+1;  // no need to allocate slop - realloc does a good job
                int j;
                v->data = (data*)realloc(v->data,sizeof(data) * newsize);
                assert(v->data);
                for(j=v->size; j<newsize; ++j) {
                        v->data[j] = baddata;
                }
                v->size = newsize;
        }
        v->data[i] = d;
}

data getV(vector v, int i)
{
        assert(isVector(v));
        assert(i >= 0);
        return ((i < v->size) ? v->data[i] : baddata);
}

/* search unsorted vector for element */
int elemV(vector v, data d)
{
        int i;
        assert(isVector(v));
        for(i=0; i<v->size; ++i) {
                if (v->data[i] == d) {
                        return 1;
                }
        }
        return 0;
}

/* insert element at end of vector */
void appendV(vector v, data d)
{
        assert(isVector(v));
        setV(v,v->size,d);
}

int sizeV(vector v)
{
        assert(isVector(v));
        return v->size;
}

vector intersectV(vector x, vector y)
{
        vector r = emptyV();
        int i;

        assert(isVector(x));
        assert(isVector(y));
        for(i=0; i<x->size; ++i) {
                if (elemV(y,getV(x,i))) {
                        appendV(r,getV(x,i));
                }
        }
        return r;
}

struct matrix {
        int  magic; // serious bug-hunting effort...
        int  refs;  // how many pointers are there to matrix
        int size;
        vector* data; // if m->data, then m->data[0..m->size-1] != NULL
};

#define m_magic 0x87654321

int isMatrix(matrix m)
{
        return m != NULL && m->magic == m_magic && m->refs > 0;
}

void freeM(matrix m)
{
        assert(isMatrix(m));
//        printf("freeM: %p\n",m);

        if (--m->refs) return;
        m->magic = 0;
        if (m->data) {
                int i;
                for(i=0; i<m->size; ++i) {
                        assert(m->data[i] != NULL);
                        freeV(m->data[i]);
                }
                free(m->data);
        }
        free(m);
}

void addRefM(matrix m)
{
        assert(isMatrix(m));
        ++m->refs;
}

matrix emptyM(void)
{
        matrix m = (matrix) malloc(sizeof(struct matrix));
        assert(m);
        m->magic = m_magic;
        m->refs = 1;
        m->size = 0;
        m->data = NULL;
//        printf("emptyM: %p\n",m);
        return m;
}

int sizeM(matrix m)
{
        assert(isMatrix(m));
        return m->size;
}

void setM(matrix m, int i, vector v)
{
        assert(isMatrix(m));
        assert(i >= 0);
        assert(isVector(v));

//        printf("setM: %p %d %p\n",m,i,v);
        if (i >= m->size) {
                int newsize = i + 1;   // no need to allocate slop - realloc does a good job
                int j;
                m->data = (vector*)realloc(m->data,sizeof(vector) * newsize);
                assert(m->data);
                for(j=m->size; j<newsize; ++j) {
                        m->data[j] = emptyV();
                }
                m->size = newsize;
                assert(m->data);
        }
        freeV(m->data[i]);
        m->data[i] = v;
        addRefV(v);
}

vector getM(matrix m, int i)
{
        vector v;
        assert(isMatrix(m));
        assert(i >= 0);
        if (i >= m->size) {
                setM(m,i,emptyV());
        }
        assert(i < m->size);
        v = m->data[i];
//        printf("getM: %p %d %p\n",m,i,v);
        return v;
}

void appendM(matrix m, vector v)
{
        assert(isMatrix(m));
        setM(m, m->size, v);
}

data get(matrix m, int i, int j)
{
        return getV(getM(m,i),j);
}

void set(matrix m, int i, int j, data d)
{
        assert(isMatrix(m));
        if (i > m->size) {
                setM(m,i,emptyV());
        }
        setV(getM(m,i),j,d);
}

void add(matrix m, int i, int j)
{
        if (!elemV(getM(m,i),j)) {
                appendV(getM(m,i),j);
        }
}

void printV(vector v)
{
        int i;
        printf("{");
        for(i=0; i<v->size; ++i) {
                if (i > 0) {
                        printf(",");
                }
                printf("%d", getV(v,i));
        }
        printf("}");
}

void printM(matrix m)
{
        int i;
        for(i=0; i<m->size; ++i) {
                printf("%d", i);
                printV(getM(m,i));
                printf("\n");
        }
}

/*
 * trees and forests
 */

struct tree {
        data   data;
        forest forest;
};

struct forest {
        int  size;
        tree* data;
};

extern void freeF(forest f);
void freeT(tree t)
{
        assert(t != NULL);
        freeF(t->forest);
        free(t);
}

void freeF(forest f)
{
        assert(f != NULL);
        if (f->data) {
                int i;
                for(i=0; i<f->size; ++i) {
                        assert(f->data[i] != NULL);
                        freeT(f->data[i]);
                }
                free(f->data);
        }
        free(f);
}

forest emptyF(void)
{
        forest f = (forest) malloc(sizeof(struct forest));
        assert(f != NULL);
        f->size = 0;
        f->data = NULL;
        return f;
}

tree mkT(data d, forest f)
{ 
        tree t = (tree) malloc(sizeof(struct tree));
        assert(t != NULL);
        t->data = d;
        t->forest = f;
        return t;
}

int isSingletonT(tree t)
{
        assert(t != NULL);
        return isEmptyF(t->forest);
}        

int isEmptyF(forest f)
{
        assert(f != NULL);
        return (f->size == 0);
}

void appendF(forest f, tree t)
{
        int newsize;
        assert(f != NULL);
        newsize = f->size+1;  // no need to allocate slop - realloc does a good job
        f->data = (tree*)realloc(f->data,sizeof(tree) * newsize);
        assert(f->data);
        f->data[f->size] = t;
        f->size = newsize;
}

extern void printF(forest f);
void printT(tree t)
{
        assert(t != NULL);
        printf("%d", t->data);
        printF(t->forest);
}

void printF(forest f)
{
        int i;
        assert(f != NULL);
        printf("{");
        for(i=0; i<f->size; ++i) {
                if (i > 0) {
                        printf(",");
                }
                printT(f->data[i]);
        }
        printf("}");
}

/*
 * reachableBFS
 *
 * Returns list of elements reachable from v using m as an
 * adjacency matrix (ie assuming that m[i] is a list of 
 * things reachable from i).
 *
 * Uses breadth-first search
 */

vector reachableBFS(matrix m, vector v)
{
        vector r = emptyV();  // visited list
        matrix worklist = emptyM();
        int j = 0;
        appendM(worklist,v);

        for(j=0; j<sizeM(worklist); ++j) {
                int i;
                vector w = getM(worklist,j);
                for(i=0; i < sizeV(w); ++i) {
                        data x = getV(w,i);
                        if (!elemV(r,x)) {
                                appendV(r,x);
                                appendM(worklist,getM(m,x));
                        }                        
                }
        }
        freeM(worklist);
        return r;
}

/*
 * reachableDFS
 *
 * Find list of elements reachable from v using m as an
 * adjacency matrix (ie assuming that m[i] is a list of 
 * things reachable from i).  Updates vector r in place.
 *
 * Uses depth-first search
 */

void reachableDFS(matrix m, vector r, vector v)
{
        int i;
        for(i=0; i < v->size; ++i) {
                data x = getV(v,i);
                if (!elemV(r,x)) {
                        appendV(r,x);
                        reachableDFS(m,r,getM(m,x));
                }                        
        }
}

/*
 * Transpose
 */
matrix transpose(matrix m)
{
        matrix r = emptyM();
        int i;

        assert(m);
        for(i=0; i < m->size; ++i) {
                vector v = getM(m,i);
                int j;
                for(j=0; j<v->size; ++j) {
                        appendV(getM(r,getV(v,j)),i);
                }
        }
        return r;
}

/*
 * Depth First Forest
 */
forest dfs_aux(matrix m, vector v, vector seen)
{
        forest r = emptyF();
        int i;
        for(i=0; i<v->size; ++i) {
                data x = getV(v,i);
                if (!elemV(seen,x)) {
                        tree t;
                        appendV(seen,x);
                        t = mkT(x,dfs_aux(m,getM(m,x),seen));
                        appendF(r,t);
                }
        }
        return r;                        
}

forest dfs(matrix m, vector v)
{
        vector seen = emptyV();
        return dfs_aux(m,v,seen);
}

void reverseF(forest f)
{
        int i, j;
        assert(f != NULL);
        for(i = 0, j=f->size-1; i < j; ++i, --j) {
                tree tmp;
                assert(i+j == f->size-1);
                tmp = f->data[i];
                f->data[i] = f->data[j];
                f->data[j] = tmp;
        }
}

void reverseV(vector v)
{
        int i, j;
        assert(isVector(v));
        for(i = 0, j=v->size-1; i < j; ++i, --j) {
                data tmp;
                assert(i+j == v->size-1);
                tmp = v->data[i];
                v->data[i] = v->data[j];
                v->data[j] = tmp;
        }
}

extern void postorderF(vector v, forest f);
void postorderT(vector v, tree t)
{
        assert(t != NULL);
        postorderF(v, t->forest);
        appendV(v, t->data);
}

void postorderF(vector v, forest f)
{
        int i;
        assert(f != NULL);
        for(i=0; i<f->size; ++i) {
                postorderT(v, f->data[i]);
        }
}

/*
 * topSort
 *
 * Sort list v according to dependency matrix m (ie assuming that m[i]
 * is a list of things greater than i and contains no duplicates).
 */

vector topSort(matrix m, vector v)
{
        vector r = emptyV();
        forest f = dfs(m,v);
        // printF(f); printf("\n");
        postorderF(r,f);
        reverseV(r);
        return r;
}

/*
 * topSort2
 *
 * if acyclic, returns topological sort
 * if cyclic, returns first cycle
 */
vector topSort2(matrix m, vector v, vector interesting, int *cyclic)
{
        vector r = emptyV();
        forest f = dfs(m,v);
        // printF(f); printf("\n");
        postorderF(r,f);
        reverseV(r);
        {
                matrix tm = transpose(m);
                forest cycles = dfs(tm,r);
                int i;
                for(i=0; i<cycles->size; ++i) {
                        /* 
                         * We look for cycles with more than 1 interesting node
                         * in them.
                         */
                        if (!isSingletonT(cycles->data[i])) {
                                vector r1, r2;
                                r1 = emptyV();
                                postorderT(r1,cycles->data[i]);
                                r2 = intersectV(r1,interesting);
                                if (sizeV(r2) > 1) {
                                        *cyclic = 1;
                                        freeV(r);
                                        freeV(r1);
                                        return r2;
                                }
                                freeV(r1);
                                freeV(r2);
                        }
                }
        }
        *cyclic = 0;
        return r;
}

#ifdef TESTMAIN /* for testing purposes */
int
main(int argc, char** argv)
{
        matrix m = emptyM();
        vector v = emptyV();
        vector r = NULL;

        /* 
         * Read a bunch of lines of the form x : y, 
         * build adjacency matrix with result 
         */
        {
                int x;
                int y;
                while (!feof(stdin) && fscanf(stdin, " %d : %d ",&x,&y)) {
                        add(m,x,y);
                }
        }

        /*
         * Let's see what we have.
         */
        printM(m);

        /* 
         * Topological sort
         */
        appendV(v,0);  // use 0 as root
        printf("Sorting\n");
        r = topSort(m,v);
        printV(r);
        printf("\n");

        return 0;
}
#endif /* TESTMAIN */
