/*
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Knit component composition software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <libelf/libelf.h>
#include <stdlib.h>

void panic(char* x) 
{
        printf("panic: %s\n", x);
        exit(1);
}

struct assoclist
{
  const char *name;
  const char *val;
  struct assoclist *next;
};

struct assoclist *rn_list = NULL;

static char* stringCopy(const char* x)
{
      char *copy;
      int len = strlen(x);
      copy = malloc (len + 1);
      strncpy (copy, x, len);
      copy[len] = '\0';
      return copy;
}

static void add_assoc_symbol(const char *, const char *, struct assoclist **);
static const char* find_assoc_symbol(const char *, struct assoclist *);

/* Add a symbol to strip_specific_list.  */

static void 
add_assoc_symbol (name, val, list)
     const char *name;
     const char *val;
     struct assoclist **list;
{
  struct assoclist *tmp_list;

  tmp_list = (struct assoclist *) malloc (sizeof (struct assoclist));
  tmp_list->name = name;
  tmp_list->val  = val;
  tmp_list->next = *list;
  *list = tmp_list;
}

/* See whether a symbol should be stripped or kept based on
   strip_specific_list and keep_symbols.  */

static const char*
find_assoc_symbol (name, list)
     const char *name;
     struct assoclist *list;
{
  struct assoclist *tmp_list;

  for (tmp_list = list; tmp_list; tmp_list = tmp_list->next)
    {
      if (strcmp (name, tmp_list->name) == 0)
	return tmp_list->val;
    }
  return NULL;
}

char* prefix; // prefix to apply to symbols

static char* sh_strings; // section header strings

static char* get_strings(Elf *elf, int i)
{
        Elf32_Shdr *shdr;
        Elf_Scn  *scn; // need this to interpret strings
        Elf_Data *data;

        scn = elf_getscn(elf, i);
        shdr = elf32_getshdr(scn);
        if (shdr->sh_type != SHT_STRTAB) {
                panic("No string table\n");
        }
        data = 0;
        if ((data = elf_getdata(scn, data)) == 0 || data->d_size == 0) {
                panic("Missing/empty string table\n");
        }
        return data->d_buf;
}

void set_strings(Elf *elf, int i,char* strings,int strsize)
{
        Elf32_Shdr *shdr;
        Elf_Scn  *scn; // need this to interpret strings
        Elf_Data *data;

        scn = elf_getscn(elf, i);
        shdr = elf32_getshdr(scn);
        if (shdr->sh_type != SHT_STRTAB) {
                panic("No string table\n");
        }
        data = 0;
        if ((data = elf_getdata(scn, data)) == 0) {
                panic("Missing string table\n");
        }
        data->d_buf = strings;
        data->d_size = strsize;
        elf_flagdata(data, ELF_C_SET, ELF_F_DIRTY);
}

Elf_Scn *find_scn(Elf* elf, char* name)
{
        Elf_Scn  *scn;
        scn = 0;
        while ((scn = elf_nextscn(elf, scn)) != 0) {
                Elf32_Shdr *shdr = elf32_getshdr(scn);
                char* scn_name = sh_strings + shdr->sh_name;
                if (strcmp(scn_name,name) == 0) return scn;
        }
        return 0;
}

#if 0
void print_sections(Elf *elf)
{
        Elf_Scn  *scn;
        scn = 0;
        while ((scn = elf_nextscn(elf, scn)) != 0) {
                Elf32_Shdr *shdr = elf32_getshdr(scn);
                printf("Section: '%s'\n", sh_strings + shdr->sh_name);
        }
}

void print_strtab(Elf_Scn* scn)
{
        Elf32_Shdr *shdr;
        Elf_Data *data;

        shdr = elf32_getshdr(scn);
        if (shdr->sh_type != SHT_STRTAB) {
                panic("Not a string table");
        }

        data = 0;
        if ((data = elf_getdata(scn, data)) == 0 || data->d_size == 0) {
                panic("Missing/empty string table");
        }
        {
                int i = 0;
                while (i <= data->d_size) {
                        printf("%s\n", data->d_buf + i);
                        i += strlen(data->d_buf + i) + 1;
                }
        }
}

void print_symtab(Elf *elf, Elf_Scn* scn)
{
        Elf32_Shdr *shdr;
        Elf_Data *data;
        char* strings;

        shdr = elf32_getshdr(scn);
        if (shdr->sh_type != SHT_SYMTAB) {
                panic("Not a symbol table");
        }
        strings = get_strings(elf,shdr->sh_link);

        data = 0;
        if ((data = elf_getdata(scn, data)) == 0 || data->d_size == 0) {
                panic("Missing/empty symbol table");
        }
        {
                Elf32_Sym *syms = (Elf32_Sym*)data->d_buf;
                int n_syms = data->d_size / sizeof(Elf32_Sym);
                int i = 0;
                for( i = 0; i < n_syms; ++i ) {
                        if (ELF32_ST_BIND(syms[i].st_info) == STB_LOCAL) {
                                printf("Local ");
                        } else {
                                printf("Global ");
                        }
                        printf("Symbol: '%s'\n", 
                               strings + syms[i].st_name
                                );
                }
        }
}
#endif /* 0 */

const char* rename_sym(char* name)
{
        const char* foundname = find_assoc_symbol(name,rn_list);
        if (foundname) {
                return foundname;
        } else {
                int newsize = strlen(name) + strlen(prefix);
                char* newname = (char*)malloc(newsize+1);
                strcpy(newname,prefix);
                strcpy(newname + strlen(prefix),name);
                return newname;
        }
}

const char* rename_symbol(Elf32_Sym *sym, char* name)
{
        if (ELF32_ST_BIND(sym->st_info) == STB_LOCAL) { // Don't rename locals
                return name;
        } else {
                return rename_sym(name);
        }
}

void mangle_symtab(Elf *elf, Elf_Scn* scn)
{
        Elf32_Shdr *shdr;
        Elf_Data *data;
        char* strings;

        char* newstrings = NULL;
        int newstr_size = 0;

        shdr = elf32_getshdr(scn);
        if (shdr->sh_type != SHT_SYMTAB) {
                panic("Not a symbol table");
        }
        strings = get_strings(elf,shdr->sh_link);

        data = 0;
        if ((data = elf_getdata(scn, data)) == 0 || data->d_size == 0) {
                panic("Missing/empty symbol table");
        }
        {
                Elf32_Sym *syms = (Elf32_Sym*)data->d_buf;
                int n_syms = data->d_size / sizeof(Elf32_Sym);
                int i = 0;
                for( i = 0; i < n_syms; ++i ) {
                        char* name = strings + syms[i].st_name;
                        const char* newname = rename_symbol(&syms[i], name);
                        int new_size = strlen(newname);

                        syms[i].st_name = newstr_size;
                        newstrings = (char*)realloc(newstrings,newstr_size + new_size+1);
                        strcpy(newstrings+newstr_size,newname);
                        newstr_size += strlen(newname);
                        newstr_size += 1;
                        // Don't free - it might be from renaming list
                        // if (newname != name) free(newname);
                }
                set_strings(elf,shdr->sh_link,newstrings,newstr_size);
        }
}

/*
   The following comment from bfd/stabs.c is useful for understanding a
   little of this code:

   Stabs entries use a 12 byte format:
     4 byte string table index
     1 byte stab type
     1 byte stab other field
     2 byte stab desc field
     4 byte stab value
   FIXME: This will have to change for a 64 bit object format.

   The stabs symbols are divided into compilation units.  For the
   first entry in each unit, the type of 0, the value is the length of
   the string table for this unit, and the desc field is the number of
   stabs symbols for this unit.

*/

// Note: endianness should not be an issue except when cross-compiling
typedef struct {
        Elf32_Word    n_name;  // stab string table index (see sh_link)
        unsigned char n_type;
        unsigned char n_other;
        Elf32_Half    n_desc;
        Elf32_Word    n_value;
} Stab;

// All these stab types have string pointers in their n_value field 
// so we have to watch out for them
#define N_GSYM  0x20
#define N_FUN   0x24
#define N_STSYM 0x26
#define N_LCSYM 0x28
#define N_OPT   0x3c
#define N_RSYM  0x40
#define N_SLINE 0x44
#define N_SO    0x64
#define N_LSYM  0x80
#define N_BINCL 0x82
#define N_SOL   0x84
#define N_PSYM  0xa0
#define N_EINCL 0xa2
#define N_LBRAC 0xc0
#define N_RBRAC 0xe0

static char* rename_stab(Stab *stab, char *name)
{
        char* colon = strchr(name,':');
        const char* new_name;
        int len = strlen(name);
        int info_len;
        int name_len;
        int new_len;
        char *copy, *cpy;

        if (colon == NULL) {
                return name;
        }

        // sleazy hack... set ':' to '\0', do lookup then switch back
        *colon = '\0';
        name_len = strlen(name);
        info_len = len - (name_len + 1);
        new_name = rename_sym(name);
        new_len = strlen(new_name) + 1 + info_len;
        cpy = copy = malloc (new_len + 1);

        strcpy(cpy,new_name); cpy += strlen(new_name);
        *cpy = ':'; cpy += 1;
        strcpy (cpy, colon+1);

        *colon = ':';

        return copy;
}

void rename_stabs(Elf *elf, Elf_Scn* scn)
{
        Elf32_Shdr *shdr;
        Elf_Data *data;
        char* strings;
        int offset = 0;
        int next_offset = 0;

        char* ostrings  = 0;
        int   ostr_size = 1;
        int last_table = -1;

        if (scn == 0) return; // That was easy...

        ostrings = malloc(1);
        ostrings[0] = '\0';

        shdr = elf32_getshdr(scn);
#if 0 // not one of the recognised formats
        if (shdr->sh_type != SHT_SYMTAB) {
                panic("Not a symbol table");
        }
#endif
        strings = get_strings(elf,shdr->sh_link);

        data = 0;
        if ((data = elf_getdata(scn, data)) == 0 || data->d_size == 0) {
                panic("Missing/empty symbol table");
        }
        {
                Stab *stabs = (Stab*)data->d_buf;
                int n_stabs = data->d_size / sizeof(Stab);
                int i = 0;
                for( i = 0; i < n_stabs; ++i ) {
                        char* old_name = strings + offset + stabs[i].n_name;
                        char* new_name = old_name;
                        // Note: if the n_name field ends in '\\' then
                        // we really ought to tack on the n_name from
                        // the next stab entry.  In practice, I've never seen
                        // this happen so I'm going to ignore it.
                        // Presumably generated strings should be chopped
                        // in the same manner - but what is the length limit?

                        // It's not clear whether we should key off the
                        // type or off of the first character after the colon.
                        // For now, we key off the type.  If this is
                        // changed, we should record why the type isn't used.
                        switch (stabs[i].n_type) {
                        case 0:
                        {
                                offset += next_offset;
                                next_offset = stabs[i].n_value;
//                                printf("Num stab symbols: %d\n", stabs[i].n_desc);
#if 1 // more to do here
                                if (last_table != -1) {
                                        panic("Multi-part (partially linked?) .o files can't be handled yet");
                                        stabs[last_table].n_value = ostr_size;
                                }
#endif
                                last_table = i;
                                goto next; // continue;
                        }
                        case N_GSYM:
                        case N_FUN:
                        case N_STSYM:
                        case N_LCSYM:
                        {
                                new_name = rename_stab(&stabs[i],old_name);
//                                printf("Stab[%d]: '%s'", i, old_name);
//                                printf(" -> '%s'\n", new_name);
                                
                        }
                        break;
                        case N_OPT   :
                        case N_SLINE :
                        case N_SO    :
                        case N_LSYM  :
                        case N_RSYM  :
                        case N_PSYM  :
                        case N_BINCL :
                        case N_SOL   :
                        case N_EINCL :
                        case N_LBRAC :
                        case N_RBRAC :
                        {
#if 0
                                printf("Stab: 0x%x '%s'\n", stabs[i].n_type, strings + offset + stabs[i].n_name);
#endif
                        }
                        break;
                        default:
                        {
                                printf("Unrecognised stab type: 0x%x '%s'\n", stabs[i].n_type, strings + offset + stabs[i].n_name);                        
                        }
                        break;
                        }
                        if (new_name[0] == '\0') { 
                                // special case for empty string
                                stabs[i].n_name = 0;
                        } else {
                                stabs[i].n_name = ostr_size;
                                ostrings = realloc(ostrings,ostr_size + strlen(new_name)+1);
                                strcpy(ostrings+ostr_size, new_name);
                                ostr_size += strlen(new_name)+1;
                        }
                next:
                }
                if (last_table != -1) {
                        stabs[last_table].n_value = ostr_size;
                }
        }
        set_strings(elf,shdr->sh_link,ostrings,ostr_size);
        elf_flagdata(data, ELF_C_SET, ELF_F_DIRTY);
}
        
void read_renaming(char* fname)
{
        FILE *f;
        char sym[4000];
        f = fopen (fname, "r");
        if (f == NULL)
        {
//                fprintf (stderr, "%s: ", program_name);
                perror (fname);
                exit (1);
        }
        while (fgets(sym,sizeof(sym),f)) 
        {
                char *eq = strchr(sym,'=');
                char *end = sym + strlen(sym) - 1; 
                if (eq == NULL || *end != '\n') {
//                        fprintf (stderr, "%s: ", program_name);
                        fprintf (stderr, "Ill-formed renaming '%s'\n", sym);
                        exit (1);
                }
                *eq = '\0'; 
                *end = '\0'; // ignore \n at end
                add_assoc_symbol (stringCopy(sym), stringCopy(eq+1), &rn_list);
        }
        fclose (f);
}

int main(int argc, char** argv)
{
        Elf *arf;
        Elf_Cmd cmd;
        Elf *elf;
        int fildes;

        if (argc != 4) {
                panic("Usage: <rn> prefix rename_file object_file");
        }
        prefix = argv[1];
        read_renaming(argv[2]);

//        fildes = open(argv[3], O_RDWR|O_TRUNC|O_CREAT, 0666);
        fildes = open(argv[3], O_RDWR, 0666);
//        fildes = open(argv[3], O_RDONLY, 0666);
        if (fildes < 0) {
                panic("Can't open .o file");
        }

        // First we coordinate version numbers between the program 
        // (probably really libelf.h) and the (dynamic) library libelf.so
        // If we don't do this, elf_begin and friends will fail...
        if (elf_version(EV_CURRENT) == EV_NONE)
        {
                printf("Library out of date\n");
                exit(1);
        }

        cmd = ELF_C_RDWR;
        arf = elf_begin(fildes, cmd, (Elf*)0);
        while ((elf = elf_begin(fildes, cmd, arf)) != 0)
        {
                Elf32_Ehdr *ehdr;
                if ((ehdr = elf32_getehdr(elf)) != 0)
                {
                        sh_strings = get_strings(elf,(size_t)ehdr->e_shstrndx);

//                        print_sections(elf);
//                        print_strtab(find_scn(elf,".stabstr"));
//                        print_strtab(find_scn(elf,".strtab"));
//                        print_symtab(elf, find_scn(elf,".symtab"));
                        mangle_symtab(elf, find_scn(elf,".symtab"));
//                        print_symtab(elf, find_scn(elf,".symtab"));
                        rename_stabs(elf, find_scn(elf,".stab"));
                }
                cmd = elf_next(elf);
                elf_end(elf);
        }

        /* The following makes absolutely certain that the changes get written
         * by marking everything dirty
         */
        /* dirty ehdr ... */
        {
                Elf32_Ehdr *ehdr;
                ehdr = elf32_getehdr(arf);
                elf_flagehdr(arf, ELF_C_SET, ELF_F_DIRTY);
        }
        /* ensure that the memory image of the file is complete */
        elf_update(arf, ELF_C_NULL);
        elf_update(arf, ELF_C_WRITE);   /* update file */

        elf_end(arf);
        close(fildes);
        exit(0);
}
