/*
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Knit component composition software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "sexpr.h"

typedef enum { 
	UNUSED, 
	CONS, NIL, 
	STRING, 
	THING,
	OTHER 
} Tag;

struct unknown {
  Tag           tag;
};

struct cons {
  Tag           tag;
  union	sexpr *car;
  union	sexpr *cdr;
};

struct string {
  Tag   tag;
  char *chars;
};

struct thing {
  Tag   tag;
  char *chars;
  union	sexpr *args;
};

struct other {
  Tag   tag;
  char *chars;
};

union sexpr {
  struct unknown unknown;
  struct cons   cons;
  struct string string;
  struct thing  thing;
  struct other  other;
};

//typedef union sexpr *sexpr;

sexpr cons( sexpr the_car, sexpr the_cdr )
{
	sexpr r = malloc(sizeof(r->cons));
	assert(r);
	r->cons.tag = CONS;
	r->cons.car = the_car;
	r->cons.cdr = the_cdr;
	return r;
}

sexpr car( sexpr x ) 
{
	assert(x && x->unknown.tag == CONS);
	return x->cons.car;
}

sexpr cdr( sexpr x ) 
{
	assert(x && x->unknown.tag == CONS);
	return x->cons.cdr;
}

int is_cons( sexpr x ) 
{
	return x && x->unknown.tag == CONS;
}

static char* stringCopy(char* s)
{
	char* r = malloc(1 + strlen(s));
	assert(r);
	strcpy(r,s);
	return r;
}

sexpr string( char* s )
{
	sexpr r = malloc(sizeof(r->string));
	assert(r);
	r->string.tag   = STRING;
	r->string.chars = stringCopy(s);
	return r;
}

sexpr fast_string( char* s )
{
	sexpr r = malloc(sizeof(r->string));
	assert(r);
	r->string.tag   = STRING;
	r->string.chars = s;
	return r;
}

char* stringOf( sexpr x ) 
{
	assert(x && x->unknown.tag == STRING);
	return x->string.chars;
}

int is_string( sexpr x ) 
{
	return x && x->unknown.tag == STRING;
}

sexpr thing( char* s, sexpr args )
{
	sexpr r = malloc(sizeof(r->thing));
	assert(r);
	r->thing.tag   = THING;
	r->thing.chars = stringCopy(s);
	r->thing.args  = args;
	return r;
}

char* thingOf( sexpr x ) 
{
	assert(x && x->unknown.tag == THING);
	return x->thing.chars;
}

sexpr argsOf( sexpr x ) 
{
	assert(x && x->unknown.tag == THING);
	return x->thing.args;
}

sexpr other( char* s )
{
	sexpr r = malloc(sizeof(r->other));
	assert(r);
	r->other.tag   = OTHER;
	r->other.chars = stringCopy(s);
	return r;
}

char* otherOf( sexpr x ) 
{
	assert(x && x->unknown.tag == OTHER);
	return x->other.chars;
}

void release( sexpr x ) 
{
	if (!x) return;
	switch (x->unknown.tag) {
	case CONS:
		release(x->cons.car);
		release(x->cons.cdr);
		break;
	case STRING:
		free(x->string.chars);
		break;
	case OTHER:
		free(x->other.chars);
		break;
	default:
		fprintf(stderr, "sexpr::release %d\n", x->unknown.tag);
		exit(1);
	}
	free(x);
}

// The pretty-printer puts a space between every token - to keep
// things even easier than they were already.
void pp( sexpr x ) 
{
	if (!x) {
		printf("nil ");
		return;
	}
	switch (x->unknown.tag) {
	case CONS:
		printf("( ");
		while (x && x->unknown.tag == CONS) {
		     pp(x->cons.car);
		     printf(" ");
		     x = x->cons.cdr;
                }
		if (x) {
			printf(". ");
			pp(x);
		}
		printf(") ");
		break;
	case STRING:
		printf("\"%s\" ", x->string.chars);
		break;
	case THING:
		printf("( ");
		printf("%s ", x->thing.chars);
		printf(" ");
		pp(x->thing.args);
		printf(") ");
		break;
	case OTHER:
		printf("%s ", x->other.chars);
		break;
	default:
		fprintf(stderr, "sexpr::pp %d\n", x->unknown.tag);
		exit(1);
	}
}

#include <ctype.h>

void str(char* s) 
{
	// putchar('"');
	for(; *s; ++s) {
		if (isspace(*s)) {
			// ugly hack to guard against possibility that
			// next character is a digit
			if (*s<8) printf("\\00%o", *s);
			else if (*s<64) printf("\\0%o", *s);
			else printf("\\%o", *s);
		} else {
			putchar(*s);
		}
	}
	// putchar('"');
}


// The ugly-printer does nothing fancy and puts a space between every
// token - to keep things even easier than they were already.
void up( sexpr x ) 
{
	if (!x) {
		/* The braces are to distinguish it from the identifier nil.
		 * This (ie the problem this avoids) showed up while testing
		 * the system on itself.
		 */
		printf("{nil} ");
		return;
	}
	switch (x->unknown.tag) {
	case CONS:
		printf("( ");
		/* This little bit of pretty-printing was added to cure a
		 * space leak in my sexpr parser when processing long
		 * initialisation lists.  Hack, hack.
		 */
		do {
			up(x->cons.car);
			printf(", ");
			x = x->cons.cdr;
		} while (x && x->unknown.tag == CONS);
		up(x);
		printf(") ");
		break;
	case STRING:
		printf("\"");
		str(x->string.chars);
		printf("\" ");
		break;
	case THING:
		printf("[ ");
		printf("%s ", x->thing.chars);
		printf(" ");
		up(x->thing.args);
		printf("] ");
		break;
	case OTHER:
		str(x->other.chars);
		printf(" ");
		break;
	default:
		fprintf(stderr, "sexpr::up %d\n", x->unknown.tag);
		exit(1);
	}
}

sexpr reverse( sexpr x )
{
	sexpr r = nil;
	while (x && x->unknown.tag == CONS) {
		sexpr t = x->cons.cdr;
		x->cons.cdr = r;
		r = x;
		x = t;
	}
	if (x) r = cons(x,r); // late night hack
	return r;
}
