/*
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this file
 * for any purpose with or without restriction is hereby granted.
 */

#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include "expr.h"
#include "number.h"
#include "token.h"

typedef struct expr_struct {
	expr_kind kind;
	union {
		const char *	err;		/* Error message. */
		token		tok;		/* Error token. */
		number		num;		/* Number value. */
		expr		unary_sub;	/* Unary subexpression. */
		struct {
			expr	left;
			expr	right;
		}		binary_sub;	/* Binary subexpressions. */
	} u;
} expr_struct;

/*****************************************************************************/

static expr_struct bad_arg;
static expr_struct bad_op;
static expr_struct no_mem;
static expr_struct no_child;
static expr_struct math_error;
static expr_struct div_by_zero;

int
init_exprs(void)
{
	/* Initialize our error exprs. */
	bad_arg.kind		= EXPR_KIND_ERROR;
	bad_arg.u.err		= "internal error (bad argument)";
	
	bad_op.kind		= EXPR_KIND_ERROR;
	bad_op.u.err		= "internal error (unknown operator)";
	
	no_mem.kind		= EXPR_KIND_ERROR;
	no_mem.u.err		= "out of memory";
	
	no_child.kind		= EXPR_KIND_ERROR;
	no_child.u.err		= "internal error (null subexpression)";
	
	math_error.kind		= EXPR_KIND_ERROR;
	math_error.u.err	= "math error";
	
	div_by_zero.kind	= EXPR_KIND_ERROR;
	div_by_zero.u.err	= "division by zero";
	
	/* Successful initialization.  Return 0 to indicate success. */
	return 0;
}

/*****************************************************************************/

static int
expr_kind_is_erroneous(expr_kind kind)
{
	switch (kind) {
	case EXPR_KIND_ERROR:
	case EXPR_KIND_ERROR_TOKEN:
		return TRUE;
	default:
		;
	}
	return FALSE;
}

static int
expr_kind_is_unary(expr_kind kind)
{
	switch (kind) {
	case EXPR_KIND_UPOS:
	case EXPR_KIND_UNEG:
		return TRUE;
	default:
		;
	}
	return FALSE;
}

static int
expr_kind_is_binary(expr_kind kind)
{
	switch (kind) {
	case EXPR_KIND_BADD:
	case EXPR_KIND_BSUB:
	case EXPR_KIND_BMUL:
	case EXPR_KIND_BDIV:
		return TRUE;
	default:
		;
	}
	return FALSE;
}


static expr
alloc_expr(void)
{
	return ((expr) malloc(sizeof(expr_struct)));
}

expr
alloc_error_no_mem_expr(void)
{
	return &no_mem;
}

/* The returned expression assumes ``ownership'' of the token `tok'. */
expr
alloc_error_token_expr(token tok)
{
	expr e;
	
	e = alloc_expr();
	if (!e) {
		free_token(tok);
		return &no_mem;
	}
	
	e->kind = EXPR_KIND_ERROR_TOKEN;
	e->u.tok = tok;
	return e;
}

/* The returned expression assumes ``ownership'' of the number `num'. */
expr
alloc_number_expr(number num)
{
	expr e;
	
	e = alloc_expr();
	if (!e) {
		free_number(num);
		return &no_mem;
	}
	
	e->kind = EXPR_KIND_NUMBER;
	e->u.num = num;
	return e;
}

/* The returned expression assumes ``ownership'' of the expr `child'. */
expr
alloc_unary_expr(expr_kind kind, expr child)
{
	expr e;
	
	if (!child)
		return &no_child;
	if (!expr_kind_is_unary(kind))
		return &bad_op;
	
	e = alloc_expr();
	if (!e) {
		free_expr(child);
		return &no_mem;
	}
	
	e->kind = kind;
	e->u.unary_sub = child;
	return e;
}

/* The returned expr assumes ``ownership'' of the exprs `left', `right'. */
expr
alloc_binary_expr(expr_kind kind, const expr left, const expr right)
{
	expr e;
	
	if (!left || !right) {
		if (left)
			free_expr(left);
		if (right)
			free_expr(right);
		return &no_child;
	}
	if (!expr_kind_is_binary(kind)) {
		free_expr(left);
		free_expr(right);
		return &bad_op;
	}
	
	e = alloc_expr();
	if (!e) {
		free_expr(left);
		free_expr(right);
		return &no_mem;
	}
	
	e->kind = kind;
	e->u.binary_sub.left = left;
	e->u.binary_sub.right = right;
	return e;
}

static expr
copy_error_token_expr(expr e)
{
	token tok;
	
	if (!e || (e->kind != EXPR_KIND_ERROR_TOKEN))
		return &bad_arg;
	
	tok = copy_token(e->u.tok);
	return alloc_error_token_expr(tok);
}

static expr
copy_number_expr(expr e)
{
	int rc;
	number out_num;
	
	rc = copy_number(e->u.num, &out_num);
	if (rc == FALSE)
		return &no_mem;
	return alloc_number_expr(out_num);
}

void
free_expr(expr e)
{
	if (!e)
		return;
	if (e->kind == EXPR_KIND_ERROR) {
		/* All error expressions are preallocated. */
		return;
	} else if (e->kind == EXPR_KIND_ERROR_TOKEN) {
		free_token(e->u.tok);
	} else if (e->kind == EXPR_KIND_NUMBER) {
		free_number(e->u.num);
	} else if (expr_kind_is_unary(e->kind)) {
		free_expr(e->u.unary_sub);
	} else if (expr_kind_is_binary(e->kind)) {
		free_expr(e->u.binary_sub.left);
		free_expr(e->u.binary_sub.right);
	}
	free(e);
}

int
expr_is_erroneous(expr e)
{
	if (!e)
		return FALSE;
	return expr_kind_is_erroneous(e->kind);
}

/*****************************************************************************/

expr
expr_eval(expr e)
{
	expr result;
	number result_num;
	
	expr child_result, left_result, right_result;
	number child_num, left_num, right_num;
	
	int rc;
	
	/*****/
	
	if (!e)
		return &no_child;
	
	switch (e->kind) {
	default:
		/* Should never get here! */
		result = &bad_op;
		break;
		
	case EXPR_KIND_ERROR:
		/* We don't need to make a fresh expression here, because
		   all errors are preallocated and never freed. */
		result = e;
		break;
		
	case EXPR_KIND_ERROR_TOKEN:
		/* Must make a fresh expression to return as result. */
		result = copy_error_token_expr(e);
		break;
		
	case EXPR_KIND_NUMBER:
		/* Must make a fresh expression to return as result. */
		result = copy_number_expr(e);
		break;
		
	case EXPR_KIND_UPOS:
		result = expr_eval(e->u.unary_sub);
		break;
	case EXPR_KIND_UNEG:
		child_result = expr_eval(e->u.unary_sub);
		if (child_result->kind != EXPR_KIND_NUMBER) {
			result = child_result;
			break;
		}
		child_num = child_result->u.num;
		free_expr(child_result);
		
		rc = number_neg(child_num, &result_num);
		if (rc == TRUE)
			result = alloc_number_expr(result_num);
		else
			result = &math_error;
		
		free_number(child_num);
		break;
		
	case EXPR_KIND_BADD:
	case EXPR_KIND_BSUB:
	case EXPR_KIND_BMUL:
	case EXPR_KIND_BDIV:
		left_result = expr_eval(e->u.binary_sub.left);
		if (left_result->kind != EXPR_KIND_NUMBER) {
			result = left_result;
			break;
		}
		left_num = left_result->u.num;
		free_expr(left_result);
		
		right_result = expr_eval(e->u.binary_sub.right);
		if (right_result->kind != EXPR_KIND_NUMBER) {
			result = right_result;
			/* Careful, must free `left_num'! */
			free_number(left_num);
			break;
		}
		right_num = right_result->u.num;
		free_expr(right_result);
		
		switch (e->kind) {
		case EXPR_KIND_BADD:
			rc = number_add(left_num, right_num, &result_num);
			break;
		case EXPR_KIND_BSUB:
			rc = number_sub(left_num, right_num, &result_num);
			break;
		case EXPR_KIND_BMUL:
			rc = number_mul(left_num, right_num, &result_num);
			break;
		case EXPR_KIND_BDIV:
			rc = number_div(left_num, right_num, &result_num);
			break;
		default:
			/* Unreachable. */
			rc = FALSE;
			break;
		}
		if (rc == TRUE)
			result = alloc_number_expr(result_num);
		else {
			if ((e->kind == EXPR_KIND_BDIV)
			    && number_is_zero(right_num))
				result = &div_by_zero;
			else
				result = &math_error;
		}
		
		free_number(left_num);
		free_number(right_num);
		break;
	}
	
	return result;
}

/*****************************************************************************/

static void expr_print_1(expr e, int precedence);
static void expr_print_1_unary(expr e, int precedence);
static void expr_print_1_binary(expr e, int precedence);

static void
expr_print_1(expr e, int precedence)
{
	if (!e)
		return;
	if (e->kind == EXPR_KIND_ERROR) {
		printf("error: %s", e->u.err);
	} else if (e->kind == EXPR_KIND_ERROR_TOKEN) {
		printf("scanner error: ");
		if (token_is_erroneous(e->u.tok))
			token_print(e->u.tok);
		else {
			printf("unexpected ");
			token_print(e->u.tok);
		}
	} else if (e->kind == EXPR_KIND_NUMBER) {
		number_print(e->u.num);
	} else if (expr_kind_is_unary(e->kind)) {
		expr_print_1_unary(e, precedence);
	} else if (expr_kind_is_binary(e->kind)) {
		expr_print_1_binary(e, precedence);
	}
}

static void
expr_print_1_unary(expr e, int precedence)
{
	if (precedence > 2)
		printf("(");
	printf("%c",
	       ((e->kind == EXPR_KIND_UPOS) ? '+' :
		(e->kind == EXPR_KIND_UNEG) ? '-' :
		' '));
	expr_print_1(e->u.unary_sub, 2);
	if (precedence > 2)
		printf(")");
}

static void
expr_print_1_binary(expr e, int precedence)
{
	int e_precedence;
	
	switch (e->kind) {
	case EXPR_KIND_BADD:
	case EXPR_KIND_BSUB:
		e_precedence = 0;
		break;
	case EXPR_KIND_BMUL:
	case EXPR_KIND_BDIV:
		e_precedence = 1;
		break;
	default:
		e_precedence = 0;
	}
	
	if (precedence >= e_precedence)
		printf("(");
	expr_print_1(e->u.binary_sub.left, (e_precedence - 1));
	printf(" %c ",
	       ((e->kind == EXPR_KIND_BADD) ? '+' :
		(e->kind == EXPR_KIND_BSUB) ? '-' :
		(e->kind == EXPR_KIND_BMUL) ? '*' :
		(e->kind == EXPR_KIND_BDIV) ? '/' :
		' '));
	expr_print_1(e->u.binary_sub.right, e_precedence);
	if (precedence >= e_precedence)
		printf(")");
}

void
expr_print(expr e)
{
	/* Start at operator precedence -1 to avoid any outermost ()'s. */
	expr_print_1(e, -1);
}

/*****************************************************************************/

/* End of file. */

