/*
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this file
 * for any purpose with or without restriction is hereby granted.
 */

#include "common.h"
#include "expr.h"
#include "parse.h"
#include "token.h"

static expr parse_sum(void);
static expr parse_product(void);
static expr parse_unary(void);
static expr parse_term(void);

/*****************************************************************************/

expr
parse(const char *str)
{
	expr e;
	token tok;
	
	start_tokenizer(str);
	e = parse_sum();
	if (expr_is_erroneous(e)) {
		end_tokenizer();
		return e;
	}
	tok = next_token();
	end_tokenizer();
	if (token_get_kind(tok) != TOKEN_KIND_EOF) {
		free_expr(e);
		return alloc_error_token_expr(tok);
	}
	free_token(tok);
	return e;
}

/*****************************************************************************/

static expr
parse_sum(void)
{
	expr		lhs;
	token		op_tok;
	expr_kind	op_expr_kind;
	expr		rhs;
	
	/*****/
	
	lhs = 0;
	for (;;) {
		/* If needed, parse the left hand side of the (binary) sum. */
		if (!lhs)
			lhs = parse_product();
		if (expr_is_erroneous(lhs))
			return lhs;
		
		/* Parse the operator. */
		op_tok = next_token();
		switch (token_get_kind(op_tok)) {
		case TOKEN_KIND_PLUS:
			op_expr_kind = EXPR_KIND_BADD;
			break;
		case TOKEN_KIND_MINUS:
			op_expr_kind = EXPR_KIND_BSUB;
			break;
		default:
			/* Push the token back onto the input. */
			push_token(op_tok);
			/* Return the parsed sum. */
			return lhs;
		}
		free_token(op_tok);
		
		/* Parse the right hand side of the (binary) sum. */
		rhs = parse_product();
		if (expr_is_erroneous(rhs)) {
			free_expr(lhs);
			return rhs;
		}
		
		/* Construct the parsed expression.  This becomes the left hand
		   side for the next iteration through the loop. */
		lhs = alloc_binary_expr(op_expr_kind, lhs, rhs);
	}
}

static expr
parse_product(void)
{
	expr		lhs;
	token		op_tok;
	expr_kind	op_expr_kind;
	expr		rhs;
	
	/*****/
	
	lhs = 0;
	for (;;) {
		/* If needed, parse the left hand side of the product. */
		if (!lhs)
			lhs = parse_unary();
		if (expr_is_erroneous(lhs))
			return lhs;
		
		/* Parse the operator. */
		op_tok = next_token();
		switch (token_get_kind(op_tok)) {
		case TOKEN_KIND_STAR:
			op_expr_kind = EXPR_KIND_BMUL;
			break;
		case TOKEN_KIND_SLASH:
			op_expr_kind = EXPR_KIND_BDIV;
			break;
		default:
			/* Push the token back onto the input. */
			push_token(op_tok);
			/* Return the parsed product. */
			return lhs;
		}
		free_token(op_tok);
		
		/* Parse the right hand side of the product. */
		rhs = parse_unary();
		if (expr_is_erroneous(rhs)) {
			free_expr(lhs);
			return rhs;
		}
		
		/* Construct the parsed expression.  This becomes the left hand
		   side for the next iteration through the loop. */
		lhs = alloc_binary_expr(op_expr_kind, lhs, rhs);
	}
}

static expr
parse_unary(void)
{
	token		op_tok;
	expr_kind	op_expr_kind;
	expr		inner_expr;
	
	/* Parse the operator. */
	op_tok = next_token();
	
	switch (token_get_kind(op_tok)) {
	default:
		/* Push the token back onto the input. */
		push_token(op_tok);
		/* Parse and return the underlying term. */
		return parse_term();
		
	case TOKEN_KIND_PLUS:
		op_expr_kind = EXPR_KIND_UPOS;
		break;
	case TOKEN_KIND_MINUS:
		op_expr_kind = EXPR_KIND_UNEG;
		break;
	}
	
	/* Recurse: parse the inner unary expression. */
	free_token(op_tok);
	inner_expr = parse_unary();
	if (expr_is_erroneous(inner_expr))
		return inner_expr;
	
	/* Return the unary expr applied to the inner expr. */
	return alloc_unary_expr(op_expr_kind, inner_expr);
}

static expr
parse_term(void)
{
	token		tok;
	int		rc;
	number		out_num;
	expr		sum;
	
	/*****/
	
	tok = next_token();
	switch (token_get_kind(tok)) {
	case TOKEN_KIND_NUMBER:
		/* Easy case: create a return a number expression. */
		rc = copy_number(token_get_number(tok), &out_num);
		free_token(tok);
		if (rc == FALSE)
			return alloc_error_no_mem_expr();
		return alloc_number_expr(out_num);
		
	case TOKEN_KIND_OPEN_PAREN:
		/* Another easy case: parse a parenthesized sum. */
		free_token(tok);
		sum = parse_sum();
		if (expr_is_erroneous(sum))
			return sum;
		
		tok = next_token();
		switch (token_get_kind(tok)) {
		case TOKEN_KIND_CLOSE_PAREN:
			free_token(tok);
			return sum;
		default:
			/* BUG: I forgot to free_expr(sum); */
			/* See the `README' file for story about this bug. */
			return alloc_error_token_expr(tok);
		}
		/* Not reached. */
		break;
		
	default:
		/* A third easy case: anything else is an error. */
		return alloc_error_token_expr(tok);
	}
}

/*****************************************************************************/

/* End of file. */

