/*
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this file
 * for any purpose with or without restriction is hereby granted.
 */

#include <stdio.h>

#include "common.h"
#include "expr.h"
#include "input.h"
#include "parse.h"
#include "repl.h"

static void print_phase(const char *phase);

/*****************************************************************************/

#ifndef ALLOC_COUNTS

/*
 * Stub macros.
 */
#define RESET_ALL_COUNTS()
#define PRINT_AND_RESET_PHASE_COUNTS(phase)
#define PRINT_AND_RESET_TOTAL_COUNTS()

#else /* ALLOC_COUNTS */

/*
 * The types of our reset-counts and get-counts functions.
 */
typedef void get_counts_func(unsigned int *mallocs, unsigned int *frees);
typedef void reset_counts_func(void);

/*
 * Our external reset- and get-counts functions.
 */
extern get_counts_func		get_token_phase_counts;
extern reset_counts_func	reset_token_phase_counts;

extern get_counts_func		get_expr_phase_counts;
extern reset_counts_func	reset_expr_phase_counts;

extern get_counts_func		get_token_total_counts;
extern reset_counts_func	reset_token_total_counts;

extern get_counts_func		get_expr_total_counts;
extern reset_counts_func	reset_expr_total_counts;

/*
 * Invoke all the reset-counts functions.
 */
static void
reset_all_counts(void)
{
	reset_token_phase_counts();
	reset_expr_phase_counts();
	reset_token_total_counts();
	reset_expr_total_counts();
}

/*
 * The generic print-and-reset-counts function.
 */
static void
print_and_reset_counts(
	const char *		phase,
	get_counts_func *	get_token_counts,
	reset_counts_func *	reset_token_counts,
	get_counts_func *	get_expr_counts,
	reset_counts_func *	reset_expr_counts)
{
	unsigned int token_mallocs;
	unsigned int token_frees;
	unsigned int expr_mallocs;
	unsigned int expr_frees;
	
	/*****/
	
	(*get_token_counts)(&token_mallocs, &token_frees);
	(*get_expr_counts)(&expr_mallocs, &expr_frees);
	
	print_phase(phase);
	printf("(allocs/frees) %2u/%2u tokens, %2u/%2u exprs\n",
	       token_mallocs,
	       token_frees,
	       expr_mallocs,
	       expr_frees);
	
	(*reset_token_counts)();
	(*reset_expr_counts)();
}

/*
 * Macros to print and reset counts for specific interpreter phases.
 */
#define RESET_ALL_COUNTS()					\
	reset_all_counts()
#define PRINT_AND_RESET_PHASE_COUNTS(phase)			\
	print_and_reset_counts((phase),				\
			       get_token_phase_counts,		\
			       reset_token_phase_counts,	\
			       get_expr_phase_counts,		\
			       reset_expr_phase_counts)
#define PRINT_AND_RESET_TOTAL_COUNTS()				\
	print_and_reset_counts("total",				\
			       get_token_total_counts,		\
			       reset_token_total_counts,	\
			       get_expr_total_counts,		\
			       reset_expr_total_counts)

#endif /* ALLOC_COUNTS */

/*****************************************************************************/

static void
print_phase(const char *phase)
{
	printf("%-8s: ", phase);
}

/* REPL == the ``read-eval-print loop.'' */
void
repl(void)
{
	int		rc;
	char *		str;
	expr		read_expr;
	expr		eval_expr;
	
	/*****/
	
	RESET_ALL_COUNTS();
	for (;;) {
		rc = get_input(&str);
		if (rc == FALSE)
			return;
		
		read_expr = parse(str);
		print_phase("read");
		expr_print(read_expr);
		printf("\n");
		PRINT_AND_RESET_PHASE_COUNTS("read");
		
		eval_expr = expr_eval(read_expr);
		print_phase("eval");
		expr_print(eval_expr);
		printf("\n");
		PRINT_AND_RESET_PHASE_COUNTS("eval");
		
		free_input(str);
		free_expr(read_expr);
		free_expr(eval_expr);
		PRINT_AND_RESET_PHASE_COUNTS("cleanup");
		
		PRINT_AND_RESET_TOTAL_COUNTS();
		printf("\n");
	}
}

/*****************************************************************************/

/* End of file. */

