/*
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this file
 * for any purpose with or without restriction is hereby granted.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "number.h"
#include "token.h"

typedef struct token_struct {
	token_kind kind;
	union {
		const char *	err;	/* Error message. */
		number		num;	/* Number value. */
	} u;
} token_struct;

/*****************************************************************************/

static struct {
	char *		str;		/* String given to `start_tokenizer' */
	const char *	pos;		/* Current read position in `str'    */
	token		saved_token;	/* Saved token from `push_token'     */
} tokenizer;

static token_struct bad_arg;
static token_struct bad_input;
static token_struct no_mem;
static token_struct scan_eof;

int
init_tokens(void)
{
	/* Initialize the tokenizer state. */
	tokenizer.str		= 0;
	tokenizer.pos		= 0;
	tokenizer.saved_token	= 0;
	
	/* Initialize our error tokens. */
	bad_arg.kind		= TOKEN_KIND_ERROR;
	bad_arg.u.err		= "internal error (bad argument)";
	
	bad_input.kind		= TOKEN_KIND_ERROR;
	bad_input.u.err		= "unrecognized input";
	
	no_mem.kind		= TOKEN_KIND_ERROR;
	no_mem.u.err		= "out of memory";
	
	scan_eof.kind		= TOKEN_KIND_EOF;
	
	/* Successful initialization.  Return 0 to indicate success. */
	return 0;
}

/*****************************************************************************/

void
start_tokenizer(const char *str)
{
	int str_len;
	char *tok_str;
	
	tokenizer.str		= 0;
	tokenizer.pos		= 0;
	tokenizer.saved_token	= 0;
	
	if (!str) {
		tokenizer.saved_token = &bad_arg;
		return;
	}
	
	str_len = strlen(str);
	tok_str = (char *) malloc(sizeof(char) * (str_len + 1));
	if (!tok_str) {
		tokenizer.saved_token = &no_mem;
		return;
	}
	strncpy(tok_str, str, (str_len + 1));
	
	tokenizer.str		= tok_str;
	tokenizer.pos		= tok_str;
	tokenizer.saved_token	= 0;
}

void
end_tokenizer(void)
{
	if (tokenizer.str)
		free(tokenizer.str);
	tokenizer.str = 0;
	tokenizer.pos = 0;
	
	if (tokenizer.saved_token)
		free_token(tokenizer.saved_token);
	tokenizer.saved_token = 0;
}

/*****************************************************************************/

static token
alloc_token(void)
{
	return ((token) malloc(sizeof(token_struct)));
}

static token
alloc_simple_token(token_kind kind)
{
	token t;
	
	t = alloc_token();
	if (!t)
		return &no_mem;
	t->kind = kind;
	return t;
}

/* The returned expression assumes ``ownership'' of the number `num'. */
static token
alloc_number_token(number num)
{
	token t;
	
	t = alloc_token();
	if (!t) {
		free_number(num);
		return &no_mem;
	}
	t->kind = TOKEN_KIND_NUMBER;
	t->u.num = num;
	return t;
}

token
copy_token(token t)
{
	int rc;
	number out_num;
	
	if (!t)
		return &bad_arg;
	if (t->kind == TOKEN_KIND_ERROR) {
		/* Just return the same token. */
		return t;
	} else if (t->kind == TOKEN_KIND_EOF) {
		/* Just return the same token. */
		return t;
	} else if (t->kind == TOKEN_KIND_NUMBER) {
		rc = copy_number(t->u.num, &out_num);
		if (rc == FALSE)
			return &no_mem;
		return alloc_number_token(out_num);
	}
	return alloc_simple_token(t->kind);
}

void
free_token(token t)
{
	if (!t)
		return;
	if (t->kind == TOKEN_KIND_ERROR) {
		/* All error tokens are preallocated. */
		return;
	} else if (t->kind == TOKEN_KIND_EOF) {
		/* All EOF tokens are preallocated. */
		return;
	} else if (t->kind == TOKEN_KIND_NUMBER) {
		free_number(t->u.num);
	}
	free(t);
}

/*****************************************************************************/

int
token_is_erroneous(token t)
{
	if (!t)
		return FALSE;
	switch (t->kind) {
	case TOKEN_KIND_ERROR:
		return TRUE;
	default:
		;
	}
	return FALSE;
}

token_kind
token_get_kind(token t)
{
	return t->kind;
}

number
token_get_number(token t)
{
	return t->u.num;
}

/*****************************************************************************/

/*
 * This is our version of the standard C library `isspace' function.
 *
 * We use this instead of `isspace' because we're trying to make a portable
 * example for Knit.  If `isspace' is a macro, Knit needs to know about the
 * symbols that get pulled in ``under the covers,'' and that set depends on
 * which C library you use.  So, we avoid that mess and use our own function.
 */
static int
is_space(char c)
{
	switch (c) {
	case ' ':
	case '\f':
	case '\n':
	case '\r':
	case '\t':
	case '\v':
		return TRUE;
	default:
		;
	}
	return FALSE;
}

token
next_token(void)
{
	token t;
	const char *s;
	number num;
	const char *num_end;
	int rc;
	
	/*****/
	
	t = tokenizer.saved_token;
	if (t) {
		tokenizer.saved_token = 0;
		return t;
	}
	
	s = tokenizer.pos;
	if (!s)
		return &scan_eof;
	
	/* Skip forward over whitespace. */
	while (*s && is_space(*s))
		++s;
	if (!(*s)) {
		tokenizer.pos = s;
		return &scan_eof;
	}
	
	/* Recognize operators and grouping. */
	switch (*s) {
	case '+':
		tokenizer.pos = s + 1;
		return alloc_simple_token(TOKEN_KIND_PLUS);
	case '-':
		tokenizer.pos = s + 1;
		return alloc_simple_token(TOKEN_KIND_MINUS);
	case '*':
		tokenizer.pos = s + 1;
		return alloc_simple_token(TOKEN_KIND_STAR);
	case '/':
		tokenizer.pos = s + 1;
		return alloc_simple_token(TOKEN_KIND_SLASH);
	case '(':
		tokenizer.pos = s + 1;
		return alloc_simple_token(TOKEN_KIND_OPEN_PAREN);
	case ')':
		tokenizer.pos = s + 1;
		return alloc_simple_token(TOKEN_KIND_CLOSE_PAREN);
	}
	
	/* Recognize numbers. */
	rc = string_to_number(s, &num, &num_end);
	if (rc == TRUE) {
		tokenizer.pos = num_end;
		return alloc_number_token(num);
	}
	
	/* Otherwise, it's a scanner error! */
	tokenizer.pos = s;
	return &bad_input;
}

void
push_token(token t)
{
	if (tokenizer.saved_token)
		/* This shouldn't happen! */
		free_token(tokenizer.saved_token);
	tokenizer.saved_token = t;
}

/*****************************************************************************/

void
token_print(token t)
{
	const char *str;
	
	if (!t)
		return;
	str = 0;
	switch (t->kind) {
	case TOKEN_KIND_ERROR:
		str = t->u.err;
		break;
	case TOKEN_KIND_EOF:
		str = "end of input";
		break;
		
	case TOKEN_KIND_NUMBER:
		number_print(t->u.num);
		break;
		
	case TOKEN_KIND_PLUS:
		str = "+";
		break;
	case TOKEN_KIND_MINUS:
		str = "-";
		break;
	case TOKEN_KIND_STAR:
		str = "*";
		break;
	case TOKEN_KIND_SLASH:
		str = "/";
		break;
	case TOKEN_KIND_OPEN_PAREN:
		str = "(";
		break;
	case TOKEN_KIND_CLOSE_PAREN:
		str = ")";
		break;
	default:
		str = "(unknown token kind!)";
		break;
	}
	if (str)
		printf("%s", str);
}

/*****************************************************************************/

/* End of file. */

