;; knit.el --- define coding styles for Knit project source code
;;
;; Copyright (c) 2000, 2001 University of Utah and the Flux Group.
;; All rights reserved.
;; 
;; This file is part of the Knit component composition software.
;; 
;; Knit is free software, also known as "open source."  You can redistribute
;; it and/or modify it under the terms of the GNU General Public License
;; (GPL), version 2, as published by the Free Software Foundation (FSF).
;; 
;; Knit is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
;; received a copy of the GPL along with Knit; see the file COPYING.  If
;; not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
;; 
;; Users are requested, but not required, to send to csl-dist@cs.utah.edu any
;; improvements that they make and grant Univ. of Utah redistribution rights.
;;

;; Bug reports to Eric Eide <eeide@cs.utah.edu>.

(require 'cc-mode)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar knit-filename-regexp "/knit/"
  "*The regular expression that matches the absolute pathnames of Knit
source files.  If the name of a source file matches this regular expression,
then the C coding style will be set by the function `knit-set-style'.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Define the "knit-c" coding style, for Knit project C code.
;;;
;;; Some of these values are up to debate.  I expect that this specification
;;; will be refined over time.
;;;
;;; This is similar to (but not identical to) the "flick" coding style.
;;; It is identical to the Janos "oskit" coding style.

(c-add-style "knit-c"
	     '(
	       "gnu"
	       (c-basic-offset . 8)
	       (c-offsets-alist
		. (
		   (statement-cont . c-lineup-math-if-nonempty)
		   (arglist-intro . c-lineup-arglist-intro-after-paren-or-+)
		   ))
	       (c-comment-only-line-offset . 0)
	       (c-indent-comments-syntactically-p . t)
	       ;; Obsolete: (c-block-comments-indent-p . nil)
	       (c-comment-continuation-stars . "* ")
	       (c-hanging-comment-starter-p . nil)
	       (c-hanging-comment-ender-p . nil)
	       (c-hanging-braces-alist . (
					  ;; (defun-open) want default.
					  ;; (defun-close) want default.
					  (class-open after)
					  ;; (class-close) want default.
					  ;; (inline-open) want default.
					  ;; (inline-close) want default.
					  (block-open after)
					  ;; (block-close) want default.
					  (substatement-open after)
					  (statement-case-open after)
					  (brace-list-open)
					  (brace-list-close)
					  (brace-list-intro)
					  (brace-list-entry)
					  ))
	       (c-cleanup-list . (brace-else-brace
				  defun-close-semi
				  list-close-comma
				  scope-operator))
	       )
	     )

;;; Define the "knit" coding style for Knit unit files.
;;;
;;; XXX --- This is totally icky.
;;;
;;; Stolen from the Flick style, with minor modes to improve autoindent.
;;;
;;; Big change is in `c-hanging-braces-alist':
;;; XXX --- Need to double-check these.  I probably screwed some of them up!
;;;
;;;  + (brace-list-open after) for newline after `unit foo = {'.
;;;  + (brace-list-intro after) for newline after unit clauses `foo {'
;;;  + (brace-list-close before) for newline before closing `}'
;;;
;;; Should also hack on `cc-engine' to find where it looks for brace-lists,
;;; e.g., case 5A.3 of function `c-guess-basic-syntax'.  Hack it for our brace
;;; lists, e.g., files, by adding (looking-at "files[ \t\n]+") in the right
;;; proper place.  Others?

(c-add-style "knit"
	     '(
	       "gnu"
	       (c-basic-offset . 4)
	       (c-offsets-alist . (
				   (statement-cont . c-lineup-math-if-nonempty)
				   ))
	       (c-comment-only-line-offset . 0)
	       (c-indent-comments-syntactically-p . t)
	       ;; Obsolete: (c-block-comments-indent-p . nil)
	       (c-comment-continuation-stars . "* ")
	       (c-hanging-comment-starter-p . nil)
	       (c-hanging-comment-ender-p . nil)
	       (c-hanging-braces-alist . (
					  ;; (defun-open) want default.
					  ;; (defun-close) want default.
					  (class-open after)
					  ;; (class-close) want default.
					  ;; (inline-open) want default.
					  ;; (inline-close) want default.
					  (block-open after)
					  ;; (block-close) want default.
					  (substatement-open after)
					  (statement-case-open after)
					  (brace-list-open after)
					  (brace-list-close before)
					  (brace-entry-open after)
					  (brace-list-intro after)
ll					  (brace-list-entry before after)
					  ))
	       (c-cleanup-list . (brace-else-brace
				  defun-close-semi
				  list-close-comma
				  scope-operator))
	       )
	     )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Add the hook function to set the coding styles for Knit source files.

(defun knit-set-style ()
  "Set the CC Mode codeing style for Knit sources.
Use the \"knit\" style for Knit unit files.
Use the \"knit-c\" style for Knit C/C++/... files."
  (interactive)
  (save-match-data
    (if (and (stringp buffer-file-name)
	     (stringp knit-filename-regexp)
	     (string-match knit-filename-regexp buffer-file-name))
	(progn
	  (cond ((eq major-mode 'knit-mode)
		 (c-set-style "knit"))
		(t
		 (c-set-style "knit-c"))))
      )))

(add-hook 'c-mode-common-hook
  (function knit-set-style)
  ;; For now, append this to `c-mode-common-hook' so that other style-munging
  ;; stuff happens before this function.  This allows us to override any other
  ;; style customizations that people may have.
  'append)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun c-lineup-arglist-intro-after-paren-or-+ (langelem)
  "Like function `c-lineup-arglist-intro-after-paren', but won't indent more
than variable `c-basic-offset'.

Works with: defun-block-intro, brace-list-intro,
statement-block-intro, statement-case-intro, arglist-intro."
  (min (c-lineup-arglist-intro-after-paren langelem)
       c-basic-offset))

;;; XXX --- I think that this is obsolete!

(defun c-lineup-math-if-nonempty (langelem)
  ;; line up math statement-cont after the equals,
  ;; but only if there is something after the `=' on the statement line.
  (save-excursion
    (let* ((relpos (cdr langelem))
	   (equalp (save-excursion
		     (goto-char (c-point 'boi))
		     (skip-chars-forward "^=" (c-point 'eol))
		     (and (= (following-char) ?=)
			  (- (point) (c-point 'boi)))))
	   (curcol (progn
		     (goto-char relpos)
		     (current-column)))
	   donep)
      (while (and (not donep)
		  (< (point) (c-point 'eol)))
	(skip-chars-forward "^=" (c-point 'eol))
	(if (c-in-literal (cdr langelem))
	    (forward-char 1)
	  (setq donep t)))
      (if (or (/= (following-char) ?=)
	      (save-excursion (skip-chars-forward "= \t" (c-point 'eol))
			      (eolp)))
	  ;; there's no equal sign on the line,
	  ;; or there's nothing after the equal sign.
	  c-basic-offset
	;; calculate indentation column after equals and ws, unless
	;; our line contains an equals sign
	(if (not equalp)
	    (progn
	      (forward-char 1)
	      (skip-chars-forward " \t")
	      (setq equalp 0)))
	(- (current-column) equalp curcol))
      )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; End of file.

